// -*- mode: C++; -*-

/* LibGM: A Geometrical Library
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef GM_MATRIX3_H
#define GM_MATRIX3_H

#include <assert.h>

#include "gm_utils.h"
#include "gm_vector2.h"
#include "gm_point2.h"
#include "gm_vector3.h"
#include "gm_point3.h"


class gmMatrix3 {
public:
  gmMatrix3 ();
  gmMatrix3 (const gmMatrix3&);
  gmMatrix3 (double, double, double,
             double, double, double,
             double, double, double);

  // array access

  double* operator[] (int);
  const double* operator[] (int) const;

  // assignment

  gmMatrix3& assign (double, double, double,
                     double, double, double,
                     double, double, double);
  gmMatrix3& operator= (const gmMatrix3&);

  // operators

  gmMatrix3& operator+= (const gmMatrix3&);
  gmMatrix3& operator-= (const gmMatrix3&);
  gmMatrix3& operator*= (const gmMatrix3&);
  gmMatrix3& operator*= (double);
  gmMatrix3& operator/= (double);

  gmMatrix3 operator+ (const gmMatrix3&) const;
  gmMatrix3 operator- (const gmMatrix3&) const;
  gmMatrix3 operator- () const;
  gmMatrix3 operator* (const gmMatrix3&) const;
  gmMatrix3 operator* (double) const;
  gmMatrix3 operator/ (double) const;

  friend gmMatrix3 operator* (double, const gmMatrix3&);

  gmBool operator== (const gmMatrix3&) const;
  gmBool operator!= (const gmMatrix3&) const;

  gmVector3 operator* (const gmVector3&) const;
  gmPoint3 operator* (const gmPoint3&) const;

  friend gmVector3 operator* (const gmVector3&, const gmMatrix3&);
  friend gmPoint3 operator* (const gmPoint3&, const gmMatrix3&);

  // operations

  gmMatrix3 inverse () const;
  gmMatrix3 transposed () const;
  gmMatrix3 adjoint () const;

  double determinant () const;
  gmBool isSingular () const;

  gmVector2 transform (const gmVector2&) const;
  gmPoint2 transform (const gmPoint2&) const;

  void copyTo (float [3][3]) const;
  void copyTo (double [3][3]) const;

  // transformation matrices

  static gmMatrix3 identity ();
  static gmMatrix3 rotation (double angle);
  static gmMatrix3 scaling (double x, double y);
  static gmMatrix3 translation (double x, double y);
  static gmMatrix3 shearing (double a, double b);

  // output

  friend ostream& operator<< (ostream&, const gmMatrix3&);

protected:
  double m_[3][3];
};


// array access

inline double* gmMatrix3::operator[] (int i)
{
  assert(i == 0 || i == 1 || i == 2 || i == 3);
  return &m_[i][0];
}


inline const double* gmMatrix3::operator[] (int i) const
{
  assert(i == 0 || i == 1 || i == 2 || i == 3);
  return &m_[i][0];
}


inline void gmMatrix3::copyTo (float f[3][3]) const
{
  f[0][0] = m_[0][0]; f[0][1] = m_[0][1]; f[0][2] = m_[0][2];
  f[1][0] = m_[1][0]; f[1][1] = m_[1][1]; f[1][2] = m_[1][2];
  f[2][0] = m_[2][0]; f[2][1] = m_[2][1]; f[2][2] = m_[2][2];
}


inline void gmMatrix3::copyTo (double d[3][3]) const
{
  d[0][0] = m_[0][0]; d[0][1] = m_[0][1]; d[0][2] = m_[0][2];
  d[1][0] = m_[1][0]; d[1][1] = m_[1][1]; d[1][2] = m_[1][2];
  d[2][0] = m_[2][0]; d[2][1] = m_[2][1]; d[2][2] = m_[2][2];
}


#endif // GM_MATRIX3_H
