// -*- mode: C++; -*-

/* LibGM: A Geometrical Library
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef GM_MATRIX4_H
#define GM_MATRIX4_H

#include <assert.h>

#include "gm_utils.h"
#include "gm_vector3.h"
#include "gm_point3.h"
#include "gm_vector4.h"
#include "gm_point4.h"

class gmMatrix4 {
public:
  gmMatrix4 ();
  gmMatrix4 (const gmMatrix4&);
  gmMatrix4 (double, double, double, double,
             double, double, double, double,
             double, double, double, double,
             double, double, double, double);

  // array access

  double* operator[] (int);
  const double* operator[] (int) const;

  // assignment

  gmMatrix4& assign (double, double, double, double,
                     double, double, double, double,
                     double, double, double, double,
                     double, double, double, double);
  gmMatrix4& operator= (const gmMatrix4&);

  // operators

  gmMatrix4& operator+= (const gmMatrix4&);
  gmMatrix4& operator-= (const gmMatrix4&);
  gmMatrix4& operator*= (const gmMatrix4&);
  gmMatrix4& operator*= (double);
  gmMatrix4& operator/= (double);

  gmMatrix4 operator+ (const gmMatrix4&) const;
  gmMatrix4 operator- (const gmMatrix4&) const;
  gmMatrix4 operator- () const;
  gmMatrix4 operator* (const gmMatrix4&) const;
  gmMatrix4 operator* (double) const;
  gmMatrix4 operator/ (double) const;

  friend gmMatrix4 operator* (double, const gmMatrix4&);

  gmBool operator== (const gmMatrix4&) const;
  gmBool operator!= (const gmMatrix4&) const;

  gmVector4 operator* (const gmVector4&) const;
  gmPoint4 operator* (const gmPoint4&) const;

  friend gmVector4 operator* (const gmVector4&, const gmMatrix4&);
  friend gmPoint4 operator* (const gmPoint4&, const gmMatrix4&);

  // operations

  gmMatrix4 inverse () const;
  gmMatrix4 transposed () const;
  gmMatrix4 adjoint () const;

  double determinant () const;
  gmBool isSingular () const;

  gmVector3 transform (const gmVector3&) const;
  gmPoint3 transform (const gmPoint3&) const;

  void copyTo (float [4][4]) const;
  void copyTo (double [4][4]) const;

  // transformation matrices

  static gmMatrix4 identity ();
  static gmMatrix4 rotation (double angle, const gmVector3& axis);
  static gmMatrix4 scaling (double x, double y, double z);
  static gmMatrix4 scaling (double factor, gmVector3 dir);
  static gmMatrix4 translation (double x, double y, double z);
  static gmMatrix4 shearing (double a, double b, double c,
                             double d, double e, double f);

  // cubic basis matrices

  static gmMatrix4 bezierBasis ();
  static gmMatrix4 bsplineBasis ();
  static gmMatrix4 catmullromBasis ();
  static gmMatrix4 hermiteBasis ();

  static gmMatrix4 tensedBSplineBasis (double);
  static gmMatrix4 cardinalBasis (double);
  static gmMatrix4 tauBasis (double, double);
  static gmMatrix4 betaSplineBasis (double, double);

  // output

  friend ostream& operator<< (ostream&, const gmMatrix4&);

protected:
  double m_[4][4];
};


// array access

inline double* gmMatrix4::operator[] (int i)
{
  assert(i == 0 || i == 1 || i == 2 || i == 3);
  return &m_[i][0];
}


inline const double* gmMatrix4::operator[] (int i) const
{
  assert(i == 0 || i == 1 || i == 2 || i == 3);
  return &m_[i][0];
}


inline void gmMatrix4::copyTo (float f[4][4]) const
{
  f[0][0] = m_[0][0];
  f[0][1] = m_[0][1];
  f[0][2] = m_[0][2];
  f[0][3] = m_[0][3];

  f[1][0] = m_[1][0];
  f[1][1] = m_[1][1];
  f[1][2] = m_[1][2];
  f[1][3] = m_[1][3];

  f[2][0] = m_[2][0];
  f[2][1] = m_[2][1];
  f[2][2] = m_[2][2];
  f[2][3] = m_[2][3];

  f[3][0] = m_[3][0];
  f[3][1] = m_[3][1];
  f[3][2] = m_[3][2];
  f[3][3] = m_[3][3];
}


inline void gmMatrix4::copyTo (double d[4][4]) const
{
  d[0][0] = m_[0][0];
  d[0][1] = m_[0][1];
  d[0][2] = m_[0][2];
  d[0][3] = m_[0][3];

  d[1][0] = m_[1][0];
  d[1][1] = m_[1][1];
  d[1][2] = m_[1][2];
  d[1][3] = m_[1][3];

  d[2][0] = m_[2][0];
  d[2][1] = m_[2][1];
  d[2][2] = m_[2][2];
  d[2][3] = m_[2][3];

  d[3][0] = m_[3][0];
  d[3][1] = m_[3][1];
  d[3][2] = m_[3][2];
  d[3][3] = m_[3][3];
}


#endif // GM_MATRIX4_H
