// -*- mode: C++; -*-

/* LibGM: A Geometrical Library
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef GM_POINT2_H
#define GM_POINT2_H

#include <math.h>
#include <assert.h>
#include <iostream.h>

#include "gm_utils.h"
#include "gm_vector2.h"


class gmPoint2 {
public:
  gmPoint2 ();
  gmPoint2 (const gmPoint2&);
  gmPoint2 (double, double);

  // array access

  double& operator[] (int);
  const double& operator[] (int) const;

  // assignment

  gmPoint2& assign (double, double);
  gmPoint2& operator= (const gmPoint2&);

  // math operators

  gmPoint2& operator+= (const gmVector2&);

  gmPoint2 operator+ (const gmVector2&) const;
  gmVector2 operator- (const gmPoint2&) const;

  friend gmPoint2 operator+ (const gmVector2&, const gmPoint2&);

  gmBool operator== (const gmPoint2&) const;
  gmBool operator!= (const gmPoint2&) const;

  // operations

  gmPoint2& clamp (double, double);
  double originDistance () const;
  double originDistanceSquared () const;

  void copyTo (float [2]) const;
  void copyTo (double [2]) const;

  friend double gmDistance (const gmPoint2&, const gmPoint2&);
  friend double gmDistanceSquared (const gmPoint2&, const gmPoint2&);
  friend gmPoint2 gmLerp (double, const gmPoint2&, const gmPoint2&);

  // output

  friend ostream& operator<< (ostream&, const gmPoint2&);

protected:
  double p_[2];
};


// constructors

inline gmPoint2::gmPoint2 ()
{
  p_[0] = p_[1] = 0;
}


inline gmPoint2::gmPoint2 (const gmPoint2& p)
{
  p_[0] = p.p_[0];
  p_[1] = p.p_[1];
}


inline gmPoint2::gmPoint2 (double x, double y)
{
  p_[0] = x;
  p_[1] = y;
}


// array access

inline double& gmPoint2::operator[] (int i)
{
  assert(i == 0 || i == 1);
  return p_[i];
}

inline const double& gmPoint2::operator[] (int i) const
{
  assert(i == 0 || i == 1);
  return p_[i];
}


// assignment

inline gmPoint2& gmPoint2::assign (double x, double y)
{
  p_[0] = x;
  p_[1] = y;
  return *this;
}


inline gmPoint2& gmPoint2::operator= (const gmPoint2& p)
{
  p_[0] = p[0];
  p_[1] = p[1];
  return *this;
}


// math operators

inline gmPoint2& gmPoint2::operator+= (const gmVector2& v)
{
  p_[0] += v[0];
  p_[1] += v[1];
  return *this;
}


inline gmPoint2 gmPoint2::operator+ (const gmVector2& v) const
{
  return gmPoint2(p_[0] + v[0], p_[1] + v[1]);
}


inline gmVector2 gmPoint2::operator- (const gmPoint2& p) const
{
  return gmVector2(p_[0] - p[0], p_[1] - p[1]);
}


inline gmPoint2 operator+ (const gmVector2& v, const gmPoint2& p)
{
  return gmPoint2(p[0] + v[0], p[1] + v[1]);
}


inline gmBool gmPoint2::operator== (const gmPoint2& p) const
{
  return (gmFuzEQ(p_[0], p[0]) && gmFuzEQ(p_[1], p[1]));
}


inline gmBool gmPoint2::operator!= (const gmPoint2& p) const
{
  return (!(*this == p));
}


// operations

inline gmPoint2& gmPoint2::clamp (double lo, double hi)
{
  gmClamp(p_[0], lo, hi);
  gmClamp(p_[1], lo, hi);
  return *this;
}


inline double gmPoint2::originDistance () const
{
  return sqrt(gmSqr(p_[0]) + gmSqr(p_[1]));
}


inline double gmPoint2::originDistanceSquared () const
{
  return gmSqr(p_[0]) + gmSqr(p_[1]);
}


inline void gmPoint2::copyTo (float f[2]) const
{
  f[0] = p_[0];
  f[1] = p_[1];
}


inline void gmPoint2::copyTo (double d[2]) const
{
  d[0] = p_[0];
  d[1] = p_[1];
}


inline double gmDistance (const gmPoint2& p1, const gmPoint2& p2)
{
  return sqrt(gmSqr(p1[0] - p2[0]) + gmSqr(p1[1] - p2[1]));
}


inline double gmDistanceSquared (const gmPoint2& p1, const gmPoint2& p2)
{
  return gmSqr(p1[0] - p2[0]) + gmSqr(p1[1] - p2[1]);
}


inline gmPoint2 gmLerp (double f, const gmPoint2& p1, const gmPoint2& p2)
{
  return p1 + ((p2 - p1) * f);
}


// output

inline ostream& operator<< (ostream& os, const gmPoint2& p)
{
  os << '(' << p[0] << ',' << p[1] << ')';
  return os;
}


#endif // GM_POINT2_H
