/* LibGM: A Geometrical Library
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "gm_point3.h"


gmPoint3 gmSph2Cart (double rho, double theta, double phi)
{
  double cp, sp, ct, st;

  gmClamp(theta, 0.0, 180.0);
  theta=gmRadians(theta);
  st=sin(theta);
  ct=cos(theta);

  phi=gmRadians(phi);
  sp=sin(phi);
  cp=cos(phi);

  return gmPoint3(rho*st*sp, rho*ct, rho*st*cp);
}


void gmCart2Sph (gmPoint3 p, double *rho, double *theta, double *phi)
{
  *rho=p.originDistance();
  if (gmAbs(*rho) < GM_EPSILON) {
    *theta=0.0;
    *phi=0.0;
    return;
  }

  double tmp=acos(p[1]/(*rho));
  *theta=gmDegrees(tmp);

  tmp=(*rho)*sin(tmp);
  if (gmAbs(tmp) < GM_EPSILON)
    *phi=0.0;
  else
    *phi=360.0-gmDegrees(acos(p[2]/tmp));
}


double gmSignedDistance (gmPoint3 p, gmVector3 n, double d)
{
  n.normalize();
  return n[0]*p[0]+n[1]*p[1]+n[2]*p[2]+d;
}


gmBool gmIntersectPlane (gmPoint3 p, gmVector3 v, gmVector3 n, double d,
                         gmPoint3 *i)
{
  double t=gmDot(n, v);
  if (gmAbs(t) < GM_EPSILON)
    return FALSE;
  t=-(n[0]*p[0]+n[1]*p[1]+n[2]*p[2]+d)/t;
  (*i)=p+t*v;
  return TRUE;
}


void gmIntersectSphere (gmPoint3 p, gmVector3 v, gmPoint3 c, double r,
                        gmBool front, gmPoint3 *i)
{
  gmVector3 a;
  double b, d;

  a=c-p;
  v.normalize();
  b=gmDot(a, v);
  d=r*r-a.lengthSquared()+b*b;
  if (d >= 0.0) {
    if (front)
      *i=(b-sqrt(d))*v+p; // hit: get nearest intersection
    else
      *i=(b+sqrt(d))*v+p; // hit: get farthest intersection
  }
  else
    *i=r*(b*v-a).normalize()+c; // missed sphere: get nearest point
}


gmBool gmCalcPlane (gmPoint3 p1, gmPoint3 p2, gmPoint3 p3,
                    gmVector3 *n, double *d)
{
  gmVector3 nn=gmCross(p3-p2, p2-p1);
  double m=nn.length();
  if (gmAbs(m) < GM_EPSILON)
    return FALSE;
  *n=nn/m;
  *d=-((*n)[0]*p1[0]+(*n)[1]*p1[1]+(*n)[2]*p1[2]);
  return TRUE;
}


gmBool gmNearestPoint (gmPoint3 p1, gmVector3 v1, gmPoint3 p2, gmVector3 v2,
                       gmPoint3 *i)
{
  gmVector3 v3=gmCross(v1, v2);
  if (gmAbs(v3.lengthSquared()) < GM_EPSILON2)
    return FALSE;
  gmVector3 n=gmCross(v2, v3);
  if (!gmIntersectPlane(p1, v1, n, -(n[0]*p2[0]+n[1]*p2[1]+n[2]*p2[2]), i))
    return FALSE;
  return TRUE;
}
