// -*- mode: C++; -*-

/* LibGM: A Geometrical Library
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef GM_POINT3_H
#define GM_POINT3_H

#include <math.h>
#include <assert.h>
#include <iostream.h>

#include "gm_utils.h"
#include "gm_vector3.h"


class gmPoint3 {
public:
  gmPoint3 ();
  gmPoint3 (const gmPoint3&);
  gmPoint3 (double, double, double);

  // array access

  double& operator[] (int);
  const double& operator[] (int) const;

  // assignment

  gmPoint3& assign (double, double, double);
  gmPoint3& operator= (const gmPoint3&);

  // math operators

  gmPoint3& operator+= (const gmVector3&);

  gmPoint3 operator+ (const gmVector3&) const;
  gmVector3 operator- (const gmPoint3&) const;

  friend gmPoint3 operator+ (const gmVector3&, const gmPoint3&);

  gmBool operator== (const gmPoint3&) const;
  gmBool operator!= (const gmPoint3&) const;

  // operations

  gmPoint3& clamp (double, double);
  double originDistance () const;
  double originDistanceSquared () const;

  void copyTo (float [3]) const;
  void copyTo (double [3]) const;

  friend double gmDistance (const gmPoint3&, const gmPoint3&);
  friend double gmDistanceSquared (const gmPoint3&, const gmPoint3&);
  friend gmPoint3 gmLerp (double, const gmPoint3&, const gmPoint3&);

  // output

  friend ostream& operator<< (ostream&, const gmPoint3&);

protected:
  double p_[3];
};


// constructors

inline gmPoint3::gmPoint3 ()
{
  p_[0] = p_[1] = p_[2] = 0;
}


inline gmPoint3::gmPoint3 (const gmPoint3& p)
{
  p_[0] = p.p_[0];
  p_[1] = p.p_[1];
  p_[2] = p.p_[2];
}


inline gmPoint3::gmPoint3 (double x, double y, double z)
{
  p_[0] = x;
  p_[1] = y;
  p_[2] = z;
}


// array access

inline double& gmPoint3::operator[] (int i)
{
  assert(i == 0 || i == 1 || i == 2);
  return p_[i];
}


inline const double& gmPoint3::operator[] (int i) const
{
  assert(i == 0 || i == 1 || i == 2);
  return p_[i];
}


// assignment

inline gmPoint3& gmPoint3::assign (double x, double y, double z)
{
  p_[0] = x;
  p_[1] = y;
  p_[2] = z;
  return *this;
}


inline gmPoint3& gmPoint3::operator= (const gmPoint3& p)
{
  p_[0] = p[0];
  p_[1] = p[1];
  p_[2] = p[2];
  return *this;
}


// math operators

inline gmPoint3& gmPoint3::operator+= (const gmVector3& v)
{
  p_[0] += v[0];
  p_[1] += v[1];
  p_[2] += v[2];
  return *this;
}


inline gmPoint3 gmPoint3::operator+ (const gmVector3& v) const
{
  return gmPoint3(p_[0] + v[0], p_[1] + v[1], p_[2] + v[2]);
}


inline gmVector3 gmPoint3::operator- (const gmPoint3& p) const
{
  return gmVector3(p_[0] - p[0], p_[1] - p[1], p_[2] - p[2]);
}


inline gmPoint3 operator+ (const gmVector3& v, const gmPoint3& p)
{
  return gmPoint3(p[0] + v[0], p[1] + v[1], p[2] + v[2]);
}


inline gmBool gmPoint3::operator== (const gmPoint3& p) const
{
  return (gmFuzEQ(p_[0], p[0]) && gmFuzEQ(p_[1], p[1]) &&
          gmFuzEQ(p_[2], p[2]));
}


inline gmBool gmPoint3::operator!= (const gmPoint3& p) const
{
  return (!(*this == p));
}


// operations

inline gmPoint3& gmPoint3::clamp (double lo, double hi)
{
  gmClamp(p_[0], lo, hi);
  gmClamp(p_[1], lo, hi);
  gmClamp(p_[2], lo, hi);
  return *this;
}


inline double gmPoint3::originDistance () const
{
  return sqrt(gmSqr(p_[0]) + gmSqr(p_[1]) + gmSqr(p_[2]));
}


inline double gmPoint3::originDistanceSquared () const
{
  return gmSqr(p_[0]) + gmSqr(p_[1]) + gmSqr(p_[2]);
}


inline void gmPoint3::copyTo (float f[3]) const
{
  f[0] = p_[0];
  f[1] = p_[1];
  f[2] = p_[2];
}


inline void gmPoint3::copyTo (double d[3]) const
{
  d[0] = p_[0];
  d[1] = p_[1];
  d[2] = p_[2];
}


inline double gmDistance (const gmPoint3& p1, const gmPoint3& p2)
{
  return sqrt(gmSqr(p1[0] - p2[0]) + gmSqr(p1[1] - p2[1]) +
              gmSqr(p1[2] - p2[2]));
}


inline double gmDistanceSquared (const gmPoint3& p1, const gmPoint3& p2)
{
  return gmSqr(p1[0] - p2[0]) + gmSqr(p1[1] - p2[1]) + gmSqr(p1[2] - p2[2]);
}


inline gmPoint3 gmLerp (double f, const gmPoint3& p1, const gmPoint3& p2)
{
  return p1 + ((p2 - p1) * f);
}


// output

inline ostream& operator<< (ostream& os, const gmPoint3& p)
{
  os << '(' << p[0] << ',' << p[1] << ',' << p[2] << ')';
  return os;
}


// complex utility functions

gmPoint3 gmSph2Cart (double rho, double theta, double phi);
void gmCart2Sph (gmPoint3 p, double *rho, double *theta, double *phi);

double gmSignedDistance (gmPoint3 p, gmVector3 n, double d);
gmBool gmIntersectPlane (gmPoint3 p, gmVector3 v, gmVector3 n, double d,
                         gmPoint3 *i);
void gmIntersectSphere (gmPoint3 p, gmVector3 v, gmPoint3 c, double r,
                        gmBool front, gmPoint3 *i);
gmBool gmCalcPlane (gmPoint3 p1, gmPoint3 p2, gmPoint3 p3,
                    gmVector3 *n, double *d);
gmBool gmNearestPoint (gmPoint3 p1, gmVector3 v1, gmPoint3 p2, gmVector3 v2,
                       gmPoint3 *i);


#endif // GM_POINT3_H
