// -*- mode: C++; -*-

/* LibGM: A Geometrical Library
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef GM_POINT4_H
#define GM_POINT4_H

#include <math.h>
#include <assert.h>
#include <iostream.h>

#include "gm_utils.h"
#include "gm_vector4.h"


class gmPoint4 {
public:
  gmPoint4 ();
  gmPoint4 (const gmPoint4&);
  gmPoint4 (double, double, double, double);

  // array access

  double& operator[] (int);
  const double& operator[] (int) const;

  // assignment

  gmPoint4& assign (double, double, double, double);
  gmPoint4& operator= (const gmPoint4&);

  // math operators

  gmPoint4& operator+= (const gmVector4&);

  gmPoint4 operator+ (const gmVector4&) const;
  gmVector4 operator- (const gmPoint4&) const;

  friend gmPoint4 operator+ (const gmVector4&, const gmPoint4&);

  gmBool operator== (const gmPoint4&) const;
  gmBool operator!= (const gmPoint4&) const;

  // operations

  gmPoint4& clamp (double, double);
  double originDistance () const;
  double originDistanceSquared () const;

  void copyTo (float [4]) const;
  void copyTo (double [4]) const;

  friend double gmDistance (const gmPoint4&, const gmPoint4&);
  friend double gmDistanceSquared (const gmPoint4&, const gmPoint4&);
  friend gmPoint4 gmLerp (double, const gmPoint4&, const gmPoint4&);

  // output

  friend ostream& operator<< (ostream&, const gmPoint4&);

protected:
  double p_[4];
};


// constructors

inline gmPoint4::gmPoint4 ()
{
  p_[0] = p_[1] = p_[2] = p_[3] = 0;
}


inline gmPoint4::gmPoint4 (const gmPoint4& p)
{
  p_[0] = p.p_[0];
  p_[1] = p.p_[1];
  p_[2] = p.p_[2];
  p_[3] = p.p_[3];
}


inline gmPoint4::gmPoint4 (double x, double y, double z, double a)
{
  p_[0] = x;
  p_[1] = y;
  p_[2] = z;
  p_[3] = a;
}


// array access

inline double& gmPoint4::operator[] (int i)
{
  assert(i == 0 || i == 1 || i == 2 || i == 3);
  return p_[i];
}


inline const double& gmPoint4::operator[] (int i) const
{
  assert(i == 0 || i == 1 || i == 2 || i == 3);
  return p_[i];
}


// assignment

inline gmPoint4& gmPoint4::assign (double x, double y, double z, double a)
{
  p_[0] = x;
  p_[1] = y;
  p_[2] = z;
  p_[3] = a;
  return *this;
}


inline gmPoint4& gmPoint4::operator= (const gmPoint4& p)
{
  p_[0] = p[0];
  p_[1] = p[1];
  p_[2] = p[2];
  p_[3] = p[3];
  return *this;
}


// math operators

inline gmPoint4& gmPoint4::operator+= (const gmVector4& v)
{
  p_[0] += v[0];
  p_[1] += v[1];
  p_[2] += v[2];
  p_[3] += v[3];
  return *this;
}


inline gmPoint4 gmPoint4::operator+ (const gmVector4& v) const
{
  return gmPoint4(p_[0] + v[0], p_[1] + v[1], p_[2] + v[2], p_[3] + v[3]);
}


inline gmVector4 gmPoint4::operator- (const gmPoint4& p) const
{
  return gmVector4(p_[0] - p[0], p_[1] - p[1], p_[2] - p[2], p_[3] - p[3]);
}


inline gmPoint4 operator+ (const gmVector4& v, const gmPoint4& p)
{
  return gmPoint4(p[0] + v[0], p[1] + v[1], p[2] + v[2], p[3] + v[3]);
}


inline gmBool gmPoint4::operator== (const gmPoint4& p) const
{
  return (gmFuzEQ(p_[0], p[0]) && gmFuzEQ(p_[1], p[1]) &&
          gmFuzEQ(p_[2], p[2]) && gmFuzEQ(p_[3], p[3]));
}


inline gmBool gmPoint4::operator!= (const gmPoint4& p) const
{
  return (!(*this == p));
}


// operations

inline gmPoint4& gmPoint4::clamp (double lo, double hi)
{
  gmClamp(p_[0], lo, hi);
  gmClamp(p_[1], lo, hi);
  gmClamp(p_[2], lo, hi);
  gmClamp(p_[3], lo, hi);
  return *this;
}


inline double gmPoint4::originDistance () const
{
  return sqrt(gmSqr(p_[0]) + gmSqr(p_[1]) + gmSqr(p_[2]) + gmSqr(p_[3]));
}


inline double gmPoint4::originDistanceSquared () const
{
  return gmSqr(p_[0]) + gmSqr(p_[1]) + gmSqr(p_[2]) + gmSqr(p_[3]);
}


inline void gmPoint4::copyTo (float f[4]) const
{
  f[0] = p_[0];
  f[1] = p_[1];
  f[2] = p_[2];
  f[3] = p_[3];
}


inline void gmPoint4::copyTo (double d[4]) const
{
  d[0] = p_[0];
  d[1] = p_[1];
  d[2] = p_[2];
  d[3] = p_[3];
}


inline double gmDistance (const gmPoint4& p1, const gmPoint4& p2)
{
  return sqrt(gmSqr(p1[0] - p2[0]) + gmSqr(p1[1] - p2[1]) +
              gmSqr(p1[2] - p2[2]) + gmSqr(p1[3] - p2[3]));
}


inline double gmDistanceSquared (const gmPoint4& p1, const gmPoint4& p2)
{
  return gmSqr(p1[0] - p2[0]) + gmSqr(p1[1] - p2[1]) +
         gmSqr(p1[2] - p2[2]) + gmSqr(p1[3] - p2[3]);
}


inline gmPoint4 gmLerp (double f, const gmPoint4& p1, const gmPoint4& p2)
{
  return p1 + ((p2 - p1) * f);
}


// output

inline ostream& operator<< (ostream& os, const gmPoint4& p)
{
  os << '(' << p[0] << ',' << p[1] << ',' << p[2] << ',' << p[3] << ')';
  return os;
}


#endif // GM_POINT4_H
