// -*- mode: C++; -*-

/* LibGM: A Geometrical Library
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef GM_VECTOR3_H
#define GM_VECTOR3_H

#include <math.h>
#include <assert.h>
#include <iostream.h>

#include "gm_utils.h"


class gmVector3 {
public:
  gmVector3 ();
  gmVector3 (const gmVector3&);
  gmVector3 (double, double, double);

  // array access

  double& operator[] (int);
  const double& operator[] (int) const;

  // assignment

  gmVector3& assign (double, double, double);
  gmVector3& operator= (const gmVector3&);

  // math operators

  gmVector3& operator+= (const gmVector3&);
  gmVector3& operator-= (const gmVector3&);
  gmVector3& operator*= (double);
  gmVector3& operator/= (double);

  gmVector3 operator+ (const gmVector3&) const;
  gmVector3 operator- (const gmVector3&) const;
  gmVector3 operator- () const;
  gmVector3 operator* (double) const;
  gmVector3 operator/ (double) const;

  friend gmVector3 operator* (double, const gmVector3&);

  gmBool operator== (const gmVector3&) const;
  gmBool operator!= (const gmVector3&) const;

  // operations

  gmVector3& clamp (double, double);
  double length () const;
  double lengthSquared () const;
  gmVector3& normalize ();

  void copyTo (float [3]) const;
  void copyTo (double [3]) const;

  friend gmVector3 gmCross (const gmVector3&, const gmVector3&);
  friend double gmDot (const gmVector3&, const gmVector3&);
  friend gmVector3 gmLerp (double, const gmVector3&, const gmVector3&);

  // output

  friend ostream& operator<< (ostream&, const gmVector3&);

protected:
  double v_[3];
};


// constructors

inline gmVector3::gmVector3 ()
{
  v_[0] = v_[1] = v_[2] = 0;
}


inline gmVector3::gmVector3 (const gmVector3& v)
{
  v_[0] = v.v_[0];
  v_[1] = v.v_[1];
  v_[2] = v.v_[2];
}


inline gmVector3::gmVector3 (double x, double y, double z)
{
  v_[0] = x;
  v_[1] = y;
  v_[2] = z;
}


// array access

inline double& gmVector3::operator[] (int i)
{
  assert(i == 0 || i == 1 || i == 2);
  return v_[i];
}


inline const double& gmVector3::operator[] (int i) const
{
  assert(i == 0 || i == 1 || i == 2);
  return v_[i];
}


// assignment

inline gmVector3& gmVector3::assign (double x, double y, double z)
{
  v_[0] = x;
  v_[1] = y;
  v_[2] = z;
  return *this;
}


inline gmVector3& gmVector3::operator= (const gmVector3& v)
{
  v_[0] = v[0];
  v_[1] = v[1];
  v_[2] = v[2];
  return *this;
}


// math operators

inline gmVector3& gmVector3::operator+= (const gmVector3& v)
{
  v_[0] += v[0];
  v_[1] += v[1];
  v_[2] += v[2];
  return *this;
}


inline gmVector3& gmVector3::operator-= (const gmVector3& v)
{
  v_[0] -= v[0];
  v_[1] -= v[1];
  v_[2] -= v[2];
  return *this;
}


inline gmVector3& gmVector3::operator*= (double c)
{
  v_[0] *= c;
  v_[1] *= c;
  v_[2] *= c;
  return *this;
}


inline gmVector3& gmVector3::operator/= (double c)
{
  assert(!gmIsZero(c));
  v_[0] /= c;
  v_[1] /= c;
  v_[2] /= c;
  return *this;
}


inline gmVector3 gmVector3::operator+ (const gmVector3& v) const
{
  return gmVector3(v_[0] + v[0], v_[1] + v[1], v_[2] + v[2]);
}


inline gmVector3 gmVector3::operator- (const gmVector3& v) const
{
  return gmVector3(v_[0] - v[0], v_[1] - v[1], v_[2] - v[2]);
}


inline gmVector3 gmVector3::operator- () const
{
  return gmVector3(-v_[0], -v_[1], -v_[2]);
}


inline gmVector3 gmVector3::operator* (double c) const
{
  return gmVector3(v_[0] * c, v_[1] * c, v_[2] * c);
}


inline gmVector3 gmVector3::operator/ (double c) const
{
  assert(!gmIsZero(c));
  return gmVector3(v_[0] / c, v_[1] / c, v_[2] / c);
}


inline gmVector3 operator* (double c, const gmVector3& v)
{
  return gmVector3(c * v[0], c * v[1], c * v[2]);
}


inline gmBool gmVector3::operator== (const gmVector3& v) const
{
  return (gmFuzEQ(v_[0], v[0]) && gmFuzEQ(v_[1], v[1]) &&
          gmFuzEQ(v_[2], v[2]));
}


inline gmBool gmVector3::operator!= (const gmVector3& v) const
{
  return (!(*this == v));
}


// operations

inline gmVector3& gmVector3::clamp (double lo, double hi)
{
  gmClamp(v_[0], lo, hi);
  gmClamp(v_[1], lo, hi);
  gmClamp(v_[2], lo, hi);
  return *this;
}


inline double gmVector3::length () const
{
  return sqrt(gmSqr(v_[0]) + gmSqr(v_[1]) + gmSqr(v_[2]));
}


inline double gmVector3::lengthSquared () const
{
  return gmSqr(v_[0]) + gmSqr(v_[1]) + gmSqr(v_[2]);
}


inline gmVector3& gmVector3::normalize ()
{
  double len = length();
  //assert(!gmIsZero(len));
  if (len > GM_EPSILON)
    *this /= len;
  return *this;
}


inline void gmVector3::copyTo (float f[3]) const
{
  f[0] = v_[0];
  f[1] = v_[1];
  f[2] = v_[2];
}


inline void gmVector3::copyTo (double d[3]) const
{
  d[0] = v_[0];
  d[1] = v_[1];
  d[2] = v_[2];
}


inline gmVector3 gmCross (const gmVector3& v1, const gmVector3& v2)
{
  return gmVector3(v1[1] * v2[2] - v1[2] * v2[1],
                   v1[2] * v2[0] - v1[0] * v2[2],
                   v1[0] * v2[1] - v1[1] * v2[0]);
}


inline double gmDot (const gmVector3& v1, const gmVector3& v2)
{
  return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2];
}


inline gmVector3 gmLerp (double f, const gmVector3& v1, const gmVector3& v2)
{
  return v1 + ((v2 - v1) * f);
}


// output

inline ostream& operator<< (ostream& os, const gmVector3& v)
{
  os << '<' << v[0] << ' ' << v[1] << ' ' << v[2] << '>';
  return os;
}


#endif // GM_VECTOR3_H
