
/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>

#include "mtk_constraint.h"
#include "mtk.h"


mtkCoreGL::mtkCoreGL ()
{
  shade_model_=MTK_SMOOTH;
}


//----------------------------------------------------------------//

mtkCameraId mtkCoreGL::createCamera ()
{
  mtkCameraGL c;
  cameras_.insert(mtkCameraPair(c.getId(), c));
  return c.getId();
}


void mtkCoreGL::cameraInitialize (mtkCameraId id)
{
  // unused parameter
  id=id;

  glClearColor(0.0, 0.0, 0.0, 0.0);
  glEnable(GL_DEPTH_TEST);
}

// Added by Leonardo Mesquita - 15/02/2001
void mtkCoreGL::changeCameras()
{
   mtkCameraMap::iterator ci;

  for (ci=cameras_.begin(); ci!=cameras_.end(); ci++)
    (*ci).second.setChanged(true);
}

void mtkCoreGL::cameraDraw (mtkCameraId id)
{
  mtkCameraMap::iterator ci=cameras_.find(id);
  if (ci == cameras_.end())
    return;

  // setup camera
  (*ci).second.setup();

  // set shade model
  if (shade_model_ == MTK_FLAT)
    glShadeModel(GL_FLAT);
  else
    glShadeModel(GL_SMOOTH);

  // clear frame and z buffers
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


  light_.draw();

  // draw display list, draggers, guides, and lights
  // display_list_.draw(); - Changed by Leonardo Mesquita - 15/02/2001
  display_list_.draw( (*ci).second.getChanged() );

  mtkDraggerMap::iterator i;
  for (i=draggers_.begin(); i!=draggers_.end(); i++)
    ((*i).second())->draw(false);
  mtkGuideMap::iterator j;
  for (j=guides_.begin(); j!=guides_.end(); j++)
    ((*j).second())->draw();

  // Included by Leonardo Mesquita - 01/02/2000
  if(cursor_.getState() == MTK_CURSOR_ACTIVE)
     cursor_.draw();


  // Included by Leonardo Mesquita - 15/02/2001
  (*ci).second.setChanged(false);
}


bool mtkCoreGL::cameraPointer (mtkCameraId id, int x, int y, mtkEvent type)
{
  mtkCameraMap::iterator ci=cameras_.find(id);
  if (ci == cameras_.end())
    return false;

  bool ret=false;
  static mtkDragger *current_dragger=0;
  static mtkDraggerPart current_part;
  list<mtkSelectionDataGL> changes;
  mtkElementCallback *callback;
  void *data;
  mtkDraggerMap::iterator i;
  mtkGuideMap::iterator j;
  list<mtkSelectionDataGL>::iterator k;

  // setup camera
  (*ci).second.setup();

  switch (type) {
    case MTK_PICK:
      // start picking
      selection_.startPicking(x, y);

      // draw display list, draggers and guides
      light_.draw();

      // display_list_.draw(); - Changed by Leonardo Mesquita - 15/02/2001
      display_list_.draw( (*ci).second.getChanged() );
      for (i=draggers_.begin(); i!=draggers_.end(); i++)
        ((*i).second())->draw(true);
      for (j=guides_.begin(); j!=guides_.end(); j++)
        ((*j).second())->draw();

      // end picking
      changes=selection_.endPicking();

      // no changes to the selection
      if (!changes.size())
        break;

      // test if some dragger was picked
      if (changes.front().type == MTK_DRAGGER) {
        mtkDraggerMap::iterator it=draggers_.find(changes.front().id);
        if (it == draggers_.end())
          break;
        current_dragger=(*it).second();
        current_part=changes.front().part;
        current_dragger->interaction(current_part, x, y, MTK_PICK);
        break;
      }

      // call the callbacks of the changed elements
      for (k=changes.begin(); k!=changes.end(); k++) {
        list<mtkId> on, off;
        callback=display_list_.getCallback((*k).id, &data);
        if ((*k).selected)
          on.push_back((*k).id);
        else
          off.push_back((*k).id);
        if (callback)
          ret=(*callback)(on, off, data) || ret;
      }
      break;

    case MTK_MOVE:
      if (current_dragger) {
        current_dragger->interaction(current_part, x, y, MTK_MOVE);
        ret=true;
      }
      break;

    case MTK_RELEASE:
      if (current_dragger) {
        current_dragger->interaction(current_part, x, y, MTK_RELEASE);
        current_dragger=0;
      }
      break;
  }
  return ret;
}

// Included by Leonardo Mesquita - 02/02/2000
bool mtkCoreGL::cameraPointer3D (mtkCameraId id, double x, double y, 
			         double z, mtkEvent type)
{
  mtkCameraMap::iterator ci=cameras_.find(id);
  if (ci == cameras_.end())
    return false;

  bool ret=false;

  // setup camera
  (*ci).second.setup();

  switch (type) {
    case MTK_PICK:
       break;
    
    case MTK_MOVE:
      cursor_.move(gmPoint3(x,y,z));
//      changeCameras();
      break;

    case MTK_RELEASE:
      break;
  }
  return ret;
}


void mtkCoreGL::cameraResize (mtkCameraId id, int x, int y,
                             int width, int height)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.resize(x, y, width, height);
}


void mtkCoreGL::cameraCopy (mtkCameraId from, mtkCameraId to)
{
  mtkCameraMap::iterator fi=cameras_.find(from);
  if (fi == cameras_.end())
    return;
  mtkCameraMap::iterator ti=cameras_.find(to);
  if (ti == cameras_.end())
    return;
  (*ti).second.copy((*fi).second);
}


void mtkCoreGL::setCameraState (mtkCameraId id, mtkState state)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.setState(state);
}


bool mtkCoreGL::getCameraState (mtkCameraId id, mtkState *state)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return false;
  *state=(*i).second.getState();
  return true;
}


void mtkCoreGL::setCameraAim (mtkCameraId id, double a, double b, double c,
                             mtkCameraMode mode)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.setAim(a, b, c, mode);
}


bool mtkCoreGL::getCameraAim (mtkCameraId id, double *a, double *b, double *c,
                             mtkCameraMode mode)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return false;
  (*i).second.getAim(a, b, c, mode);
  return true;
}


void mtkCoreGL::setCameraPosition (mtkCameraId id, double a, double b, double c,
                                  mtkCameraMode mode)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.setPosition(a, b, c, mode);
}


bool mtkCoreGL::getCameraPosition (mtkCameraId id,
                                  double *a, double *b, double *c,
                                  mtkCameraMode mode)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return false;
  (*i).second.getPosition(a, b, c, mode);
  return true;
}


void mtkCoreGL::setCameraNear (mtkCameraId id, double n)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.setNear(n);
}


bool mtkCoreGL::getCameraNear (mtkCameraId id, double *n)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return false;
  *n=(*i).second.getNear();
  return true;
}


void mtkCoreGL::setCameraFar (mtkCameraId id, double f)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.setFar(f);
}


bool mtkCoreGL::getCameraFar (mtkCameraId id, double *f)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return false;
  *f=(*i).second.getFar();
  return true;
}


void mtkCoreGL::setCameraWindow (mtkCameraId id, double w)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.setWindow(w);
}


bool mtkCoreGL::getCameraWindow (mtkCameraId id, double *w)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return false;
  *w=(*i).second.getWindow();
  return true;
}


void mtkCoreGL::setCameraFOV (mtkCameraId id, double a)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.setFOV(a);
}


bool mtkCoreGL::getCameraFOV (mtkCameraId id, double *a)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return false;
  *a=(*i).second.getFOV();
  return true;
}


void mtkCoreGL::setCameraRoll (mtkCameraId id, double a)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.setRoll(a);
}


bool mtkCoreGL::getCameraRoll (mtkCameraId id, double *a)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return false;
  *a=(*i).second.getRoll();
  return true;
}


void mtkCoreGL::setCameraProjection (mtkCameraId id, mtkCameraProjType type)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.setProjection(type);
}


bool mtkCoreGL::getCameraProjection (mtkCameraId id, mtkCameraProjType *type)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return false;
  *type=(*i).second.getProjection();
  return true;
}


void mtkCoreGL::cameraDolly (mtkCameraId id, double forward)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.dolly(forward);
}


void mtkCoreGL::cameraTracking (mtkCameraId id, double up, double right)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.tracking(up, right);
}


void mtkCoreGL::cameraOrbit (mtkCameraId id, double up, double right)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.orbit(up, right);
}


void mtkCoreGL::cameraPan (mtkCameraId id, double right)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.pan(right);
}


void mtkCoreGL::cameraTilt (mtkCameraId id, double up)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.tilt(up);
}


void mtkCoreGL::cameraRoll (mtkCameraId id, double ccw)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.roll(ccw);
}


void mtkCoreGL::cameraZoom (mtkCameraId id, double out)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.zoom(out);
}


void mtkCoreGL::cameraLook (mtkCameraId id, list<mtkId> id_list)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  gmPoint3 c;
  double r;
  display_list_.calcBoundingSphere(id_list, &c, &r);
  (*i).second.look(c, r);
}


void mtkCoreGL::cameraLevel (mtkCameraId id)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.level();
}


void mtkCoreGL::cameraReset (mtkCameraId id)
{
  mtkCameraMap::iterator i=cameras_.find(id);
  if (i == cameras_.end())
    return;
  (*i).second.reset();
}


//----------------------------------------------------------------//

void mtkCoreGL::setLightType (mtkLightId id, mtkLightType type)
{
  light_.setType(id, type);
}


void mtkCoreGL::setLightSize (float size)
{
  light_.setSize(size);
}


float mtkCoreGL::getLightSize ()
{
  return light_.getSize();
}

void mtkCoreGL::setLightPosition (mtkLightId id, gmPoint3 position)
{
  light_.setPosition(id, position);
}


void mtkCoreGL::setLightComponents (mtkLightId id,
                                    mtkMaterialComponent component,
                                    gmPoint4 value)
{
  light_.setComponents(id, component, value);
}


bool mtkCoreGL::setSpotDir (mtkLightId id, mtkSpotLight spot, gmPoint3 value)
{
  return light_.setSpotDir(id, spot, value);
}


bool mtkCoreGL::setSpot (mtkLightId id, mtkSpotLight spot, float value)
{
  return light_.setSpot(id, spot, value);
}


bool mtkCoreGL::setLightAttenuation (mtkLightId id, mtkAttenuation att,
                                     float value)
{
  return light_.setAttenuation(id, att, value);
}


void mtkCoreGL::setLight (mtkLightId id, mtkState state)
{
  light_.setLight(id, state);
}


mtkState mtkCoreGL::getLight (mtkLightId id)
{
  return light_.getLight(id);
}


void mtkCoreGL::setLighting (mtkState state)
{
  light_.setLighting(state);
}


mtkState mtkCoreGL::getLighting ()
{
  return light_.getLighting();
}


void mtkCoreGL::setLightGlobal (gmPoint4 value)
{
  light_.setGlobalLight(value);
}


void mtkCoreGL::setLightTwoSide (float value)
{
  light_.setTwoSide(value);
}


void mtkCoreGL::setLightLocalViewer (float value)
{
  light_.setLocalViewer(value);
}


void mtkCoreGL::setLightDraw (bool draw)
{
  light_.setLightDraw(draw);
}


//----------------------------------------------------------------//

void mtkCoreGL::setRenderType (mtkRenderType r)
{
  display_list_.setRenderType(r);

  // Included by Leonardo Mesquita - 15/02/2001  
  changeCameras();
}


mtkRenderType mtkCoreGL::getRenderType()
{
  return display_list_.getRenderType();
}


void mtkCoreGL::setShadeModel (mtkShadeModel m)
{
  shade_model_=m;
}


mtkShadeModel mtkCoreGL::getShadeModel ()
{
  return shade_model_;
}


void mtkCoreGL::setHighlightColor (float r, float g, float b)
{
  display_list_.setHighlightColor(r, g, b);

  // Included by Leonardo Mesquita - 15/02/2001  
  changeCameras();
}


bool mtkCoreGL::createElement (mtkId id)
{
  return display_list_.createElement(id);
}


bool mtkCoreGL::appendElement (mtkId id)
{
  return display_list_.appendElement(id);
}


void mtkCoreGL::finishElement ()
{
  display_list_.finishElement();

  // Included by Leonardo Mesquita - 15/02/2001  
  changeCameras();
}


bool mtkCoreGL::checkElement (mtkId id)
{
  return display_list_.checkElement(id);
}


bool mtkCoreGL::deleteElement (mtkId id)
{
  return display_list_.deleteElement(id);

  // Included by Leonardo Mesquita - 15/02/2001  
  changeCameras();
}


bool mtkCoreGL::clearElement (mtkId id)
{
  return display_list_.clearElement(id);

  // Included by Leonardo Mesquita - 15/02/2001  
  changeCameras();
}


list<mtkId> mtkCoreGL::getAllElements ()
{
  return display_list_.getAllElements();
}


void mtkCoreGL::clearDisplayList ()
{
  display_list_.clearDisplayList();

  // Included by Leonardo Mesquita - 15/02/2001  
  changeCameras();
}


void mtkCoreGL::begin (mtkGeometricType g)
{
  display_list_.begin(g);
}


void mtkCoreGL::setMaterialv (mtkFaceMode fmode, mtkMaterialComponent component,
                             gmPoint4 value)
{
  display_list_.setMaterialv(fmode, component, value);
}


void mtkCoreGL::setMaterialf (mtkFaceMode fmode, mtkMaterialComponent component,
                             float value)
{
  display_list_.setMaterialf(fmode, component, value);
}


void mtkCoreGL::setColor (float r, float g, float b)
{
  display_list_.setColor(r, g, b);
}


void mtkCoreGL::setVertex (double x, double y, double z)
{
  display_list_.setVertex(gmPoint3(x, y, z));
}


void mtkCoreGL::setVertex (gmPoint3 v)
{
  display_list_.setVertex(v);
}


void mtkCoreGL::setNormal (double i, double j, double k)
{
  display_list_.setNormal(gmVector3(i, j, k));
}


void mtkCoreGL::setNormal (gmVector3 n)
{
  display_list_.setNormal(n);
}


void mtkCoreGL::bezierSurface (int uorder, int vorder, int nu, int vu, gmPoint3 *ctrlpoints)
{
  display_list_.bezier(uorder, vorder, nu, vu, ctrlpoints);
}


void mtkCoreGL::end ()
{
  display_list_.end();
}


void mtkCoreGL::applyElementMatrix (list<mtkId> id_list, gmMatrix4 m)
{
  display_list_.applyMatrix(id_list, m);

  // Included by Leonardo Mesquita - 15/02/2001  
  changeCameras();
}


bool mtkCoreGL::calcBoundingBox (list<mtkId> id_list, gmPoint3 *min,
                                gmPoint3 *max)
{
  return display_list_.calcBoundingBox(id_list, min, max);
}


bool mtkCoreGL::calcBoundingSphere (list<mtkId> id_list, gmPoint3 *center,
                                   double *radius)
{
  return display_list_.calcBoundingSphere(id_list, center, radius);
}


void mtkCoreGL::setElementState (mtkId id, mtkState state)
{
  display_list_.setState(id, state);

  // Included by Leonardo Mesquita - 15/02/2001  
  changeCameras();
}


mtkState mtkCoreGL::getElementState (mtkId id)
{
  return display_list_.getState(id);
}


void mtkCoreGL::setHighlight (mtkId id, bool s)
{
  display_list_.setHighlight(id, s);

  // Included by Leonardo Mesquita - 15/02/2001  
  changeCameras();
}


bool mtkCoreGL::isHighlighted (mtkId id)
{
  return display_list_.isHighlighted(id);
}


bool mtkCoreGL::setElementCallback (mtkId id, mtkElementCallback *callback,
                                   void *data)
{
  return display_list_.setCallback(id, callback, data);
}


mtkElementCallback *mtkCoreGL::getElementCallback (mtkId id, void **datap)
{
  return display_list_.getCallback(id, datap);
}


//----------------------------------------------------------------//

void mtkCoreGL::setSelectionMode (mtkSelectionMode mode)
{
  selection_.setSelectionMode(mode);
}


mtkSelectionMode mtkCoreGL::getSelectionMode ()
{
  return selection_.getSelectionMode();
}


bool mtkCoreGL::setSelection (mtkId id, bool s)
{
  list<mtkId> on, off;
  if (s) {
    selection_.select(id);
    on.push_back(id);
  }
  else {
    selection_.unselect(id);
    off.push_back(id);
  }
  void *data;
  mtkElementCallback *callback;
  if ((callback=display_list_.getCallback(id, &data)))
    return (*callback)(on, off, data);
  return false;
}


bool mtkCoreGL::isSelected (mtkId id)
{
  return selection_.isSelected(id);
}


list<mtkId> mtkCoreGL::getAllSelected ()
{
  return selection_.getAll();
}

//---------------------------------------------------//

void mtkCoreGL::setGuideState (mtkGuideId id, mtkState state)
{
  mtkGuideMap::iterator i=guides_.find(id);
  if (i == guides_.end())
    return;
  ((*i).second())->setState(state);
}


bool mtkCoreGL::getGuideState (mtkGuideId id, mtkState *state)
{
  mtkGuideMap::iterator i=guides_.find(id);
  if (i == guides_.end())
    return false;
  *state=((*i).second())->getState();
  return true;
}


bool mtkCoreGL::getGuideType (mtkGuideId id, mtkGuideType *type)
{
  mtkGuideMap::iterator i=guides_.find(id);
  if (i == guides_.end())
    return false;
  *type=((*i).second())->getType();
  return true;
}


mtkAxisGL *mtkCoreGL::findAxis_ (mtkGuideId id)
{
  mtkGuideMap::iterator i=guides_.find(id);
  if (i == guides_.end())
    return 0;
  if (((*i).second())->getType() != MTK_AXIS)
    return 0;
  return (mtkAxisGL *)(*i).second();
}


mtkGridGL *mtkCoreGL::findGrid_ (mtkGuideId id)
{
  mtkGuideMap::iterator i=guides_.find(id);
  if (i == guides_.end())
    return 0;
  if (((*i).second())->getType() != MTK_GRID)
    return 0;
  return (mtkGridGL *)(*i).second();
}


mtkGuideId mtkCoreGL::createAxis()
{
  mtkAxisGL *d=new mtkAxisGL();
  mtkGuideWrapper dw(d);
  guides_.insert(mtkGuidePair(d->getId(), dw));
  return d->getId();
}


void mtkCoreGL::setAxisOrigin (mtkGuideId id, gmPoint3 origin)
{
  mtkAxisGL *d=findAxis_(id);
  if (!d)
    return;
  d->setOrigin(origin);
}


bool mtkCoreGL::getAxisOrigin(mtkGuideId id, gmPoint3 *origin)
{
  mtkAxisGL *d=findAxis_(id);
  if (!d)
    return false;
  *origin=d->getOrigin();
  return true;
}


void mtkCoreGL::setAxis (mtkGuideId id, gmVector3 axis[3])
{
  mtkAxisGL *d=findAxis_(id);
  if (!d)
    return;
  d->setAxis(axis);
}


void mtkCoreGL::getAxis (mtkGuideId id, gmVector3 axis[3])
{
  mtkAxisGL *d=findAxis_(id);
  if (!d)
    return;
  d->getAxis(axis);
}


void mtkCoreGL::setAxisStyle (mtkGuideId id, mtkAxisStyle style)
{
  mtkAxisGL *d=findAxis_(id);
  if (!d)
    return;
  d->setAxisStyle(style);
}


bool mtkCoreGL::getAxisStyle (mtkGuideId id, mtkAxisStyle *style)
{
  mtkAxisGL *d=findAxis_(id);
  if (!d)
    return false;
  *style=d->getAxisStyle();
  return true;
}


void mtkCoreGL::setAxisThickness (mtkGuideId id, unsigned int thickness)
{
  mtkAxisGL *d=findAxis_(id);
  if (!d)
    return;
  d->setThickness(thickness);
}


void mtkCoreGL::setAxisColor (mtkGuideId id, float r, float g, float b)
{
  mtkAxisGL *d=findAxis_(id);
  if (!d)
    return;
  d->setColor(r, g, b);
}


mtkGuideId mtkCoreGL::createGrid()
{
  mtkGridGL *d=new mtkGridGL();
  mtkGuideWrapper dw(d);
  guides_.insert(mtkGuidePair(d->getId(), dw));
  return d->getId();
}


void mtkCoreGL::setGridOrigin (mtkGuideId id, gmPoint3 origin)
{
  mtkGridGL *d=findGrid_(id);
  if (!d)
    return;
  d->setOrigin(origin);
}


bool mtkCoreGL::getGridOrigin (mtkGuideId id, gmPoint3 *origin)
{
  mtkGridGL *d=findGrid_(id);
  if (!d)
    return false;
  *origin=d->getOrigin();
  return true;
}


void mtkCoreGL::setGridDir (mtkGuideId id, gmVector3 dir[2])
{
  mtkGridGL *d=findGrid_(id);
  if (!d)
    return;
  d->setDir(dir);
}


void mtkCoreGL::getGridDir (mtkGuideId id, gmVector3 dir[2])
{
  mtkGridGL *d=findGrid_(id);
  if (!d)
    return;
  d->getDir(dir);
}


void mtkCoreGL::setGridNumLines (mtkGuideId id, int nLines1, int nLines2)
{
  mtkGridGL *d=findGrid_(id);
  if (!d)
    return;
  d->setNumLines(nLines1, nLines2);
}


void mtkCoreGL::getGridNumLines (mtkGuideId id, int *nLines1, int *nLines2)
{
  mtkGridGL *d=findGrid_(id);
  if (!d)
    return;
  d->getNumLines(nLines1, nLines2);
}


void mtkCoreGL::setGridEdgeThickness (mtkGuideId id, unsigned int thickness)
{
  mtkGridGL *d=findGrid_(id);
  if (!d)
    return;
  d->setEdgeThickness(thickness);
}


void mtkCoreGL::setGridEdgeColor (mtkGuideId id, float r, float g, float b)
{
  mtkGridGL *d=findGrid_(id);
  if (!d)
    return;
  d->setEdgeColor(r, g, b);
}


void mtkCoreGL::setGridFaceColor (mtkGuideId id, float r, float g, float b, float a)
{
  mtkGridGL *d=findGrid_(id);
  if (!d)
    return;
  d->setFaceColor(r, g, b, a);
}


void mtkCoreGL::setGridFillFace (mtkGuideId id, bool fill_face)
{
  mtkGridGL *d=findGrid_(id);
  if (!d)
    return;
  d->setFillFace(fill_face);
}


//----------------------------------------------------------------//

void mtkCoreGL::setDraggerState (mtkDraggerId id, mtkState state)
{
  mtkDraggerMap::iterator i=draggers_.find(id);
  if (i == draggers_.end())
    return;
  ((*i).second())->setState(state);
}

bool mtkCoreGL::getDraggerState (mtkDraggerId id, mtkState *state)
{
  mtkDraggerMap::iterator i=draggers_.find(id);
  if (i == draggers_.end())
    return false;
  *state=((*i).second())->getState();
  return true;
}


bool mtkCoreGL::getDraggerType (mtkDraggerId id, mtkDraggerType *type)
{
  mtkDraggerMap::iterator i=draggers_.find(id);
  if (i == draggers_.end())
    return false;
  *type=((*i).second())->getType();
  return true;
}


void mtkCoreGL::setDraggerSize (mtkDraggerId id, double size)
{
  mtkDraggerMap::iterator i=draggers_.find(id);
  if (i == draggers_.end())
    return;
  ((*i).second())->setSize(size);
}


bool mtkCoreGL::getDraggerSize (mtkDraggerId id, double *size)
{
  mtkDraggerMap::iterator i=draggers_.find(id);
  if (i == draggers_.end())
    return false;
  *size=((*i).second())->getSize();
  return true;
}


bool mtkCoreGL::setConstraint_ (mtkDraggerId id, mtkDraggerPart part,
                                mtkConstraintGL c)
{
  mtkDraggerMap::iterator i=draggers_.find(id);
  if (i == draggers_.end())
    return false;
  bool ret=false;
  if (((*i).second())->getType() == MTK_BOUNDING_BOX) {
    mtkBoundingBoxGL *d=(mtkBoundingBoxGL *)(*i).second();
    ret=d->setConstraint(part, c);
  }
  else if (((*i).second())->getType() == MTK_TRACKBALL) {
    mtkTrackballGL *d=(mtkTrackballGL *)(*i).second();
    ret=d->setConstraint(part, c);
  }
  else if (((*i).second())->getType() == MTK_JACK) {
    mtkJackGL *d=(mtkJackGL *)(*i).second();
    ret=d->setConstraint(part, c);
  }
  else if (((*i).second())->getType() == MTK_CURSOR3D) {
    mtkCursor3DGL *d=(mtkCursor3DGL *)(*i).second();
    ret=d->setConstraint(part, c);
  }
  return ret;
}


bool mtkCoreGL::setSameDepth (mtkDraggerId id, mtkDraggerPart part)
{
  mtkConstraintGL c;
  c.setSameDepth();
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSameDepthHorizontal (mtkDraggerId id, mtkDraggerPart part)
{
  mtkConstraintGL c;
  c.setSameDepthHorizontal();
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSameDepthVertical (mtkDraggerId id, mtkDraggerPart part)
{
  mtkConstraintGL c;
  c.setSameDepthVertical();
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapAxis (mtkDraggerId id, mtkDraggerPart part,
                             gmPoint3 p, gmVector3 v)
{
  mtkConstraintGL c;
  c.setSnapAxis(p, v);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapAxisD (mtkDraggerId id, mtkDraggerPart part,
                              gmPoint3 p, gmVector3 v)
{
  mtkConstraintGL c;
  c.setSnapAxisD(p, v);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapParallelAxis (mtkDraggerId id, mtkDraggerPart part,
                                     gmVector3 v)
{
  mtkConstraintGL c;
  c.setSnapParallelAxis(v);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapParallelAxisD (mtkDraggerId id, mtkDraggerPart part,
                                      gmVector3 v)
{
  mtkConstraintGL c;
  c.setSnapParallelAxisD(v);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapNearestAxis (mtkDraggerId id, mtkDraggerPart part,
                                    gmPoint3 p, gmVector3 v1, gmVector3 v2,
                                    gmVector3 v3)
{
  mtkConstraintGL c;
  c.setSnapNearestAxis(p, v1, v2, v3);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapNearestAxisD (mtkDraggerId id, mtkDraggerPart part,
                                     gmPoint3 p, gmVector3 v1, gmVector3 v2,
                                     gmVector3 v3)
{
  mtkConstraintGL c;
  c.setSnapNearestAxisD(p, v1, v2, v3);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapNearestParallelAxis (mtkDraggerId id,
                                            mtkDraggerPart part,
                                            gmVector3 v1, gmVector3 v2,
                                            gmVector3 v3)
{
  mtkConstraintGL c;
  c.setSnapNearestParallelAxis(v1, v2, v3);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapNearestParallelAxisD (mtkDraggerId id,
                                             mtkDraggerPart part,
                                             gmVector3 v1, gmVector3 v2,
                                             gmVector3 v3)
{
  mtkConstraintGL c;
  c.setSnapNearestParallelAxisD(v1, v2, v3);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapPlane (mtkDraggerId id, mtkDraggerPart part,
                              gmVector3 n, double d)
{
  mtkConstraintGL c;
  c.setSnapPlane(n, d);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapPlane (mtkDraggerId id, mtkDraggerPart part,
                              gmPoint3 p, gmVector3 v1, gmVector3 v2)
{
  mtkConstraintGL c;
  c.setSnapPlane(p, v1, v2);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapPlaneD (mtkDraggerId id, mtkDraggerPart part,
                               gmPoint3 p, gmVector3 v1, gmVector3 v2)
{
  mtkConstraintGL c;
  c.setSnapPlaneD(p, v1, v2);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapParallelPlane (mtkDraggerId id, mtkDraggerPart part,
                                      gmVector3 n)
{
  mtkConstraintGL c;
  c.setSnapParallelPlane(n);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapParallelPlane (mtkDraggerId id, mtkDraggerPart part,
                                      gmVector3 v1, gmVector3 v2)
{
  mtkConstraintGL c;
  c.setSnapParallelPlane(v1, v2);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapParallelPlaneD (mtkDraggerId id, mtkDraggerPart part,
                                       gmVector3 v1, gmVector3 v2)
{
  mtkConstraintGL c;
  c.setSnapParallelPlaneD(v1, v2);
  return setConstraint_(id, part, c);
}


bool mtkCoreGL::setSnapSphere (mtkDraggerId id, mtkDraggerPart part,
                               gmPoint3 c, double r)
{
  mtkConstraintGL ct;
  ct.setSnapSphere(c, r);
  return setConstraint_(id, part, ct);
}


//----------------------------------------------------------------//

mtkDraggerId mtkCoreGL::createBoxDragger ()
{
  mtkBoundingBoxGL *d=new mtkBoundingBoxGL();
  mtkDraggerWrapper dw(d);
  draggers_.insert(mtkDraggerPair(d->getId(), dw));
  return d->getId();
}


mtkBoundingBoxGL *mtkCoreGL::findBox_ (mtkDraggerId id)
{
  mtkDraggerMap::iterator i=draggers_.find(id);
  if (i == draggers_.end())
    return 0;
  if (((*i).second())->getType() != MTK_BOUNDING_BOX)
    return 0;
  return (mtkBoundingBoxGL *)(*i).second();
}


void  mtkCoreGL::setBoxGeometry (mtkDraggerId id, gmPoint3 v0,
                                 gmVector3 e01, gmVector3 e02, gmVector3 e04)
{
  mtkBoundingBoxGL *d=findBox_(id);
  if (!d)
    return;
  d->setGeometry(v0, e01, e02, e04);
}


bool mtkCoreGL::getBoxVertex (mtkDraggerId id, mtkDraggerPart part,
                              gmPoint3 *p)
{
  mtkBoundingBoxGL *d=findBox_(id);
  if (!d)
    return false;
  *p=d->getVertex(part);
  return true;
}


bool mtkCoreGL::getBoxCenter (mtkDraggerId id, gmPoint3 *p)
{
  mtkBoundingBoxGL *d=findBox_(id);
  if (!d)
    return false;
  *p=d->getCenter();
  return true;
}


void mtkCoreGL::setBoxVisibility (mtkDraggerId id, mtkDraggerPart part,
                                  bool b)
{
  mtkBoundingBoxGL *d=findBox_(id);
  if (!d)
    return;
  d->setVisibility(part, b);
}


bool mtkCoreGL::getBoxVisibility (mtkDraggerId id, mtkDraggerPart part)
{
  mtkBoundingBoxGL *d=findBox_(id);
  if (!d)
    return false;
  return d->getVisibility(part);
}


void mtkCoreGL::setBoxVertexSize (mtkDraggerId id, unsigned int size)
{
  mtkBoundingBoxGL *d=findBox_(id);
  if (!d)
    return;
  d->setVertexSize(size);
}


void mtkCoreGL::setBoxVertexColor (mtkDraggerId id, float r, float g, float b)
{
  mtkBoundingBoxGL *d=findBox_(id);
  if (!d)
    return;
  d->setVertexColor(r, g, b);
}


void mtkCoreGL::setBoxEdgeThickness (mtkDraggerId id, unsigned int thickness)
{
  mtkBoundingBoxGL *d=findBox_(id);
  if (!d)
    return;
  d->setEdgeThickness(thickness);
}


void mtkCoreGL::setBoxEdgeColor (mtkDraggerId id, float r, float g, float b)
{
  mtkBoundingBoxGL *d=findBox_(id);
  if (!d)
    return;
  d->setEdgeColor(r, g, b);
}


void mtkCoreGL::setBoxFaceColor (mtkDraggerId id,
                                float r, float g, float b, float a)
{
  mtkBoundingBoxGL *d=findBox_(id);
  if (!d)
    return;
  d->setFaceColor(r, g, b, a);
}


void mtkCoreGL::applyBoxMatrix (mtkDraggerId id, gmMatrix4 m)
{
  mtkBoundingBoxGL *d=findBox_(id);
  if (!d)
    return;
  d->applyMatrix(m);
}


void mtkCoreGL::setBoxCallback (mtkDraggerId id, mtkBoxCallback *callback,
                               void *data)
{
  mtkBoundingBoxGL *d=findBox_(id);
  if (!d)
    return;
  d->setCallback(callback, data);
}


//----------------------------------------------------------------//

mtkDraggerId mtkCoreGL::createTrackballDragger ()
{
  mtkTrackballGL *d=new mtkTrackballGL();
  mtkDraggerWrapper dw(d);
  draggers_.insert(mtkDraggerPair(d->getId(), dw));
  return d->getId();
}


mtkTrackballGL *mtkCoreGL::findBall_ (mtkDraggerId id)
{
  mtkDraggerMap::iterator i=draggers_.find(id);
  if (i == draggers_.end())
    return 0;
  if (((*i).second())->getType() != MTK_TRACKBALL)
    return 0;
  return (mtkTrackballGL *)(*i).second();
}


void mtkCoreGL::setBallGeometry (mtkDraggerId id, gmPoint3 center,
                                 double radius, double arrow)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return;
  d->setGeometry(center, radius, arrow);
}


void mtkCoreGL::setBallResolution (mtkDraggerId id, int slices, int stacks)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return;
  d->setResolution(slices, stacks);
}


bool mtkCoreGL::getBallSphere (mtkDraggerId id, gmPoint3 *center,
                               double *radius)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return false;
  d->getSphere(center, radius);
  return true;
}


bool mtkCoreGL::getBallArrow (mtkDraggerId id, gmPoint3 *p)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return false;
  *p=d->getArrow();
  return true;
}


bool mtkCoreGL::getBallRing (mtkDraggerId id, mtkDraggerPart part,
                             gmVector3 *v)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return false;
  *v=d->getRing(part);
  return true;
}


void mtkCoreGL::setBallVisibility (mtkDraggerId id, mtkDraggerPart part,
                                   bool b)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return;
  d->setVisibility(part, b);
}


bool mtkCoreGL::getBallVisibility (mtkDraggerId id, mtkDraggerPart part)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return false;
  return d->getVisibility(part);
}


void mtkCoreGL::setBallVertexSize (mtkDraggerId id, unsigned int size)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return;
  d->setVertexSize(size);
}


void mtkCoreGL::setBallVertexColor (mtkDraggerId id, float r, float g, float b)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return;
  d->setVertexColor(r, g, b);
}


void mtkCoreGL::setBallEdgeThickness (mtkDraggerId id, unsigned int thickness)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return;
  d->setEdgeThickness(thickness);
}


void mtkCoreGL::setBallEdgeColor (mtkDraggerId id, float r, float g, float b)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return;
  d->setEdgeColor(r, g, b);
}


void mtkCoreGL::setBallFaceColor (mtkDraggerId id,
                                float r, float g, float b, float a)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return;
  d->setFaceColor(r, g, b, a);
}


void mtkCoreGL::applyBallMatrix (mtkDraggerId id, gmMatrix4 m)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return;
  d->applyMatrix(m);
}


void mtkCoreGL::setBallCallback (mtkDraggerId id,
                                 mtkTrackballCallback *callback,
                                 void *data)
{
  mtkTrackballGL *d=findBall_(id);
  if (!d)
    return;
  d->setCallback(callback, data);
}


//----------------------------------------------------------------//

mtkDraggerId mtkCoreGL::createJackDragger ()
{
  mtkJackGL *d=new mtkJackGL();
  mtkDraggerWrapper dw(d);
  draggers_.insert(mtkDraggerPair(d->getId(), dw));
  return d->getId();
}


mtkJackGL *mtkCoreGL::findJack_ (mtkDraggerId id)
{
  mtkDraggerMap::iterator i=draggers_.find(id);
  if (i == draggers_.end())
    return 0;
  if (((*i).second())->getType() != MTK_JACK)
    return 0;
  return (mtkJackGL *)(*i).second();
}


void mtkCoreGL::setJackGeometry (mtkDraggerId id, gmPoint3 v0,
                                 gmVector3 e01, gmVector3 e03, gmVector3 e05)
{
  mtkJackGL *d=findJack_(id);
  if (!d)
    return;
  d->setGeometry(v0, e01, e03, e05);
}


void mtkCoreGL::setJackVisibility (mtkDraggerId id, mtkDraggerPart part,
                                   bool b)
{
  mtkJackGL *d=findJack_(id);
  if (!d)
    return;
  d->setVisibility(part, b);
}


bool mtkCoreGL::getJackVisibility (mtkDraggerId id, mtkDraggerPart part)
{
  mtkJackGL *d=findJack_(id);
  if (!d)
    return false;
  return d->getVisibility(part);
}


bool mtkCoreGL::getJackVertex (mtkDraggerId id, mtkDraggerPart part,
                               gmPoint3 *p)
{
  mtkJackGL *d=findJack_(id);
  if (!d)
    return false;
  *p=d->getVertex(part);
  return true;
}


void mtkCoreGL::setJackVertexSize (mtkDraggerId id, unsigned int size)
{
  mtkJackGL *d=findJack_(id);
  if (!d)
    return;
  d->setVertexSize(size);
}


void mtkCoreGL::setJackVertexColor (mtkDraggerId id, float r, float g, float b)
{
  mtkJackGL *d=findJack_(id);
  if (!d)
    return;
  d->setVertexColor(r, g, b);
}


void mtkCoreGL::setJackEdgeThickness (mtkDraggerId id, unsigned int thickness)
{
  mtkJackGL *d=findJack_(id);
  if (!d)
    return;
  d->setEdgeThickness(thickness);
}


void mtkCoreGL::setJackEdgeColor (mtkDraggerId id, float r, float g, float b)
{
  mtkJackGL *d=findJack_(id);
  if (!d)
    return;
  d->setEdgeColor(r, g, b);
}


void mtkCoreGL::applyJackMatrix (mtkDraggerId id, gmMatrix4 m)
{
  mtkJackGL *d=findJack_(id);
  if (!d)
    return;
  d->applyMatrix(m);
}


void mtkCoreGL::setJackCallback (mtkDraggerId id, mtkJackCallback *callback,
                                void *data)
{
  mtkJackGL *d=findJack_(id);
  if (!d)
    return;
  d->setCallback(callback, data);
}


//----------------------------------------------------------------//

mtkDraggerId mtkCoreGL::createCursor3DDragger ()
{
  mtkCursor3DGL *d=new mtkCursor3DGL();
  mtkDraggerWrapper dw(d);
  draggers_.insert(mtkDraggerPair(d->getId(), dw));
  return d->getId();
}


mtkCursor3DGL *mtkCoreGL::findCursor3D_ (mtkDraggerId id)
{
  mtkDraggerMap::iterator i=draggers_.find(id);
  if (i == draggers_.end())
    return 0;
  if (((*i).second())->getType() != MTK_CURSOR3D)
    return 0;
  return (mtkCursor3DGL *)(*i).second();
}


void mtkCoreGL::setCursor3DGeometry (mtkDraggerId id, gmPoint3 v0,
                                 gmVector3 e01, gmVector3 e03, gmVector3 e05)
{
  mtkCursor3DGL *d=findCursor3D_(id);
  if (!d)
    return;
  d->setGeometry(v0, e01, e03, e05);
}


void mtkCoreGL::setCursor3DVisibility (mtkDraggerId id, mtkDraggerPart part,
                                   bool b)
{
  mtkCursor3DGL *d=findCursor3D_(id);
  if (!d)
    return;
  d->setVisibility(part, b);
}


bool mtkCoreGL::getCursor3DVisibility (mtkDraggerId id, mtkDraggerPart part)
{
  mtkCursor3DGL *d=findCursor3D_(id);
  if (!d)
    return false;
  return d->getVisibility(part);
}


bool mtkCoreGL::getCursor3DVertex (mtkDraggerId id, mtkDraggerPart part,
                               gmPoint3 *p)
{
  mtkCursor3DGL *d=findCursor3D_(id);
  if (!d)
    return false;
  *p=d->getVertex(part);
  return true;
}


void mtkCoreGL::setCursor3DVertexSize (mtkDraggerId id, unsigned int size)
{
  mtkCursor3DGL *d=findCursor3D_(id);
  if (!d)
    return;
  d->setVertexSize(size);
}


void mtkCoreGL::setCursor3DVertexColor (mtkDraggerId id, float r, float g, float b)
{
  mtkCursor3DGL *d=findCursor3D_(id);
  if (!d)
    return;
  d->setVertexColor(r, g, b);
}


void mtkCoreGL::setCursor3DEdgeThickness (mtkDraggerId id, unsigned int thickness)
{
  mtkCursor3DGL *d=findCursor3D_(id);
  if (!d)
    return;
  d->setEdgeThickness(thickness);
}


void mtkCoreGL::setCursor3DEdgeColor (mtkDraggerId id, float r, float g, float b)
{
  mtkCursor3DGL *d=findCursor3D_(id);
  if (!d)
    return;
  d->setEdgeColor(r, g, b);
}


void mtkCoreGL::applyCursor3DMatrix (mtkDraggerId id, gmMatrix4 m)
{
  mtkCursor3DGL *d=findCursor3D_(id);
  if (!d)
    return;
  d->applyMatrix(m);
}


void mtkCoreGL::setCursor3DCallback (mtkDraggerId id, mtkCursor3DCallback *callback,
                                void *data)
{
  mtkCursor3DGL *d=findCursor3D_(id);
  if (!d)
    return;
  d->setCallback(callback, data);
}

//Included by Leonardo Mesquita - 01/02/2000
void mtkCoreGL::activateCursor()
{
   cursor_.setState(MTK_CURSOR_ACTIVE);
}

void mtkCoreGL::deactivateCursor()
{
   cursor_.setState(MTK_CURSOR_INACTIVE);
}

gmPoint3 mtkCoreGL::getCursorPosition()
{
   return cursor_.getPosition();
}

