// -*- mode: C++; -*-

/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef MTK_H
#define MTK_H

#include <map.h>
#include <list.h>

#include "mtk_camera.h"
#include "mtk_displaylist.h"
#include "mtk_dragger.h"
#include "mtk_guide.h"
#include "mtk_light.h"
#include "mtk_selection.h"
#include "mtk_wrapper.h"

// Included by Leonardo Mesquita - 01/02/2000
#include "mtk_cursor.h"

class mtkCoreGL {
public:
  mtkCoreGL ();

  //---------------- interface to camera module ----------------//

  mtkCameraId createCamera ();

  void cameraInitialize (mtkCameraId id);
  void cameraDraw (mtkCameraId id);
  bool cameraPointer (mtkCameraId id, int x, int y, mtkEvent type);

  // Included by Leonardo Mesquita - 02/02/2000
  bool cameraPointer3D (mtkCameraId id, double x, double y, double z,
			mtkEvent type);

  void cameraResize (mtkCameraId id, int x, int y, int width, int height);
  void cameraCopy (mtkCameraId from, mtkCameraId to);

  void setCameraState (mtkCameraId id, mtkState state);
  bool getCameraState (mtkCameraId id, mtkState *state);

  void setCameraAim (mtkCameraId id, double a, double b, double c,
                     mtkCameraMode mode);
  bool getCameraAim (mtkCameraId id, double *a, double *b, double *c,
                     mtkCameraMode mode);
  void setCameraPosition (mtkCameraId id, double a, double b, double c,
                          mtkCameraMode mode);
  bool getCameraPosition (mtkCameraId id, double *a, double *b, double *c,
                          mtkCameraMode mode);

  void setCameraNear (mtkCameraId id, double n);
  bool getCameraNear (mtkCameraId id, double *n);
  void setCameraFar (mtkCameraId id, double f);
  bool getCameraFar (mtkCameraId id, double *f);
  void setCameraWindow (mtkCameraId id, double w);
  bool getCameraWindow (mtkCameraId id, double *w);
  void setCameraFOV (mtkCameraId id, double a);
  bool getCameraFOV (mtkCameraId id, double *a);
  void setCameraRoll (mtkCameraId id, double a);
  bool getCameraRoll (mtkCameraId id, double *a);

  void setCameraProjection (mtkCameraId id, mtkCameraProjType type);
  bool getCameraProjection (mtkCameraId id, mtkCameraProjType *type);

  void cameraDolly (mtkCameraId id, double forward);
  void cameraTracking (mtkCameraId id, double up, double right);
  void cameraOrbit (mtkCameraId id, double up, double right);
  void cameraPan (mtkCameraId id, double right);
  void cameraTilt (mtkCameraId id, double up);
  void cameraRoll (mtkCameraId id, double ccw);
  void cameraZoom (mtkCameraId id, double out);
  void cameraLook (mtkCameraId id, list<mtkId> id_list);
  void cameraLevel (mtkCameraId id);
  void cameraReset (mtkCameraId id);

  //---------------- interface to light module ----------------//

  void setLightType (mtkLightId id, mtkLightType type);
  void setLightSize (float size);
  float getLightSize ();

  void setLight (mtkLightId id, mtkState state);
  mtkState getLight (mtkLightId id);
  void setLightPosition (mtkLightId id, gmPoint3 position);
  void setLightComponents (mtkLightId id, mtkMaterialComponent component,
                           gmPoint4 value);
  bool setSpotDir (mtkLightId id, mtkSpotLight spot, gmPoint3 value);
  bool setSpot (mtkLightId id, mtkSpotLight spot, float value);
  bool setLightAttenuation (mtkLightId id, mtkAttenuation att, float value);

  void setLighting (mtkState state);
  mtkState getLighting ();
  void setLightGlobal (gmPoint4 value);
  void setLightTwoSide (float value);
  void setLightLocalViewer (float value);
  void setLightDraw (bool draw);

  //---------------- interface to display list module ----------------//

  /// Sets how objects are rendered in the display list: wire frame or shade
  void setRenderType (mtkRenderType r);
  /// Get rendering mode
  mtkRenderType getRenderType ();
  /// Sets how objects are shaded in the display list: flat or smooth
  void setShadeModel (mtkShadeModel m);
  /// Get shade model
  mtkShadeModel getShadeModel ();
  void setHighlightColor (float r, float g, float b);

  /// Create a display list element
  bool createElement (mtkId id);
  /// Appends geometric data to an existing element
  bool appendElement (mtkId id);
  /// Finish a display list element
  void finishElement ();
  /// Test if there is an element in the display list with the given id
  bool checkElement (mtkId id);
  /// Erase an element in the display list, given its identifier
  bool deleteElement (mtkId id);
  bool clearElement (mtkId id);
  list<mtkId> getAllElements ();

  /// Erase all elements currently in the display list
  void clearDisplayList ();

  /// Starts the definition of a geometric type (POINT, LINE, POLYGON)
  void begin (mtkGeometricType g);
  /// Sets the current material for elements sent to display list
  void setMaterialv (mtkFaceMode fmode, mtkMaterialComponent component,
                             gmPoint4 value);
  void setMaterialf (mtkFaceMode fmode, mtkMaterialComponent component,
                             float value);
  /// Sets the current color for elements sent to display list
  void setColor (float r, float g, float b);
  /// Declare a display list coordinate
  void setVertex (double x, double y, double z);
  void setVertex (gmPoint3 v);
  void setNormal (double i, double j, double k);
  void setNormal (gmVector3 n);
  /// Creates a bezier surface
  void bezierSurface (int uorder, int vorder, int nu, int vu, gmPoint3 *ctrlpoints);
  /// Finishes a polygon or a group of points or lines sent to display list
  void end ();

  void applyElementMatrix (list<mtkId> id_list, gmMatrix4 m);
  bool calcBoundingBox (list<mtkId> id_list, gmPoint3 *min, gmPoint3 *max);
  bool calcBoundingSphere (list<mtkId> id_list, gmPoint3 *center,
                           double *radius);

  void setElementState (mtkId id, mtkState state);
  mtkState getElementState (mtkId id);
  void setHighlight (mtkId id, bool s);
  bool isHighlighted (mtkId id);

  bool setElementCallback (mtkId id, mtkElementCallback *callback, void *data);
  mtkElementCallback *getElementCallback (mtkId id, void **datap);

  //---------------- interface to selection module ----------------//

  void setSelectionMode (mtkSelectionMode mode);
  mtkSelectionMode getSelectionMode ();
  bool setSelection (mtkId id, bool s);
  bool isSelected (mtkId id);
  list<mtkId> getAllSelected ();

  //---------------- interface to guide module ----------------//

  void setGuideState (mtkGuideId id, mtkState state);
  bool getGuideState (mtkGuideId id, mtkState *state);
  bool getGuideType (mtkGuideId id, mtkGuideType *type);

  mtkGuideId createAxis();
  void setAxisOrigin (mtkGuideId id, gmPoint3 origin);
  bool getAxisOrigin(mtkGuideId id, gmPoint3 *origin);
  void setAxis (mtkGuideId id, gmVector3 axis[3]);
  void getAxis (mtkGuideId id, gmVector3 axis[3]);
  void setAxisStyle (mtkGuideId id, mtkAxisStyle style);
  bool getAxisStyle (mtkGuideId id, mtkAxisStyle *style);
  void setAxisThickness (mtkGuideId id, unsigned int thickness);
  void setAxisColor (mtkGuideId id, float r, float g, float b);

  mtkGuideId createGrid();
  void setGridOrigin (mtkGuideId id, gmPoint3 origin);
  bool getGridOrigin (mtkGuideId id, gmPoint3 *origin);
  void setGridDir (mtkGuideId id, gmVector3 dir[2]);
  void getGridDir (mtkGuideId id, gmVector3 dir[2]);
  void setGridNumLines (mtkGuideId id, int nLines1, int nLines2);
  void getGridNumLines (mtkGuideId id, int *nLines1, int *nLines2);
  void setGridEdgeThickness (mtkGuideId id, unsigned int thickness);
  void setGridEdgeColor (mtkGuideId id, float r, float g, float b);
  void setGridFaceColor (mtkGuideId id, float r, float g, float b, float a);
  void setGridFillFace (mtkGuideId id, bool fill_face);

  //---------------- interface to dragger module ----------------//

  void setDraggerState (mtkDraggerId id, mtkState state);
  bool getDraggerState (mtkDraggerId id, mtkState *state);
  bool getDraggerType (mtkDraggerId id, mtkDraggerType *type);
  void setDraggerSize (mtkDraggerId id, double size);
  bool getDraggerSize (mtkDraggerId id, double *size);

  bool setSameDepth (mtkDraggerId id, mtkDraggerPart part);
  bool setSameDepthHorizontal (mtkDraggerId id, mtkDraggerPart part);
  bool setSameDepthVertical (mtkDraggerId id, mtkDraggerPart part);
  bool setSnapAxis (mtkDraggerId id, mtkDraggerPart part,
                    gmPoint3 p, gmVector3 v);
  bool setSnapAxisD (mtkDraggerId id, mtkDraggerPart part,
                     gmPoint3 p, gmVector3 v);
  bool setSnapParallelAxis (mtkDraggerId id, mtkDraggerPart part, gmVector3 v);
  bool setSnapParallelAxisD (mtkDraggerId id, mtkDraggerPart part, gmVector3 v);
  bool setSnapNearestAxis (mtkDraggerId id, mtkDraggerPart part,
                           gmPoint3 p, gmVector3 v1, gmVector3 v2,
                           gmVector3 v3);
  bool setSnapNearestAxisD (mtkDraggerId id, mtkDraggerPart part,
                            gmPoint3 p, gmVector3 v1, gmVector3 v2,
                            gmVector3 v3);
  bool setSnapNearestParallelAxis (mtkDraggerId id, mtkDraggerPart part,
                                   gmVector3 v1, gmVector3 v2, gmVector3 v3);
  bool setSnapNearestParallelAxisD (mtkDraggerId id, mtkDraggerPart part,
                                    gmVector3 v1, gmVector3 v2, gmVector3 v3);
  bool setSnapPlane (mtkDraggerId id, mtkDraggerPart part,
                     gmVector3 n, double d);
  bool setSnapPlane (mtkDraggerId id, mtkDraggerPart part,
                     gmPoint3 p, gmVector3 v1, gmVector3 v2);
  bool setSnapPlaneD (mtkDraggerId id, mtkDraggerPart part,
                      gmPoint3 p, gmVector3 v1, gmVector3 v2);
  bool setSnapParallelPlane (mtkDraggerId id, mtkDraggerPart part,
                             gmVector3 n);
  bool setSnapParallelPlane (mtkDraggerId id, mtkDraggerPart part,
                             gmVector3 v1, gmVector3 v2);
  bool setSnapParallelPlaneD (mtkDraggerId id, mtkDraggerPart part,
                              gmVector3 v1, gmVector3 v2);
  bool setSnapSphere (mtkDraggerId id, mtkDraggerPart part,
                      gmPoint3 c, double r);

  mtkDraggerId createBoxDragger ();
  void setBoxGeometry (mtkDraggerId id, gmPoint3 v0,
                       gmVector3 e01, gmVector3 e02, gmVector3 e04);
  bool getBoxVertex (mtkDraggerId id, mtkDraggerPart part, gmPoint3 *p);
  bool getBoxCenter (mtkDraggerId id, gmPoint3 *p);
  void setBoxVisibility (mtkDraggerId id, mtkDraggerPart part, bool b);
  bool getBoxVisibility (mtkDraggerId id, mtkDraggerPart part);
  void setBoxVertexSize (mtkDraggerId id, unsigned int size );
  void setBoxVertexColor (mtkDraggerId id, float r, float g, float b);
  void setBoxEdgeThickness (mtkDraggerId id, unsigned int thickness);
  void setBoxEdgeColor (mtkDraggerId id, float r, float g, float b);
  void setBoxFaceColor (mtkDraggerId id, float r, float g, float b, float a);
  void applyBoxMatrix (mtkDraggerId id, gmMatrix4 m);
  void setBoxCallback (mtkDraggerId id, mtkBoxCallback *callback,
                       void *data);

  mtkDraggerId createTrackballDragger ();
  void setBallGeometry (mtkDraggerId id, gmPoint3 center, double radius,
                        double arrow);
  void setBallResolution (mtkDraggerId id, int slices, int stacks);
  bool getBallSphere (mtkDraggerId id, gmPoint3 *center, double *radius);
  bool getBallArrow (mtkDraggerId id, gmPoint3 *p);
  bool getBallRing (mtkDraggerId id, mtkDraggerPart part, gmVector3 *v);
  void setBallVisibility (mtkDraggerId id, mtkDraggerPart part, bool b);
  bool getBallVisibility (mtkDraggerId id, mtkDraggerPart part);
  void setBallVertexSize (mtkDraggerId id, unsigned int size );
  void setBallVertexColor (mtkDraggerId id, float r, float g, float b);
  void setBallEdgeThickness (mtkDraggerId id, unsigned int thickness);
  void setBallEdgeColor (mtkDraggerId id, float r, float g, float b);
  void setBallFaceColor (mtkDraggerId id, float r, float g, float b, float a);
  void applyBallMatrix (mtkDraggerId id, gmMatrix4 m);
  void setBallCallback (mtkDraggerId id, mtkTrackballCallback *callback,
                        void *data);

  mtkDraggerId createJackDragger ();
  void setJackGeometry (mtkDraggerId id, gmPoint3 v0,
                        gmVector3 e01, gmVector3 e03, gmVector3 e05);
  bool getJackVertex (mtkDraggerId id, mtkDraggerPart part, gmPoint3 *p);
  void setJackVisibility (mtkDraggerId id, mtkDraggerPart part, bool b);
  bool getJackVisibility (mtkDraggerId id, mtkDraggerPart part);
  void setJackVertexSize (mtkDraggerId id, unsigned int size);
  void setJackVertexColor (mtkDraggerId id, float r, float g, float b);
  void setJackEdgeThickness (mtkDraggerId id, unsigned int thickness);
  void setJackEdgeColor (mtkDraggerId id, float r, float g, float b);
  void applyJackMatrix (mtkDraggerId id, gmMatrix4 m);
  void setJackCallback (mtkDraggerId id, mtkJackCallback *callback,
                        void *data);

  // Cursor3D - variation of Jack - by Gonzaga

  mtkDraggerId createCursor3DDragger ();
  void setCursor3DGeometry (mtkDraggerId id, gmPoint3 v0,
                        gmVector3 e01, gmVector3 e03, gmVector3 e05);
  bool getCursor3DVertex (mtkDraggerId id, mtkDraggerPart part, gmPoint3 *p);
  void setCursor3DVisibility (mtkDraggerId id, mtkDraggerPart part, bool b);
  bool getCursor3DVisibility (mtkDraggerId id, mtkDraggerPart part);
  void setCursor3DVertexSize (mtkDraggerId id, unsigned int size);
  void setCursor3DVertexColor (mtkDraggerId id, float r, float g, float b);
  void setCursor3DEdgeThickness (mtkDraggerId id, unsigned int thickness);
  void setCursor3DEdgeColor (mtkDraggerId id, float r, float g, float b);
  void applyCursor3DMatrix (mtkDraggerId id, gmMatrix4 m);
  void setCursor3DCallback (mtkDraggerId id, mtkCursor3DCallback *callback,
                        void *data);

  //---------------- interface to cursor module ----------------//


  // Included by Leonardo Mesquita - 01/02/2000
  void activateCursor();
  void deactivateCursor();
  gmPoint3 getCursorPosition();

private:
  typedef map<mtkCameraId, mtkCameraGL, less<mtkCameraId> > mtkCameraMap;
  typedef pair<const mtkCameraId, mtkCameraGL> mtkCameraPair;

  typedef mtkWrapper<mtkDragger> mtkDraggerWrapper;
  typedef map<mtkDraggerId, mtkDraggerWrapper, less<mtkDraggerId> > mtkDraggerMap;
  typedef pair<const mtkDraggerId, mtkDraggerWrapper> mtkDraggerPair;

  typedef mtkWrapper<mtkGuide> mtkGuideWrapper;
  typedef map<mtkGuideId, mtkGuideWrapper, less<mtkGuideId> > mtkGuideMap;
  typedef pair<const mtkGuideId, mtkGuideWrapper> mtkGuidePair;

  bool setConstraint_ (mtkDraggerId id, mtkDraggerPart part, mtkConstraintGL c);
  mtkBoundingBoxGL *findBox_ (mtkDraggerId id);
  mtkTrackballGL *findBall_ (mtkDraggerId id);
  mtkJackGL *findJack_ (mtkDraggerId id);
  mtkCursor3DGL *findCursor3D_ (mtkDraggerId id);
  mtkAxisGL *findAxis_ (mtkGuideId id);
  mtkGridGL *findGrid_ (mtkGuideId id);

  mtkCameraMap cameras_;
  mtkLightGL light_;
  mtkDraggerMap draggers_;
  mtkGuideMap guides_;

  mtkDisplayListGL display_list_;
  mtkSelectionGL selection_;

  mtkShadeModel shade_model_;

  // Included by Leonardo Mesquita - 01/02/2000
  mtkCursorGL cursor_;
  void changeCameras();
};


#endif // MTK_H
