/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <math.h>
#include <GL/glu.h>

#include "mtk_camera.h"


mtkCameraId mtkCamera::counter_=1;


mtkCamera::mtkCamera ()
{
  id_=counter_++;
  state_=MTK_ACTIVE;

  // Added by Leonardo Mesquita - 15/02/2001
  changed_ = true;

  reset();
}


mtkCameraId mtkCamera::getId ()
{
  return id_;
}


// Added by Leonardo Mesquite - 15/02/2001
void mtkCamera::setChanged (bool c)
{
  changed_=c;
}

// Added by Leonardo Mesquite - 15/02/2001
bool mtkCamera::getChanged ()
{
  return changed_;
}


void mtkCamera::setState (mtkState state)
{
  state_=state;
}


mtkState mtkCamera::getState ()
{
  return state_;
}


void mtkCamera::setAim (double a, double b, double c, mtkCameraMode mode)
{
  gmPoint3 new_aim;

  switch (mode) {
    case MTK_ABSOLUTE:
      new_aim.assign(a, b, c);
      break;
    case MTK_RELATIVE:
      new_aim.assign(pos_[0]+a, pos_[1]+b, pos_[2]+c);
      break;
    case MTK_SPHERICAL:
      new_aim=gmSph2Cart(a, b, c);
      break;
    default:
      return; /*!*/
  }

  // test if the new mtkewing direction is parallel to up vector
  if (gmCross(new_aim-pos_, up_).lengthSquared() < GM_EPSILON2) {
    // yes, so compute a new up vector
    up_=gmCross(new_aim-pos_, gmCross(up_, aim_-pos_));
    up_.normalize();
  }
  aim_=new_aim;
  need_recalc_=true;
}


void mtkCamera::getAim (double *a, double *b, double *c, mtkCameraMode mode)
{
  switch (mode) {
    case MTK_ABSOLUTE:
      *a=aim_[0];
      *b=aim_[1];
      *c=aim_[2];
      break;
    case MTK_RELATIVE:
      *a=aim_[0]-pos_[0];
      *b=aim_[1]-pos_[1];
      *c=aim_[2]-pos_[2];
      break;
    case MTK_SPHERICAL:
      gmCart2Sph(aim_, a, b, c);
      break;
    default:
      break; /*!*/
  }
}


void mtkCamera::setPosition (double a, double b, double c, mtkCameraMode mode)
{
  gmPoint3 new_pos;

  switch (mode) {
    case MTK_ABSOLUTE:
      new_pos.assign(a, b, c);
      break;
    case MTK_RELATIVE:
      new_pos.assign(aim_[0]+a, aim_[1]+b, aim_[2]+c);
      break;
    case MTK_SPHERICAL:
      new_pos=gmSph2Cart(a, b, c);
      break;
    default:
      return; /*!*/
  }

  // test if the new mtkewing direction is parallel to up vector
  if (gmCross(aim_-new_pos, up_).lengthSquared() < GM_EPSILON2) {
    // yes, so compute a new up vector
    up_=gmCross(aim_-new_pos, gmCross(up_, aim_-pos_));
    up_.normalize();
  }
  pos_=new_pos;
  need_recalc_=true;
}


void mtkCamera::getPosition (double *a, double *b, double *c, mtkCameraMode mode)
{
  switch (mode) {
    case MTK_ABSOLUTE:
      *a=pos_[0];
      *b=pos_[1];
      *c=pos_[2];
      break;
    case MTK_RELATIVE:
      *a=pos_[0]-aim_[0];
      *b=pos_[1]-aim_[1];
      *c=pos_[2]-aim_[2];
      break;
    case MTK_SPHERICAL:
      gmCart2Sph(pos_, a, b, c);
      break;
    default:
      break; /*!*/
  }
}


void mtkCamera::setNear (double n)
{
  near_=n;
  need_recalc_=true;
}


double mtkCamera::getNear ()
{
  return near_;
}


void mtkCamera::setFar (double f)
{
  far_=f;
  need_recalc_=true;
}


double mtkCamera::getFar ()
{
  return fov_;
}


void mtkCamera::setWindow (double w)
{
  window_=w;
  need_recalc_=true;
}


double mtkCamera::getWindow ()
{
  return window_;
}


void mtkCamera::setFOV (double a)
{
  fov_=a;
  need_recalc_=true;
}


double mtkCamera::getFOV ()
{
  return fov_;
}


void mtkCamera::setRoll (double a)
{
  roll_=a;
  need_recalc_=true;
}


double mtkCamera::getRoll ()
{
  return roll_;
}


void mtkCamera::setProjection (mtkCameraProjType type)
{
  type_=type;
  need_recalc_=true;
}


mtkCameraProjType mtkCamera::getProjection ()
{
  return type_;
}


void mtkCamera::dolly (double forward)
{
  gmVector3 v=forward*((aim_-pos_).normalize());
  aim_+=v;
  pos_+=v;
  need_recalc_=true;
}


void mtkCamera::tracking (double up, double right)
{
  gmVector3 vright=gmCross(aim_-pos_, up_).normalize();
  gmVector3 vup=gmCross(vright, aim_-pos_).normalize();
  aim_+=up*vup+right*vright;
  pos_+=up*vup+right*vright;
  need_recalc_=true;
}


void mtkCamera::orbit (double up, double right)
{
  gmMatrix4 m;
  gmVector3 vright=gmCross(aim_-pos_, up_).normalize();
  gmVector3 vup=gmCross(vright, aim_-pos_).normalize();
  m=m.translation(aim_[0], aim_[1], aim_[2]);
  m*=m.rotation(-right, vup);
  m*=m.rotation(up, vright);
  m*=m.translation(-aim_[0], -aim_[1], -aim_[2]);
  pos_=m.transform(pos_);
  up_=m.transform(up_);
  need_recalc_=true;
}


void mtkCamera::pan (double right)
{
  gmMatrix4 m;
  gmVector3 vright=gmCross(aim_-pos_, up_).normalize();
  gmVector3 vup=gmCross(vright, aim_-pos_).normalize();
  m=m.translation(pos_[0], pos_[1], pos_[2]);
  m*=m.rotation(-right, vup);
  m*=m.translation(-pos_[0], -pos_[1], -pos_[2]);
  aim_=m.transform(aim_);
  up_=m.transform(up_);
  need_recalc_=true;
}


void mtkCamera::tilt (double up)
{
  gmMatrix4 m;
  gmVector3 vright=gmCross(aim_-pos_, up_).normalize();
  m=m.translation(pos_[0], pos_[1], pos_[2]);
  m*=m.rotation(up, vright);
  m*=m.translation(-pos_[0], -pos_[1], -pos_[2]);
  aim_=m.transform(aim_);
  up_=m.transform(up_);
  need_recalc_=true;
}


void mtkCamera::roll (double ccw)
{
  roll_+=ccw;
  need_recalc_=true;
}


void mtkCamera::zoom (double out)
{
  if (type_ == MTK_PERSPECTIVE) {
    fov_*=(1.0+out/100.0);
    gmClamp(fov_, 1.0, 179.0);
  }
  else
    window_*=(1.0+out/100.0);
  need_recalc_=true;
}


void mtkCamera::look (gmPoint3 center, double radius)
{
  // make sphere fit in screen
  double d;
  if (type_ == MTK_PERSPECTIVE) {
    setAim(center[0], center[1], center[2], MTK_ABSOLUTE);
    d=radius/sin(gmRadians(0.5*fov_));
    pos_=aim_+d*((pos_-aim_).normalize()); // dolly
  }
  else {
    pos_+=center-aim_;
    aim_=center;
    d=gmDistance(pos_, aim_);
    window_=2.0*radius;
  }
  // update near and far to reasonable values
  near_=0.2*(d-radius);
  far_=5.0*(d+radius);
  need_recalc_=true;
}


void mtkCamera::level ()
{
  aim_[1]=pos_[1];
  up_.assign(0.0, 1.0, 0.0);
  need_recalc_=true;
}


void mtkCamera::copy (const mtkCamera& c)
{
  state_=c.state_;
  type_=c.type_;
  aim_=c.aim_;
  pos_=c.pos_;
  up_=c.up_;
  far_=c.far_;
  near_=c.near_;
  window_=c.window_;
  fov_=c.fov_;
  roll_=c.roll_;
  need_recalc_=true;
}


void mtkCamera::reset ()
{
  type_=MTK_PERSPECTIVE;
  aim_.assign(0.0, 0.0, 0.0);
  pos_.assign(0.0, 0.0, 5.0);
  up_.assign(0.0, 1.0, 0.0);
  near_=0.5;
  far_=100.0;
  window_=2.0;
  fov_=45.0;
  roll_=0.0;
  need_recalc_=true;
}


//----------------------------------------------------------------//

mtkCameraGL::mtkCameraGL () : mtkCamera()
{
  x_=y_=0;
  width_=height_=100;
}


void mtkCameraGL::resize (int x, int y, int width, int height)
{
  x_=x;
  y_=y;
  width_=width;
  height_=height;
  need_recalc_=true;
}


void mtkCameraGL::setup ()
{
  if (need_recalc_) {
    // define projection matrix
    glMatrixMode(GL_PROJECTION);
    double aspect=(double)width_/height_;
    glLoadIdentity();
    if (type_ == MTK_PERSPECTIVE)
      gluPerspective(fov_, aspect, near_, far_);
    else {
      double d=0.5*window_;
      glOrtho(-d*aspect, d*aspect, -d, d, near_, far_);
    }

    // define model mtkewing matrix
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glRotated(roll_, 0.0, 0.0, 1.0);
    gluLookAt(pos_[0], pos_[1], pos_[2], aim_[0], aim_[1], aim_[2],
              up_[0], up_[1], up_[2]);

    // set mtkewport
    glViewport(x_, y_, width_, height_);

    // clear flag
    need_recalc_=false;
  }
}
