/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <GL/glu.h>

#include "mtk_displaylist.h"


mtkMaterial::mtkMaterial ()
{
  // use OpenGL defaults
  fmode=MTK_FRONT_AND_BACK;
  ambient[0]=ambient[1]=ambient[2]=0.2;
  ambient[3]=1.0;
  diffuse[0]=diffuse[1]=diffuse[2]=0.8;
  diffuse[3]=1.0;
  specular[0]=specular[1]=specular[2]=0.0;
  specular[3]=1.0;
  emission[0]=emission[1]=emission[2]=0.0;
  emission[3]=1.0;
  shininess=0.0;
}


mtkMaterial& mtkMaterial::operator= (const mtkMaterial& m)
{
  int i;

  fmode=m.fmode;
  for (i=0; i<4; i++) {
    ambient[i]=m.ambient[i];
    diffuse[i]=m.diffuse[i];
    specular[i]=m.specular[i];
    emission[i]=m.emission[i];
  }
  shininess=m.shininess;

  return *this;
}


//----------------------------------------------------------------//

mtkGeometry::mtkGeometry ()
{
  type=MTK_POINTS;
}


mtkGeometry& mtkGeometry::operator= (const mtkGeometry& g)
{
  type=g.type;
  vlist=g.vlist;
  nlist=g.nlist;
  clist=g.clist;
  mlist=g.mlist;

  return *this;
}


void mtkGeometry::reset ()
{
  type=MTK_POINTS;
  vlist.erase(vlist.begin(), vlist.end());
  nlist.erase(nlist.begin(), nlist.end());
  clist.erase(clist.begin(), clist.end());
  mlist.erase(mlist.begin(), mlist.end());
}


//----------------------------------------------------------------//

mtkElement::mtkElement ()
{
  state=MTK_ACTIVE;
  highlight=false;
  callback=0;
  cb_data=0;
}


mtkElement& mtkElement::operator= (const mtkElement& e)
{
  state=e.state;
  highlight=e.highlight;
  geometry=e.geometry;
  callback=e.callback;
  cb_data=e.cb_data;

  return *this;
}


void mtkElement::clear ()
{
  geometry.erase(geometry.begin(), geometry.end());
}


void mtkElement::reset ()
{
  state=MTK_ACTIVE;
  highlight=false;
  geometry.erase(geometry.begin(), geometry.end());
  callback=0;
  cb_data=0;
}


//----------------------------------------------------------------//

/// Constructor
mtkDisplayList::mtkDisplayList ()
{
  changed_=true;

  render_type_=MTK_SHADED;
  highlight_color_.r=1.0;
  highlight_color_.g=1.0;
  highlight_color_.b=0.0;

  c_id_=0;
  elock_=glock_=false;

  c_color_.r=1.0;
  c_color_.g=1.0;
  c_color_.b=1.0;
}


/// Sets how objects are rendered in the display list: wireframe or shaded
void mtkDisplayList::setRenderType (mtkRenderType r)
{
  render_type_=r;
  changed_=true;
}


/// Get rendering mode
mtkRenderType mtkDisplayList::getRenderType ()
{
  return render_type_;
}


void mtkDisplayList::setHighlightColor(float r, float g, float b)
{
  highlight_color_.r=r;
  highlight_color_.g=g;
  highlight_color_.b=b;
  changed_=true;
}


/// Sets an identifier to elements sent to the display list
bool mtkDisplayList::createElement (mtkId id)
{
  if (elock_ || !id)
    return false;
  mtkDisplayMap::iterator i=elements_.find(id);
  if (i != elements_.end()) {
    // Identifier already exists: copy non-geometric data
    c_element_=(*i).second;
    c_element_.clear();
    elements_.erase(i);
  }
  c_id_=id;
  elock_=true;
  return true;
}


/// Appends geometric data to an existing identifier
bool mtkDisplayList::appendElement (mtkId id)
{
  if (elock_ || !id)
    return false;
  mtkDisplayMap::iterator i=elements_.find(id);
  if (i != elements_.end()) {
    // Identifier already exists
    c_element_=(*i).second;
    elements_.erase(i);
  }
  c_id_=id;
  elock_=true;
  return true;
}


/// Finish a display list element
void mtkDisplayList::finishElement ()
{
  if (!elock_ || glock_)
    return;
  // Put element in the display list (note: we do accept empty geometry)
  elements_.insert(mtkDisplayPair(c_id_, c_element_));
  changed_=true;
  c_element_.reset();
  c_id_=0;
  elock_=false;
}


/// Test if there is an element in the display list with the given id
bool mtkDisplayList::checkElement (mtkId id)
{
  if (elock_ || !id)
    return false;
  mtkDisplayMap::iterator i=elements_.find(id);
  return i != elements_.end();
}


/// Erase an element in the display list, given its identifier
bool mtkDisplayList::deleteElement (mtkId id)
{
  if (elock_ || !id)
    return false;
  elements_.erase(id);
  changed_=true;
  return true;
}


/// Empty an element, given its identifier
bool mtkDisplayList::clearElement (mtkId id)
{
  if (elock_ || !id)
    return false;
  mtkDisplayMap::iterator i=elements_.find(id);
  if (i == elements_.end())
    return false; // not found
  (*i).second.clear();
  changed_=true;
  return true;
}


list<mtkId> mtkDisplayList::getAllElements ()
{
  list<mtkId> ret;
  if (!elock_) {
    mtkDisplayMap::iterator i;
    for (i=elements_.begin(); i!=elements_.end(); i++)
      ret.push_back((*i).first);
  }
  return ret;
}


/// Erase all elements currently in the display list
void mtkDisplayList::clearDisplayList ()
{
  if (elock_)
    return;
  elements_.erase(elements_.begin(), elements_.end());
  changed_=true;
}


/// Sets the current material for elements sent to display list
void mtkDisplayList::setMaterialv (mtkFaceMode fmode,
                                   mtkMaterialComponent component,
                                   gmPoint4 value)
{
  c_material_.fmode=fmode;
  switch (component) {
  case MTK_AMBIENT:
    c_material_.ambient[0]=value[0];
    c_material_.ambient[1]=value[1];
    c_material_.ambient[2]=value[2];
    c_material_.ambient[3]=value[3];
    break;
  case MTK_DIFFUSE:
    c_material_.diffuse[0]=value[0];
    c_material_.diffuse[1]=value[1];
    c_material_.diffuse[2]=value[2];
    c_material_.diffuse[3]=value[3];
    break;
  case MTK_SPECULAR:
    c_material_.specular[0]=value[0];
    c_material_.specular[1]=value[1];
    c_material_.specular[2]=value[2];
    c_material_.specular[3]=value[3];
    break;
  case MTK_EMISSION:
    c_material_.emission[0]=value[0];
    c_material_.emission[1]=value[1];
    c_material_.emission[2]=value[2];
    c_material_.emission[3]=value[3];
    break;
  default:
    break;
  }
}


void mtkDisplayList::setMaterialf (mtkFaceMode fmode,
                                   mtkMaterialComponent component,
                                   float value)
{
  c_material_.fmode=fmode;
  if (component == MTK_SHININESS)
    c_material_.shininess=value;
}


/// Sets the current color for elements sent to display list
void mtkDisplayList::setColor (float r, float g, float b)
{
  c_color_.r=r;
  c_color_.g=g;
  c_color_.b=b;
}


/// Selects a geometric type
void mtkDisplayList::begin (mtkGeometricType g)
{
  if (!elock_ || glock_)
    return;
  c_geometry_.type=g;
  glock_=true;
}


/** Sets x, y, z coordinates when sending points, lines and faces
  to the display list */
void mtkDisplayList::setVertex (gmPoint3 v)
{
  if (!glock_)
    return;
  c_geometry_.vlist.push_back(v);
  c_geometry_.mlist.push_back(c_material_);
  c_geometry_.clist.push_back(c_color_);
}


void mtkDisplayList::setNormal (gmVector3 n)
{
  if (!glock_)
    return;
  c_geometry_.nlist.push_back(n);
}


/** Set parameters and control points of a Bezier surface */
void mtkDisplayList::bezier (int uorder, int vorder, int nu, int vu,
			     gmPoint3 *ctrlpoints)
{
  if (!glock_)
    return;

  nu=vu; /*FIXME: make full bezier support*/
  /*gmPoint3 order(uorder, vorder, 0.0);
  gmPoint3 u(nu, vu, 0.0);
  c_geometry_.vlist.push_back(order);
  c_geometry_.mlist.push_back(c_material_);
  c_geometry_.clist.push_back(c_color_);
  c_geometry_.vlist.push_back(u);
  c_geometry_.clist.push_back(c_color_);*/

  for(int i=0; i<uorder*vorder; i++) {
    c_geometry_.vlist.push_back(ctrlpoints[i]);
    c_geometry_.mlist.push_back(c_material_);
    c_geometry_.clist.push_back(c_color_);
  }
}


/** Finishes a polygon or a group of one or more
    points or lines sent to display list */
void mtkDisplayList::end ()
{
  if (!glock_)
    return;

  int n=c_geometry_.vlist.size();
  mtkGeometricType type=c_geometry_.type;
  if (n)
    switch (type) {
    case MTK_POINTS:
      c_element_.geometry.push_back(c_geometry_);
      break;

    case MTK_LINES:
      if (n >= 2)
        c_element_.geometry.push_back(c_geometry_);
      break;

    case MTK_LINE_STRIP:
      if (n >= 2)
        c_element_.geometry.push_back(c_geometry_);
      break;

    case MTK_POLYGON:
      if (n >= 3) {
        if (!c_geometry_.nlist.size())
          c_geometry_.nlist.push_back(calcNormal_(c_geometry_.vlist));
        c_element_.geometry.push_back(c_geometry_);
      }
      break;

    case MTK_BEZIER:
      c_element_.geometry.push_back(c_geometry_); /*FIXME: add checking*/
      break;

    default:
      break;
    }
  c_geometry_.reset();
  glock_=false;
}


void mtkDisplayList::applyMatrix (list<mtkId> id_list, gmMatrix4 m)
{
  if (elock_)
    return;

  mtkDisplayMap::iterator elem_it;
  list<mtkGeometry>::iterator geo_it;
  list<gmPoint3>::iterator vtx_it;
  list<gmVector3>::iterator nml_it;

  // pass over all points and update vertices
  for (list<mtkId>::iterator i=id_list.begin(); i!=id_list.end(); i++) {
    elem_it=elements_.find(*i);
    if (elem_it == elements_.end())
      continue;
    for (geo_it=(*elem_it).second.geometry.begin();
         geo_it!=(*elem_it).second.geometry.end(); geo_it++) {
      // transform vertices
      for (vtx_it=(*geo_it).vlist.begin();
           vtx_it!=(*geo_it).vlist.end(); vtx_it++)
        (*vtx_it)=m.transform(*vtx_it);

      // transform normals
      for (nml_it=(*geo_it).nlist.begin();
           nml_it!=(*geo_it).nlist.end(); nml_it++)
        (*nml_it)=m.transform(*nml_it);
    }
  }
  changed_=true;
}


bool mtkDisplayList::calcBoundingBox (list<mtkId> id_list, gmPoint3 *min,
                                     gmPoint3 *max)
{
  if (elock_)
    return false;

  bool first=true;
  gmPoint3 p;
  mtkDisplayMap::iterator elem_it;
  list<mtkGeometry>::iterator geo_it;
  list<gmPoint3>::iterator vtx_it;

  // pass over all points and calculate extreme coordinates
  for (list<mtkId>::iterator i=id_list.begin(); i!=id_list.end(); i++) {
    elem_it=elements_.find(*i);
    if (elem_it == elements_.end())
      continue;
    for (geo_it=(*elem_it).second.geometry.begin();
         geo_it!=(*elem_it).second.geometry.end(); geo_it++) {
      for (vtx_it=(*geo_it).vlist.begin();
           vtx_it!=(*geo_it).vlist.end(); vtx_it++) {
        p=*vtx_it;
        if (first) {
          first=false;
          (*min)=(*max)=p;
          continue;
        }
        (*min)[0]=gmMin((*min)[0], p[0]);
        (*min)[1]=gmMin((*min)[1], p[1]);
        (*min)[2]=gmMin((*min)[2], p[2]);
        (*max)[0]=gmMax((*max)[0], p[0]);
        (*max)[1]=gmMax((*max)[1], p[1]);
        (*max)[2]=gmMax((*max)[2], p[2]);
      }
    }
  }
  return !first;
}


/* Code based on
      An Efficient Bounding Sphere
      by Jack Ritter
      from "Graphics Gems", Academic Press, 1990
*/

bool mtkDisplayList::calcBoundingSphere (list<mtkId> id_list, gmPoint3 *center,
                         double *radius)
{
  if (elock_)
    return false;

  bool first=true;
  gmPoint3 p, xmin, xmax, ymin, ymax, zmin, zmax;
  mtkDisplayMap::iterator elem_it;
  list<mtkGeometry>::iterator geo_it;
  list<gmPoint3>::iterator vtx_it;
  list<mtkId>::iterator i;

  // pass over all points and calculate extreme points
  for (i=id_list.begin(); i!=id_list.end(); i++) {
    elem_it=elements_.find(*i);
    if (elem_it == elements_.end())
      continue;
    for (geo_it=(*elem_it).second.geometry.begin();
         geo_it!=(*elem_it).second.geometry.end(); geo_it++) {
      for (vtx_it=(*geo_it).vlist.begin();
           vtx_it!=(*geo_it).vlist.end(); vtx_it++) {
        p=*vtx_it;
        if (first) {
          first=false;
          xmin=xmax=ymin=ymax=zmin=zmax=p;
          continue;
        }
        if (p[0] < xmin[0])
          xmin=p;
        else if (p[0] > xmax[0])
          xmax=p;
        if (p[1] < ymin[1])
          ymin=p;
        else if (p[1] > ymax[1])
          ymax=p;
        if (p[2] < zmin[2])
          zmin=p;
        else if (p[2] > zmax[2])
          zmax=p;
      }
    }
  }
  if (first)
    return false; // no points found

  // calculate squared distance between extreme points
  double xspan, yspan, zspan;
  xspan=gmDistanceSquared(xmax, xmin);
  yspan=gmDistanceSquared(ymax, ymin);
  zspan=gmDistanceSquared(zmax, zmin);

  // find maximally separated pair
  gmPoint3 d1, d2;
  if (xspan > yspan) {
    if (xspan > zspan) {
      d1=xmin;
      d2=xmax;
    }
    else {
      d1=zmin;
      d2=zmax;
    }
  }
  else {
    if (yspan > zspan) {
      d1=ymin;
      d2=ymax;
    }
    else {
      d1=zmin;
      d2=zmax;
    }
  }

  // calculate initial center and radius of sphere
  gmPoint3 c;
  double r, r2;
  c=gmLerp(0.5, d1, d2); // get midpoint
  r=gmDistance(d2, c);
  r2=gmDistanceSquared(d2, c);

  // pass again over all points and update sphere if needed
  double pr;
  for (i=id_list.begin(); i!=id_list.end(); i++) {
    elem_it=elements_.find(*i);
    if (elem_it == elements_.end())
      continue;
    for (geo_it=(*elem_it).second.geometry.begin();
         geo_it!=(*elem_it).second.geometry.end(); geo_it++) {
      for (vtx_it=(*geo_it).vlist.begin();
           vtx_it!=(*geo_it).vlist.end(); vtx_it++) {
        p=*vtx_it;
        if (gmDistanceSquared(p, c) > r2) {
          // point is outside
          pr=gmDistance(p, c);
          r=(r+pr)/2.0; // new sphere radius
          r2=r*r;
          c=gmLerp(r/pr, p, c); // new center
        }
      }
    }
  }
  *center=c;
  *radius=r;
  return true;
}


void mtkDisplayList::setState (mtkId id, mtkState state)
{
  if (elock_ || !id)
    return;
  mtkDisplayMap::iterator i=elements_.find(id);
  if (i != elements_.end()) {
    (*i).second.state=state;
    changed_=true;
  }
}


mtkState mtkDisplayList::getState (mtkId id)
{
  if (elock_ || !id)
    return MTK_HIDDEN;
  mtkDisplayMap::iterator i=elements_.find(id);
  if (i == elements_.end())
    return MTK_HIDDEN;
  return (*i).second.state;
}


void mtkDisplayList::setHighlight (mtkId id, bool s)
{
  if (elock_ || !id)
    return;
  mtkDisplayMap::iterator i=elements_.find(id);
  if (i != elements_.end()) {
    (*i).second.highlight=s;
    changed_=true;
  }
}


bool mtkDisplayList::isHighlighted (mtkId id)
{
  if (elock_ || !id)
    return false;
  mtkDisplayMap::iterator i=elements_.find(id);
  if (i == elements_.end())
    return false;
  return (*i).second.highlight;
}


bool mtkDisplayList::setCallback (mtkId id, mtkElementCallback *callback,
                                 void *data)
{
  if (elock_ || !id)
    return false;
  mtkDisplayMap::iterator i=elements_.find(id);
  if (i == elements_.end())
    return false;
  (*i).second.callback=callback;
  (*i).second.cb_data=data;
  return true;
}


mtkElementCallback *mtkDisplayList::getCallback (mtkId id, void **datap)
{
  if (elock_ || !id)
    return 0;
  mtkDisplayMap::iterator i=elements_.find(id);
  if (i == elements_.end())
    return 0;
  if (datap)
    *datap=(*i).second.cb_data;
  return (*i).second.callback;
}


gmVector3 mtkDisplayList::calcNormal_ (list<gmPoint3> vlist)
{
  gmPoint3 p1, p2;
  gmVector3 n(0.0, 0.0, 0.0);
  list<gmPoint3>::iterator it;
  for (it=vlist.begin(); it!=vlist.end();) {
    p1=*it;
    it++;
    if (it != vlist.end())
      p2=*it;
    else
      p2=*(vlist.begin());
    n[0]+=(p1[1]-p2[1])*(p1[2]+p2[2]);
    n[1]+=(p1[2]-p2[2])*(p1[0]+p2[0]);
    n[2]+=(p1[0]-p2[0])*(p1[1]+p2[1]);
  }
  return n.normalize();
}


//----------------------------------------------------------------//

mtkDisplayListGL::mtkDisplayListGL () : mtkDisplayList()
{
  gl_list_=0;
}


void mtkDisplayListGL::setupMaterial_ (mtkMaterial mat)
{
  GLenum face_mode;
  switch (mat.fmode) {
  case MTK_FRONT:
    face_mode=GL_FRONT;
    break;
  case MTK_BACK:
    face_mode=GL_BACK;
    break;
  default:
    face_mode=GL_FRONT_AND_BACK;
    break;
  }
  glMaterialfv(face_mode, GL_AMBIENT, mat.ambient);
  glMaterialfv(face_mode, GL_DIFFUSE, mat.diffuse);
  glMaterialfv(face_mode, GL_SPECULAR, mat.specular);
  glMaterialfv(face_mode, GL_EMISSION, mat.emission);
  glMaterialfv(face_mode, GL_SHININESS, &mat.shininess);
}


void mtkDisplayListGL::draw ()
{
  int i, j;

  if (!changed_) {
    glCallList(gl_list_);
    return;
  }

  if (!gl_list_)
    gl_list_=glGenLists(1);
  glNewList(gl_list_, GL_COMPILE_AND_EXECUTE);
  glPointSize(1.0);
  glLineWidth(1.0);
  if (render_type_ == MTK_SHADED)
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  else
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

  GLboolean lighting=glIsEnabled(GL_LIGHTING);
  GLenum mode=GL_POINTS;
  mtkState state;
  bool highlight;
  mtkDisplayMap::iterator elem_it;
  list<mtkGeometry>::iterator geo_it;
  list<gmPoint3>::iterator vtx_it;
  list<gmVector3>::iterator nml_it;
  list<mtkColor>::iterator col_it;
  list<mtkMaterial>::iterator mat_it;

  for (elem_it=elements_.begin(); elem_it!=elements_.end(); elem_it++) {
    state=(*elem_it).second.state;
    highlight=(*elem_it).second.highlight;
    if (state == MTK_HIDDEN)
      continue;
    if (state == MTK_ACTIVE) {
      glPushName(MTK_ELEMENT); // type
      glPushName((*elem_it).first); // id
      glPushName(0); // part number
    }

    for (geo_it=(*elem_it).second.geometry.begin();
         geo_it!=(*elem_it).second.geometry.end(); geo_it++) {
      switch ((*geo_it).type) {
      case MTK_POINTS:
        mode=GL_POINTS;
        if (lighting)
          glDisable(GL_LIGHTING);
        glPointSize(4.0); /*FIXME: hardcoded value*/
        break;

      case MTK_LINES:
        mode=GL_LINES;
        if (lighting)
          glDisable(GL_LIGHTING);
        break;

      case MTK_LINE_STRIP:
        mode=GL_LINE_STRIP;
        if (lighting)
          glDisable(GL_LIGHTING);
        break;

      case MTK_POLYGON:
        mode=GL_POLYGON;
        break;

      case MTK_BEZIER:
        float ctrlpoints[64];
        for (vtx_it=(*geo_it).vlist.begin(), i=0;
             vtx_it!=(*geo_it).vlist.end(), i<16;
             vtx_it++, i++) {
          ctrlpoints[i*4+0]=(*vtx_it)[0];
          ctrlpoints[i*4+1]=(*vtx_it)[1];
          ctrlpoints[i*4+2]=(*vtx_it)[2];
          ctrlpoints[i*4+3]=1.0;
        }
        glMap2f(GL_MAP2_VERTEX_4, 0.0, 1.0,  4, 4, 0.0, 1.0, 16, 4,
                ctrlpoints); /*FIXME: hardcoded values*/
        glEnable(GL_MAP2_VERTEX_4);
        break;

      default:
        mode=GL_POINTS; // fallback
        if (lighting)
          glDisable(GL_LIGHTING);
        break;
      }

      // draw a Bezier surface
      if ((*geo_it).type == MTK_BEZIER) {
        if (highlight)
          glColor3f(highlight_color_.r, highlight_color_.g,
                    highlight_color_.b);
        else {
          col_it=(*geo_it).clist.begin();
          glColor3f((*col_it).r, (*col_it).g, (*col_it).b);
          mat_it=(*geo_it).mlist.begin();
          setupMaterial_(*mat_it);
	}
	for (j=0; j<8; j++) { /*FIXME: hardcoded values*/
	  glEvalCoord2f((GLfloat)j/8.0, 0.0);
	  glEvalCoord2f((GLfloat)(j+1)/8.0, 0.0);
	  glBegin(GL_QUAD_STRIP);
          for (i=0; i<=8; i++) {
	    glEvalCoord2f((GLfloat)j/8.0, (GLfloat)i/8.0);
	    glEvalCoord2f((GLfloat)(j+1)/8.0, (GLfloat)i/8.0);
	  }
	  glEnd();
	}
	glDisable(GL_MAP2_VERTEX_4);
      }
      else {
        // draw other objects
        glBegin(mode);
        for (vtx_it=(*geo_it).vlist.begin(),
               nml_it=(*geo_it).nlist.begin(),
               col_it=(*geo_it).clist.begin(),
               mat_it=(*geo_it).mlist.begin();
             vtx_it!=(*geo_it).vlist.end();
             vtx_it++, col_it++, mat_it++) {
          if (nml_it != (*geo_it).nlist.end()) {
            glNormal3d((*nml_it)[0], (*nml_it)[1], (*nml_it)[2]);
            nml_it++;
          }
          if (highlight)
            glColor3f(highlight_color_.r, highlight_color_.g,
                      highlight_color_.b);
          else {
            glColor3f((*col_it).r, (*col_it).g, (*col_it).b);
            setupMaterial_(*mat_it);
          }
          glVertex3d((*vtx_it)[0], (*vtx_it)[1], (*vtx_it)[2]);
        }
      }
      glEnd();

      if (lighting)
        glEnable(GL_LIGHTING);
    }
    if (state == MTK_ACTIVE) {
      glPopName();
      glPopName();
      glPopName();
    }
  }
  glEndList();
  changed_=false;
}

void mtkDisplayListGL::draw (bool changed)
{
  int i, j;

  if (!changed) {
    glCallList(gl_list_);
    return;
  }

  if (!gl_list_)
    gl_list_=glGenLists(1);
  glNewList(gl_list_, GL_COMPILE_AND_EXECUTE);
  glPointSize(1.0);
  glLineWidth(1.0);
  if (render_type_ == MTK_SHADED)
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  else
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

  GLboolean lighting=glIsEnabled(GL_LIGHTING);
  GLenum mode=GL_POINTS;
  mtkState state;
  bool highlight;
  mtkDisplayMap::iterator elem_it;
  list<mtkGeometry>::iterator geo_it;
  list<gmPoint3>::iterator vtx_it;
  list<gmVector3>::iterator nml_it;
  list<mtkColor>::iterator col_it;
  list<mtkMaterial>::iterator mat_it;

  for (elem_it=elements_.begin(); elem_it!=elements_.end(); elem_it++) {
    state=(*elem_it).second.state;
    highlight=(*elem_it).second.highlight;
    if (state == MTK_HIDDEN)
      continue;
    if (state == MTK_ACTIVE) {
      glPushName(MTK_ELEMENT); // type
      glPushName((*elem_it).first); // id
      glPushName(0); // part number
    }

    for (geo_it=(*elem_it).second.geometry.begin();
         geo_it!=(*elem_it).second.geometry.end(); geo_it++) {
      switch ((*geo_it).type) {
      case MTK_POINTS:
        mode=GL_POINTS;
        if (lighting)
          glDisable(GL_LIGHTING);
        glPointSize(4.0); /*FIXME: hardcoded value*/
        break;

      case MTK_LINES:
        mode=GL_LINES;
        if (lighting)
          glDisable(GL_LIGHTING);
        break;

      case MTK_LINE_STRIP:
        mode=GL_LINE_STRIP;
        if (lighting)
          glDisable(GL_LIGHTING);
        break;

      case MTK_POLYGON:
        mode=GL_POLYGON;
        break;

      case MTK_BEZIER:
        float ctrlpoints[64];
        for (vtx_it=(*geo_it).vlist.begin(), i=0;
             vtx_it!=(*geo_it).vlist.end(), i<16;
             vtx_it++, i++) {
          ctrlpoints[i*4+0]=(*vtx_it)[0];
          ctrlpoints[i*4+1]=(*vtx_it)[1];
          ctrlpoints[i*4+2]=(*vtx_it)[2];
          ctrlpoints[i*4+3]=1.0;
        }
        glMap2f(GL_MAP2_VERTEX_4, 0.0, 1.0,  4, 4, 0.0, 1.0, 16, 4,
                ctrlpoints); /*FIXME: hardcoded values*/
        glEnable(GL_MAP2_VERTEX_4);
        break;

      default:
        mode=GL_POINTS; // fallback
        if (lighting)
          glDisable(GL_LIGHTING);
        break;
      }

      // draw a Bezier surface
      if ((*geo_it).type == MTK_BEZIER) {
        if (highlight)
          glColor3f(highlight_color_.r, highlight_color_.g,
                    highlight_color_.b);
        else {
          col_it=(*geo_it).clist.begin();
          glColor3f((*col_it).r, (*col_it).g, (*col_it).b);
          mat_it=(*geo_it).mlist.begin();
          setupMaterial_(*mat_it);
	}
	for (j=0; j<8; j++) { /*FIXME: hardcoded values*/
	  glEvalCoord2f((GLfloat)j/8.0, 0.0);
	  glEvalCoord2f((GLfloat)(j+1)/8.0, 0.0);
	  glBegin(GL_QUAD_STRIP);
          for (i=0; i<=8; i++) {
	    glEvalCoord2f((GLfloat)j/8.0, (GLfloat)i/8.0);
	    glEvalCoord2f((GLfloat)(j+1)/8.0, (GLfloat)i/8.0);
	  }
	  glEnd();
	}
	glDisable(GL_MAP2_VERTEX_4);
      }
      else {
        // draw other objects
        glBegin(mode);
        for (vtx_it=(*geo_it).vlist.begin(),
               nml_it=(*geo_it).nlist.begin(),
               col_it=(*geo_it).clist.begin(),
               mat_it=(*geo_it).mlist.begin();
             vtx_it!=(*geo_it).vlist.end();
             vtx_it++, col_it++, mat_it++) {
          if (nml_it != (*geo_it).nlist.end()) {
            glNormal3d((*nml_it)[0], (*nml_it)[1], (*nml_it)[2]);
            nml_it++;
          }
          if (highlight)
            glColor3f(highlight_color_.r, highlight_color_.g,
                      highlight_color_.b);
          else {
            glColor3f((*col_it).r, (*col_it).g, (*col_it).b);
            setupMaterial_(*mat_it);
          }
          glVertex3d((*vtx_it)[0], (*vtx_it)[1], (*vtx_it)[2]);
        }
      }
      glEnd();

      if (lighting)
        glEnable(GL_LIGHTING);
    }
    if (state == MTK_ACTIVE) {
      glPopName();
      glPopName();
      glPopName();
    }
  }
  glEndList();
}

