// -*- mode: C++; -*-

/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef MTK_DISPLAYLIST_H
#define MTK_DISPLAYLIST_H

#include <map.h>
#include <list.h>

#include "gm.h"
#include "mtk_general.h"


/// Types of rendering available in the display list
enum mtkRenderType {MTK_WIREFRAME, MTK_SHADED};

/// Types of geometric entities handled by the display list
enum mtkGeometricType {MTK_POINTS, MTK_LINES, MTK_LINE_STRIP, MTK_POLYGON,
                       MTK_BEZIER};

/// Specifies which face or faces are updated
enum mtkFaceMode {MTK_FRONT, MTK_BACK, MTK_FRONT_AND_BACK};

typedef bool mtkElementCallback (list<mtkId> on, list<mtkId> off, void *data);


class mtkMaterial {
public:
  mtkMaterial ();
  mtkMaterial& operator= (const mtkMaterial& m);

  mtkFaceMode fmode;
  float ambient[4], diffuse[4], specular[4], emission[4];
  float shininess;
};


class mtkGeometry {
public:
  mtkGeometry ();
  mtkGeometry& operator= (const mtkGeometry& g);

  void reset ();

  mtkGeometricType type;
  list<gmPoint3> vlist;
  list<gmVector3> nlist;
  list<mtkColor> clist;
  list<mtkMaterial> mlist;
};


class mtkElement {
public:
  mtkElement ();
  mtkElement& operator= (const mtkElement& e);

  void clear ();
  void reset ();

  mtkState state;
  bool highlight;
  list<mtkGeometry> geometry;
  mtkElementCallback *callback;
  void *cb_data;
};


class mtkDisplayList {
public:
  /// Set default values
  mtkDisplayList ();
  virtual ~mtkDisplayList () {}

  /// Sets how objects are rendered in the display list: wireframe or shaded
  void setRenderType (mtkRenderType r);
  /// Get rendering mode
  mtkRenderType getRenderType ();
  void setHighlightColor (float r, float g, float b);

  /// Create a display list element
  bool createElement (mtkId id);
  /// Appends geometric data to an existing element
  bool appendElement (mtkId id);
  /// Finish a display list element
  void finishElement ();
  /// Test if there is an element in the display list with the given id
  bool checkElement (mtkId id);
  /// Erase an element in the display list, given its identifier
  bool deleteElement (mtkId id);
  /// Empty an element, given its identifier
  bool clearElement (mtkId id);
  list<mtkId> getAllElements ();

  /// Erase all elements currently in the display list
  void clearDisplayList ();

  /// Sets the current material for elements sent to display list
  void setMaterialv(mtkFaceMode fmode, mtkMaterialComponent component,
                    gmPoint4 value);
  void setMaterialf(mtkFaceMode fmode, mtkMaterialComponent component,
                    float value);

  /// Sets the current color for elements sent to display list
  void setColor (float r, float g, float b);

  /// Starts the definition of a geometric type
  void begin (mtkGeometricType g);
  /// Declare a display list coordinate
  void setVertex (gmPoint3 p);
  void setNormal (gmVector3 n);
  /// Create a Bezier surface
  void bezier (int uorder, int vorder, int nu, int vu, gmPoint3 *ctrlpoints);
  /// Finishes a polygon or a group of points or lines sent to display list
  void end ();

  void applyMatrix (list<mtkId> id_list, gmMatrix4 m);
  bool calcBoundingBox (list<mtkId> id_list, gmPoint3 *min, gmPoint3 *max);
  bool calcBoundingSphere (list<mtkId> id_list, gmPoint3 *center,
                           double *radius);

  void setState (mtkId id, mtkState state);
  mtkState getState (mtkId id);
  void setHighlight (mtkId id, bool h);
  bool isHighlighted (mtkId id);

  bool setCallback (mtkId id, mtkElementCallback *callback, void *data);
  mtkElementCallback *getCallback (mtkId id, void **datap);

  virtual void draw () = 0;
  // Added by Leonardo Mesquita - 15/02/2001
  virtual void draw (bool) = 0;

protected:
  // Defines the map that holds all data of the display list
  typedef map<mtkId, mtkElement, less<mtkId> > mtkDisplayMap;
  typedef pair<const mtkId, mtkElement> mtkDisplayPair;

  gmVector3 calcNormal_ (list<gmPoint3> vlist);

  /* This map holds an identifier to indicate each element
     (point, line, polygon or group of them) that can be
     selected, and associates each identifier with its
     corresponding data */
  mtkDisplayMap elements_;

  // signal changes on resulting output
  bool changed_;

  // Defines if objects are rendered in wire frame or shaded
  mtkRenderType render_type_;
  // Select color
  mtkColor highlight_color_;

  // These variables keep the current state of the display list
  mtkId c_id_;
  bool elock_; // element input lock
  bool glock_; // geometry input lock
  // Keeps vertices sent to create points, lines and polygons
  mtkGeometry c_geometry_;
  mtkElement c_element_;
  // current normal
  // Current primitive material
  mtkMaterial c_material_;
  // Current primitive color
  mtkColor c_color_;
};


/**=========================================================
   Derivates Class mtkDisplayList to include OpenGL functions
 ===========================================================
*/
class mtkDisplayListGL : public mtkDisplayList {
public:
  mtkDisplayListGL ();
  ~mtkDisplayListGL () {}

  void draw ();
  // Added by Leonardo Mesquita - 15/02/2001
  virtual void draw (bool);

protected:
  void setupMaterial_ (mtkMaterial mat);

  GLuint gl_list_;
};


#endif  // MTK_DISPLAYLIST_H
