/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <GL/glu.h>
#include "mtk_dragger.h"


/*********************************************************************
 mtkDragger
**********************************************************************/

mtkDraggerId mtkDragger::counter_=1;

mtkDragger::mtkDragger ()
{
  id_=counter_++;
  state_=MTK_ACTIVE;
  size_=0.0;
}


/// get dragger id
mtkDraggerId mtkDragger::getId ()
{
  return id_;
}


/// set dragger state
void mtkDragger::setState (mtkState state)
{
  state_=state;
}


/// get dragger state
mtkState mtkDragger::getState ()
{
  return state_;
}


void mtkDragger::setSize (double size)
{
  size_=size;
}


double mtkDragger::getSize ()
{
  return size_;
}


/*********************************************************************
 mtkBoundingBox
**********************************************************************/

mtkBoundingBox::mtkBoundingBox ()
{
  v_[0].assign(-0.5, -0.5, -0.5);
  v_[1].assign(-0.5, -0.5,  0.5);
  v_[2].assign(-0.5,  0.5, -0.5);
  v_[3].assign(-0.5,  0.5,  0.5);
  v_[4].assign( 0.5, -0.5, -0.5);
  v_[5].assign( 0.5, -0.5,  0.5);
  v_[6].assign( 0.5,  0.5, -0.5);
  v_[7].assign( 0.5,  0.5,  0.5);
  vshow_=eshow_=fshow_=true;
  vertex_size_=5;
  vertex_color_.r=1.0;
  vertex_color_.g=vertex_color_.b=0.0;
  edge_thickness_=1;
  edge_color_.r=edge_color_.g=edge_color_.b=1.0;
  face_color_.r=0.0;
  face_color_.g=face_color_.b=1.0;
  face_opacity_=0.4;
  callback_=0;
  data_=0;
}


mtkDraggerType mtkBoundingBox::getType ()
{
  return MTK_BOUNDING_BOX;
}


void mtkBoundingBox::setGeometry (gmPoint3 v0,
                                  gmVector3 e01, gmVector3 e02, gmVector3 e04)
{
  // set vertices
  v_[0]=v0;
  v_[1]=v0+e01;
  v_[2]=v0+e02;
  v_[3]=v0+e02+e01;
  v_[4]=v0+e04;
  v_[5]=v0+e04+e01;
  v_[6]=v0+e04+e02;
  v_[7]=v0+e04+e02+e01;

  // update box using size factor
  gmPoint3 c=gmLerp(0.5, v_[0], v_[7]);
  int i;
  for (i=0; i<=7; i++)
    v_[i]+=size_*(v_[i]-c);
}


gmPoint3 mtkBoundingBox::getVertex (mtkDraggerPart part)
{
  return v_[part-MTK_BOX_V_0];
}


gmPoint3 mtkBoundingBox::getCenter ()
{
  return gmLerp(0.5, v_[0], v_[7]);
}


void mtkBoundingBox::setVisibility (mtkDraggerPart part, bool b)
{
  if (part>=MTK_BOX_V_0 && part<=MTK_BOX_V_7)
    vshow_=b;
  else if (part>=MTK_BOX_E_01 && part<=MTK_BOX_E_67)
    eshow_=b;
  else if (part>=MTK_BOX_F_0132 && part<=MTK_BOX_F_0451)
    fshow_=b;
}


bool mtkBoundingBox::getVisibility (mtkDraggerPart part)
{
  bool ret=false;
  if (part>=MTK_BOX_V_0 && part<=MTK_BOX_V_7)
    ret=vshow_;
  else if (part>=MTK_BOX_E_01 && part<=MTK_BOX_E_67)
    ret=eshow_;
  else if (part>=MTK_BOX_F_0132 && part<=MTK_BOX_F_0451)
    ret=fshow_;
  return ret;
}


void mtkBoundingBox::setVertexSize (unsigned int size)
{
  vertex_size_=size;
}


void mtkBoundingBox::setVertexColor (float r, float g, float b)
{
  vertex_color_.r=r;
  vertex_color_.g=g;
  vertex_color_.b=b;
}


void mtkBoundingBox::setEdgeThickness (unsigned int thickness)
{
  edge_thickness_=thickness;
}


void mtkBoundingBox::setEdgeColor (float r, float g, float b)
{
  edge_color_.r=r;
  edge_color_.g=g;
  edge_color_.b=b;
}


void mtkBoundingBox::setFaceColor (float r, float g, float b, float a)
{
  face_color_.r=r;
  face_color_.g=g;
  face_color_.b=b;
  face_opacity_=a;
}


void mtkBoundingBox::applyMatrix (gmMatrix4 m)
{
  // applies transformations on vertices
  for(int i=0; i<8; i++)
    v_[i]=m.transform(v_[i]);
}


void mtkBoundingBox::setCallback(mtkBoxCallback *callback, void *data)
{
  callback_=callback;
  data_=data;
}


/*********************************************************************
 mtkTrackball
**********************************************************************/

mtkTrackball::mtkTrackball ()
{
  matrix_=matrix_.identity(); // center at origin and radius=1.0
  arrow_=0.5;
  slices_=28;
  stacks_=14;
  vshow_=eshow_=rshow_[0]=rshow_[1]=rshow_[2]=sshow_=true;
  vertex_size_=5;
  vertex_color_.r=1.0;
  vertex_color_.g=vertex_color_.b=0.0;
  edge_thickness_=1;
  edge_color_.r=edge_color_.g=edge_color_.b=1.0;
  face_color_.r=0.0;
  face_color_.g=face_color_.b=1.0;
  face_opacity_=0.4;
  callback_=0;
  data_=0;
}


mtkDraggerType mtkTrackball::getType ()
{
  return MTK_TRACKBALL;
}

void mtkTrackball::setGeometry (gmPoint3 center, double radius, double arrow)
{
  radius*=1.0+size_;
  matrix_=matrix_.scaling(radius, radius, radius);
  matrix_=matrix_.translation(center[0], center[1], center[2])*matrix_;
  arrow_=arrow;
}


void mtkTrackball::setResolution (int slices, int stacks)
{
  slices_=slices;
  stacks_=stacks;
}


void mtkTrackball::getSphere (gmPoint3 *center, double *radius)
{
  gmPoint3 p(0.0, 0.0, 0.0);
  *center=matrix_.transform(p);
  p.assign(1.0, 0.0, 0.0);
  *radius=gmDistance(*center, matrix_.transform(p));
}


gmPoint3 mtkTrackball::getArrow ()
{
  gmPoint3 p(1.0+arrow_, 0.0, 0.0);
  return matrix_.transform(p);
}


gmVector3 mtkTrackball::getRing (mtkDraggerPart part)
{
  gmPoint3 c, p1, p2;

  c.assign(0.0, 0.0, 0.0);
  c=matrix_.transform(c);
  if (part == MTK_BALL_R_0) {
    // XY-ring
    p1.assign(1.0, 0.0, 0.0);
    p2.assign(0.0, 1.0, 0.0);
  }
  else if (part == MTK_BALL_R_1) {
    // ZX-ring
    p1.assign(0.0, 0.0, 1.0);
    p2.assign(1.0, 0.0, 0.0);
  }
  else {
    // YZ-ring
    p1.assign(0.0, 1.0, 0.0);
    p2.assign(0.0, 0.0, 1.0);
  }
  p1=matrix_.transform(p1);
  p2=matrix_.transform(p2);
  return (gmCross(p1-c, p2-c)).normalize();
}


void mtkTrackball::setVisibility (mtkDraggerPart part, bool b)
{
  if (part == MTK_BALL_V)
    vshow_=b;
  else if (part == MTK_BALL_E)
    eshow_=b;
  else if (part>=MTK_BALL_R_0 && part<=MTK_BALL_R_2)
    rshow_[part-MTK_BALL_R_0]=b;
  else if (part == MTK_BALL_S)
    sshow_=b;
}


bool mtkTrackball::getVisibility (mtkDraggerPart part)
{
  bool ret=false;
  if (part == MTK_BALL_V)
    ret=vshow_;
  else if (part == MTK_BALL_E)
    ret=eshow_;
  else if (part>=MTK_BALL_R_0 && part<=MTK_BALL_R_2)
    ret=rshow_[part-MTK_BALL_R_0];
  else if (part == MTK_BALL_S)
    ret=sshow_;
  return ret;
}


void mtkTrackball::setVertexSize (unsigned int size)
{
  vertex_size_=size;
}


void mtkTrackball::setVertexColor (float r, float g, float b)
{
  vertex_color_.r=r;
  vertex_color_.g=g;
  vertex_color_.b=b;
}


void mtkTrackball::setEdgeThickness (unsigned int thickness)
{
  edge_thickness_=thickness;
}


void mtkTrackball::setEdgeColor (float r, float g, float b)
{
  edge_color_.r=r;
  edge_color_.g=g;
  edge_color_.b=b;
}


void mtkTrackball::setFaceColor (float r, float g, float b, float a)
{
  face_color_.r=r;
  face_color_.g=g;
  face_color_.b=b;
  face_opacity_=a;
}


void mtkTrackball::applyMatrix (gmMatrix4 m)
{
  matrix_=m*matrix_;
}


void mtkTrackball::setCallback(mtkTrackballCallback *callback, void *data)
{
  callback_=callback;
  data_=data;
}


/*********************************************************************
 mtkJack
**********************************************************************/

mtkJack::mtkJack ()
{
  v_[0].assign( 0.0,  0.0,  0.0);
  v_[1].assign( 0.5,  0.0,  0.0);
  v_[2].assign(-0.5,  0.0,  0.0);
  v_[3].assign( 0.0,  0.5,  0.0);
  v_[4].assign( 0.0, -0.5,  0.0);
  v_[5].assign( 0.0,  0.0,  0.5);
  v_[6].assign( 0.0,  0.0, -0.5);
  int i;
  for (i=0; i<7; i++)
    vshow_[i]=true;
  for (i=0; i<6; i++)
    eshow_[i]=true;
  vertex_size_=5;
  vertex_color_.r=1.0;
  vertex_color_.g=vertex_color_.b=0.0;
  edge_thickness_=1;
  edge_color_.r=edge_color_.g=edge_color_.b=1.0;
  callback_=0;
  data_=0;
}


mtkDraggerType mtkJack::getType ()
{
  return MTK_JACK;
}

void mtkJack::setGeometry (gmPoint3 v0,
                           gmVector3 e01, gmVector3 e03, gmVector3 e05)
{
  // set vertices
  e01*=0.5*(1.0+size_);
  e03*=0.5*(1.0+size_);
  e05*=0.5*(1.0+size_);
  v_[0]=v0;
  v_[1]=v0+e01;
  v_[2]=v0+(-e01);
  v_[3]=v0+e03;
  v_[4]=v0+(-e03);
  v_[5]=v0+e05;
  v_[6]=v0+(-e05);
}


gmPoint3 mtkJack::getVertex(mtkDraggerPart part)
{
  return v_[part-MTK_JACK_V_0];
}


void mtkJack::setVisibility (mtkDraggerPart part, bool b)
{
  if (part>=MTK_JACK_V_0 && part<=MTK_JACK_V_6)
    vshow_[part-MTK_JACK_V_0]=b;
  else if (part>=MTK_JACK_E_01 && part<=MTK_JACK_E_06)
    eshow_[part-MTK_JACK_E_01]=b;
}


bool mtkJack::getVisibility (mtkDraggerPart part)
{
  bool ret=false;
  if (part>=MTK_JACK_V_0 && part<=MTK_JACK_V_6)
    ret=vshow_[part-MTK_JACK_V_0];
  else if (part>=MTK_JACK_E_01 && part<=MTK_JACK_E_06)
    ret=eshow_[part-MTK_JACK_E_01];
  return ret;
}


void mtkJack::setVertexSize (unsigned int size)
{
  vertex_size_=size;
}


void mtkJack::setVertexColor (float r, float g, float b)
{
  vertex_color_.r=r;
  vertex_color_.g=g;
  vertex_color_.b=b;
}


void mtkJack::setEdgeThickness (unsigned int thickness)
{
  edge_thickness_=thickness;
}


void mtkJack::setEdgeColor (float r, float g, float b)
{
  edge_color_.r=r;
  edge_color_.g=g;
  edge_color_.b=b;
}


void mtkJack::applyMatrix (gmMatrix4 m)
{
  // applies transformations on vertices
  for(int i=0; i<7; i++)
    v_[i]=m.transform(v_[i]);
}


void mtkJack::setCallback(mtkJackCallback *callback, void *data)
{
  callback_=callback;
  data_=data;
}


/*********************************************************************
 mtkCursor3D
**********************************************************************/

mtkCursor3D::mtkCursor3D ()
{
  v_[0].assign( 0.0,  0.0,  0.0);
  v_[1].assign( 0.5,  0.0,  0.0);
  v_[2].assign(-0.5,  0.0,  0.0);
  v_[3].assign( 0.0,  0.5,  0.0);
  v_[4].assign( 0.0, -0.5,  0.0);
  v_[5].assign( 0.0,  0.0,  0.5);
  v_[6].assign( 0.0,  0.0, -0.5);
  int i;
  for (i=0; i<7; i++)
    vshow_[i]=true;
  for (i=0; i<6; i++)
    eshow_[i]=true;
  vertex_size_=5;
  vertex_color_.r=1.0;
  vertex_color_.g=vertex_color_.b=0.0;
  edge_thickness_=1;
  edge_color_.r=edge_color_.g=edge_color_.b=1.0;
  callback_=0;
  data_=0;
}

mtkDraggerType mtkCursor3D::getType ()
{
  return MTK_CURSOR3D;
}


void mtkCursor3D::setGeometry (gmPoint3 v0,
                           gmVector3 e01, gmVector3 e03, gmVector3 e05)
{
  // set vertices
  e01*=0.5*(1.0+size_);
  e03*=0.5*(1.0+size_);
  e05*=0.5*(1.0+size_);
  v_[0]=v0;
  v_[1]=v0+e01;
  v_[2]=v0+(-e01);
  v_[3]=v0+e03;
  v_[4]=v0+(-e03);
  v_[5]=v0+e05;
  v_[6]=v0+(-e05);
}


gmPoint3 mtkCursor3D::getVertex(mtkDraggerPart part)
{
  return v_[part-MTK_CURSOR3D_V_0];
}


void mtkCursor3D::setVisibility (mtkDraggerPart part, bool b)
{
  if (part>=MTK_CURSOR3D_V_0 && part<=MTK_CURSOR3D_V_6)
    vshow_[part-MTK_CURSOR3D_V_0]=b;
  else if (part>=MTK_CURSOR3D_E_01 && part<=MTK_CURSOR3D_E_06)
    eshow_[part-MTK_CURSOR3D_E_01]=b;
}


bool mtkCursor3D::getVisibility (mtkDraggerPart part)
{
  bool ret=false;
  if (part>=MTK_CURSOR3D_V_0 && part<=MTK_CURSOR3D_V_6)
    ret=vshow_[part-MTK_CURSOR3D_V_0];
  else if (part>=MTK_CURSOR3D_E_01 && part<=MTK_CURSOR3D_E_06)
    ret=eshow_[part-MTK_CURSOR3D_E_01];
  return ret;
}


void mtkCursor3D::setVertexSize (unsigned int size)
{
  vertex_size_=size;
}


void mtkCursor3D::setVertexColor (float r, float g, float b)
{
  vertex_color_.r=r;
  vertex_color_.g=g;
  vertex_color_.b=b;
}


void mtkCursor3D::setEdgeThickness (unsigned int thickness)
{
  edge_thickness_=thickness;
}


void mtkCursor3D::setEdgeColor (float r, float g, float b)
{
  edge_color_.r=r;
  edge_color_.g=g;
  edge_color_.b=b;
}


void mtkCursor3D::applyMatrix (gmMatrix4 m)
{
  // applies transformations on vertices
  for(int i=0; i<7; i++)
    v_[i]=m.transform(v_[i]);
}


void mtkCursor3D::setCallback(mtkCursor3DCallback *callback, void *data)
{
  callback_=callback;
  data_=data;
}

/*********************************************************************
 mtkBoundingBoxGL
**********************************************************************/

mtkBoundingBoxGL::mtkBoundingBoxGL () : mtkBoundingBox()
{
}


bool mtkBoundingBoxGL::setConstraint (mtkDraggerPart part, mtkConstraintGL cstr)
{
  if (part>=MTK_BOX_V_0 && part<=MTK_BOX_V_7)
    vconst_=cstr;
  else if (part>=MTK_BOX_E_01 && part<=MTK_BOX_E_67)
    econst_=cstr;
  else if (part>=MTK_BOX_F_0132 && part<=MTK_BOX_F_0451)
    fconst_=cstr;
  else
    return false;
  return true;
}


mtkDragger *mtkBoundingBoxGL::clone () const
{
  mtkBoundingBoxGL *ptr=new mtkBoundingBoxGL();
  (*ptr)=*this;
  return ptr;
}


void mtkBoundingBoxGL::draw (bool pick)
{
  if (state_ == MTK_HIDDEN)
    return;

  if (state_ == MTK_ACTIVE) {
    glPushName(MTK_DRAGGER);
    glPushName(id_);
    glPushName(0);
  }

  GLboolean lighting=glIsEnabled(GL_LIGHTING);
  if (lighting)
    glDisable(GL_LIGHTING);

  // Draws bounding box vertices
  if (vshow_) {
    glColor3f(vertex_color_.r, vertex_color_.g, vertex_color_.b);
    glPointSize(vertex_size_);

    if (state_ == MTK_ACTIVE)
      for (int i=0; i<8; i++) {
        glLoadName(MTK_BOX_V_0+i);
        glBegin(GL_POINTS);
        glVertex3f(v_[i][0], v_[i][1], v_[i][2]);
        glEnd();
      }
    else {
      glBegin(GL_POINTS);
      for (int i=0; i<8; i++)
        glVertex3f(v_[i][0], v_[i][1], v_[i][2]);
      glEnd();
    }
  }

  // Draws bounding box edges
  if (eshow_) {
    glColor3f(edge_color_.r, edge_color_.g, edge_color_.b);
    glLineWidth(edge_thickness_);

    if (state_ == MTK_ACTIVE) {
      glLoadName(MTK_BOX_E_01);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glEnd();
      glLoadName(MTK_BOX_E_02);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glEnd();
      glLoadName(MTK_BOX_E_04);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glEnd();
      glLoadName(MTK_BOX_E_31);
      glBegin(GL_LINES);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glEnd();
      glLoadName(MTK_BOX_E_32);
      glBegin(GL_LINES);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glEnd();
      glLoadName(MTK_BOX_E_37);
      glBegin(GL_LINES);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glVertex3f(v_[7][0], v_[7][1], v_[7][2]);
      glEnd();
      glLoadName(MTK_BOX_E_51);
      glBegin(GL_LINES);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glEnd();
      glLoadName(MTK_BOX_E_54);
      glBegin(GL_LINES);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glEnd();
      glLoadName(MTK_BOX_E_57);
      glBegin(GL_LINES);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glVertex3f(v_[7][0], v_[7][1], v_[7][2]);
      glEnd();
      glLoadName(MTK_BOX_E_62);
      glBegin(GL_LINES);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glEnd();
      glLoadName(MTK_BOX_E_64);
      glBegin(GL_LINES);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glEnd();
      glLoadName(MTK_BOX_E_67);
      glBegin(GL_LINES);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glVertex3f(v_[7][0], v_[7][1], v_[7][2]);
      glEnd();
    }
    else {
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glVertex3f(v_[7][0], v_[7][1], v_[7][2]);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glVertex3f(v_[7][0], v_[7][1], v_[7][2]);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glVertex3f(v_[7][0], v_[7][1], v_[7][2]);
      glEnd();
    }
  }

  // Draws bounding box faces
  if (fshow_) {
    if (face_opacity_>0.01 && face_opacity_<0.99) {
      glPolygonMode(GL_FRONT, GL_FILL);
      glEnable(GL_BLEND);
      glDepthMask(GL_FALSE);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      glColor4f(face_color_.r, face_color_.g, face_color_.b, face_opacity_);
    }
    else
    if (face_opacity_ > 0.99) {
      glColor3f(face_color_.r, face_color_.g, face_color_.b);
      glPolygonMode(GL_FRONT, GL_FILL);
    }

    if ((face_opacity_<0.01 && pick==true) || face_opacity_>=0.01) {
    if (state_ == MTK_ACTIVE) {
      glLoadName(MTK_BOX_F_0132);
      glBegin(GL_POLYGON);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glEnd();
      glLoadName(MTK_BOX_F_1573);
      glBegin(GL_POLYGON);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glVertex3f(v_[7][0], v_[7][1], v_[7][2]);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glEnd();
      glLoadName(MTK_BOX_F_4675);
      glBegin(GL_POLYGON);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glVertex3f(v_[7][0], v_[7][1], v_[7][2]);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glEnd();
      glLoadName(MTK_BOX_F_0264);
      glBegin(GL_POLYGON);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glEnd();
      glLoadName(MTK_BOX_F_2376);
      glBegin(GL_POLYGON);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glVertex3f(v_[7][0], v_[7][1], v_[7][2]);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glEnd();
      glLoadName(MTK_BOX_F_0451);
      glBegin(GL_POLYGON);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glEnd();
    }
    else {
      glBegin(GL_QUADS);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glVertex3f(v_[7][0], v_[7][1], v_[7][2]);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glVertex3f(v_[7][0], v_[7][1], v_[7][2]);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glVertex3f(v_[7][0], v_[7][1], v_[7][2]);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glEnd();
    }
    }
    if (face_opacity_>0.01 && face_opacity_<0.99) {
      glDepthMask(GL_TRUE);
      glDisable(GL_BLEND);
    }
  }

  if (lighting)
    glEnable(GL_LIGHTING);

  if (state_ == MTK_ACTIVE) {
    glPopName();
    glPopName();
    glPopName();
  }
}


bool mtkBoundingBoxGL::interaction (mtkDraggerPart part, int x, int y,
                                   mtkEvent type)
{
  bool ok, ret=false;
  gmPoint3 p;
  gmVector3 v;
  switch (type) {
    case MTK_PICK:
      if (part>=MTK_BOX_V_0 && part<=MTK_BOX_V_7) {
        // handle vertices
        p=v_[part-MTK_BOX_V_0];
        if (callback_)
          ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                           MTK_PICK, data_);
        vconst_.start(x, y, p);
      }
      else if (part>=MTK_BOX_E_01 && part<=MTK_BOX_E_67) {
        // handle edges
        gmPoint3 o1, o2;

        // get edge vertices
        switch (part) {
          case MTK_BOX_E_01:
            o1=v_[0];
            o2=v_[1];
            break;
          case MTK_BOX_E_02:
            o1=v_[0];
            o2=v_[2];
            break;
          case MTK_BOX_E_04:
            o1=v_[0];
            o2=v_[4];
            break;
          case MTK_BOX_E_31:
            o1=v_[3];
            o2=v_[1];
            break;
          case MTK_BOX_E_32:
            o1=v_[3];
            o2=v_[2];
            break;
          case MTK_BOX_E_37:
            o1=v_[3];
            o2=v_[7];
            break;
          case MTK_BOX_E_51:
            o1=v_[5];
            o2=v_[1];
            break;
          case MTK_BOX_E_54:
            o1=v_[5];
            o2=v_[4];
            break;
          case MTK_BOX_E_57:
            o1=v_[5];
            o2=v_[7];
            break;
          case MTK_BOX_E_62:
            o1=v_[6];
            o2=v_[2];
            break;
          case MTK_BOX_E_64:
            o1=v_[6];
            o2=v_[4];
            break;
          case MTK_BOX_E_67:
            o1=v_[6];
            o2=v_[7];
            break;
        }
        econst_.edgeIntersect(x, y, o1, o2, &p);
        if (callback_)
          ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                           MTK_PICK, data_);
        econst_.start(x, y, p);
      }
      else if (part>=MTK_BOX_F_0132 && part<=MTK_BOX_F_0451) {
        // handle faces
        gmPoint3 o1, o2, o3;

        // get face vertices
        switch (part) {
          case MTK_BOX_F_0132:
            o1=v_[0];
            o2=v_[1];
            o3=v_[3];
            break;
          case MTK_BOX_F_1573:
            o1=v_[1];
            o2=v_[5];
            o3=v_[7];
            break;
          case MTK_BOX_F_4675:
            o1=v_[4];
            o2=v_[6];
            o3=v_[7];
            break;
          case MTK_BOX_F_0264:
            o1=v_[0];
            o2=v_[2];
            o3=v_[6];
            break;
          case MTK_BOX_F_2376:
            o1=v_[2];
            o2=v_[3];
            o3=v_[7];
            break;
          case MTK_BOX_F_0451:
            o1=v_[0];
            o2=v_[4];
            o3=v_[5];
            break;
        }
        fconst_.faceIntersect(x, y, o1, o2, o3, &p);
        if (callback_)
          ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                           MTK_PICK, data_);
        fconst_.start(x, y, p);
      }
      break;

    case MTK_MOVE:
      ok=false;
      if (part>=MTK_BOX_V_0 && part<=MTK_BOX_V_7) {
        ok=vconst_.next(x, y, &p, &v);
        if (ok && callback_)
          ret=(*callback_)(id_, part, p, v, MTK_MOVE, data_);
        if (vconst_.getPhase() == MTK_SETUP_PHASE)
          vconst_.start(x, y, p+v);
      }
      else if (part>=MTK_BOX_E_01 && part<=MTK_BOX_E_67) {
        ok=econst_.next(x, y, &p, &v);
        if (ok && callback_)
          ret=(*callback_)(id_, part, p, v, MTK_MOVE, data_);
        if (econst_.getPhase() == MTK_SETUP_PHASE)
          econst_.start(x, y, p+v);
      }
      else if (part>=MTK_BOX_F_0132 && part<=MTK_BOX_F_0451) {
        ok=fconst_.next(x, y, &p, &v);
        if (ok && callback_)
          ret=(*callback_)(id_, part, p, v, MTK_MOVE, data_);
        if (fconst_.getPhase() == MTK_SETUP_PHASE)
          fconst_.start(x, y, p+v);
      }
      break;

    case MTK_RELEASE:
      if (part>=MTK_BOX_V_0 && part<=MTK_BOX_V_7)
        vconst_.end(&p);
      else if (part>=MTK_BOX_E_01 && part<=MTK_BOX_E_67)
        econst_.end(&p);
      else if (part>=MTK_BOX_F_0132 && part<=MTK_BOX_F_0451)
        fconst_.end(&p);
      if (callback_)
        ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                         MTK_RELEASE, data_);
      break;
  }
  return ret;
}


/*********************************************************************
 mtkTrackballGL
**********************************************************************/

mtkTrackballGL::mtkTrackballGL () : mtkTrackball()
{
}


bool mtkTrackballGL::setConstraint (mtkDraggerPart part, mtkConstraintGL cstr)
{
  if (part == MTK_BALL_V)
    vconst_=cstr;
  else if (part == MTK_BALL_E)
    econst_=cstr;
  else if (part>=MTK_BALL_R_0 && part<=MTK_BALL_R_2)
    rconst_=cstr;
  else if (part == MTK_BALL_S)
    sconst_=cstr;
  else
    return false;
  return true;
}


mtkDragger *mtkTrackballGL::clone () const
{
  mtkTrackballGL *ptr=new mtkTrackballGL();
  (*ptr)=*this;
  return ptr;
}


void mtkTrackballGL::draw (bool pick)
{
  if (state_ == MTK_HIDDEN)
    return;

  if (state_ == MTK_ACTIVE) {
    glPushName(MTK_DRAGGER);
    glPushName(id_);
    glPushName(0);
  }

  int i, j;

  GLboolean lighting=glIsEnabled(GL_LIGHTING);
  if (lighting)
    glDisable(GL_LIGHTING);

  // concatenate the current matrix (MODELmtkEW) with trackball matrix
  GLdouble m[16];
  for (i=0; i<4; i++)
    for (j=0; j<4; j++)
      m[i*4+j]=matrix_[j][i]; glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glMultMatrixd(m);

  // draw arrow vertex
  if (vshow_) {
    glColor3f(vertex_color_.r, vertex_color_.g, vertex_color_.b);
    glPointSize(vertex_size_);
    if (state_ == MTK_ACTIVE)
      glLoadName(MTK_BALL_V);
    glBegin(GL_POINTS);
    glVertex3f(1.0+arrow_, 0.0, 0.0);
    glEnd();
  }

  // draw arrow segment
  if (eshow_) {
    glColor3f(edge_color_.r, edge_color_.g, edge_color_.b);
    glLineWidth(edge_thickness_);
    if (state_ == MTK_ACTIVE)
      glLoadName(MTK_BALL_E);
    glBegin(GL_LINE_LOOP);
    glVertex3f(1.0, 0.0, 0.0);
    glVertex3f(1.0+arrow_, 0.0, 0.0);
    glEnd();
  }

  // draw rings
  double a;
  glColor3f(edge_color_.r, edge_color_.g, edge_color_.b);
  glLineWidth(edge_thickness_);
  if (rshow_[0]) {
    if (state_ == MTK_ACTIVE)
      glLoadName(MTK_BALL_R_0);
    glBegin(GL_LINE_LOOP); // XY-ring
    for (i=0; i<slices_; i++) {
      a=GM_2PI*i/slices_;
      glVertex3f(cos(a), sin(a), 0.0);
    }
    glEnd();
  }
  if (rshow_[1]) {
    if (state_ == MTK_ACTIVE)
      glLoadName(MTK_BALL_R_1);
    glBegin(GL_LINE_LOOP); // ZX-ring
    for (i=0; i<slices_; i++) {
      a=GM_2PI*i/slices_;
      glVertex3f(cos(a), 0.0, sin(a));
    }
    glEnd();
  }
  if (rshow_[2]) {
    if (state_ == MTK_ACTIVE)
      glLoadName(MTK_BALL_R_2);
    glBegin(GL_LINE_LOOP); // YZ-ring
    for (i=0; i<slices_; i++) {
      a=GM_2PI*i/slices_;
      glVertex3f(0.0, cos(a), sin(a));
    }
    glEnd();
  }

  // draw sphere surface
  if (sshow_) {

    glPolygonMode(GL_FRONT, GL_FILL);
    if (face_opacity_>0.01 && face_opacity_<0.99) {
      glEnable(GL_BLEND);
      glDepthMask(GL_FALSE);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      glColor4f(face_color_.r, face_color_.g, face_color_.b, face_opacity_);
    }
    else
    if (face_opacity_ > 0.99) {
      glColor3f(face_color_.r, face_color_.g, face_color_.b);
    }

    if ((face_opacity_<0.01 && pick==true) || face_opacity_>=0.01) {
      if (state_ == MTK_ACTIVE)
        glLoadName(MTK_BALL_S);
      glPushAttrib(GL_CURRENT_BIT);
      GLUquadricObj *sphere=gluNewQuadric();
      gluSphere(sphere, 1.0, slices_, stacks_);
      glPopAttrib();
    }

    if (face_opacity_>0.01 && face_opacity_<0.99) {
      glDepthMask(GL_TRUE);
      glDisable(GL_BLEND);
    }
  }

  // restore matrix
  glPopMatrix();

  if (lighting)
    glEnable(GL_LIGHTING);

  if (state_ == MTK_ACTIVE) {
    glPopName();
    glPopName();
    glPopName();
  }
}


bool mtkTrackballGL::interaction (mtkDraggerPart part, int x, int y,
                                 mtkEvent type)
{
  bool ok, ret=false;
  gmPoint3 c, p;
  gmVector3 v;
  double r;

  getSphere(&c, &r);
  switch (type) {
    case MTK_PICK:
      if (part == MTK_BALL_V) {
        p=getArrow();
        if (callback_)
          ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                           MTK_PICK, data_);
        vconst_.start(x, y, p);
      }
      else if (part == MTK_BALL_E) {
        p=getArrow();
        econst_.edgeIntersect(x, y, gmLerp(gmInv(1.0+arrow_), c, p), p, &p);
        if (callback_)
          ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                           MTK_PICK, data_);
        econst_.start(x, y, p);
      }
      else if (part>=MTK_BALL_R_0 && part<=MTK_BALL_R_2) {
        rconst_.sphereIntersect(x, y, c, r, &p);
        if (callback_)
          ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                           MTK_PICK, data_);
        rconst_.start(x, y, p);
      }
      else if (part == MTK_BALL_S) {
        sconst_.sphereIntersect(x, y, c, r, &p);
        if (callback_)
          ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                           MTK_PICK, data_);
        sconst_.start(x, y, p);
      }
      break;

    case MTK_MOVE:
      ok=false;
      if (part == MTK_BALL_V) {
        ok=vconst_.next(x, y, &p, &v);
        if (ok && callback_)
          ret=(*callback_)(id_, part, p, v, MTK_MOVE, data_);
        if (vconst_.getPhase() == MTK_SETUP_PHASE)
          vconst_.start(x, y, p+v);
      }
      else if (part == MTK_BALL_E) {
	ok=econst_.next(x, y, &p, &v);
        if (ok && callback_)
          ret=(*callback_)(id_, part, p, v, MTK_MOVE, data_);
        if (econst_.getPhase() == MTK_SETUP_PHASE)
          econst_.start(x, y, p+v);
      }
      else if (part>=MTK_BALL_R_0 && part<=MTK_BALL_R_2) {
        ok=rconst_.next(x, y, &p, &v);
        if (ok && callback_)
          ret=(*callback_)(id_, part, p, v, MTK_MOVE, data_);
        if (rconst_.getPhase() == MTK_SETUP_PHASE)
          rconst_.start(x, y, p+v);
      }
      else if (part == MTK_BALL_S) {
        ok=sconst_.next(x, y, &p, &v);
        if (ok && callback_)
          ret=(*callback_)(id_, part, p, v, MTK_MOVE, data_);
        if (sconst_.getPhase() == MTK_SETUP_PHASE)
          sconst_.start(x, y, p+v);
      }
      break;

    case MTK_RELEASE:
      if (part == MTK_BALL_V)
        vconst_.end(&p);
      else if (part == MTK_BALL_E)
        econst_.end(&p);
      else if (part>=MTK_BALL_R_0 && part<=MTK_BALL_R_2)
        rconst_.end(&p);
      else if (part == MTK_BALL_S)
        sconst_.end(&p);
      if (callback_)
        ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                         MTK_RELEASE, data_);
      break;
  }
  return ret;
}


/*********************************************************************
 mtkJackGL
**********************************************************************/

mtkJackGL::mtkJackGL () : mtkJack()
{
}


bool mtkJackGL::setConstraint (mtkDraggerPart part, mtkConstraintGL cstr)
{
  if (part>=MTK_JACK_V_0 && part<=MTK_JACK_V_6)
    vconst_=cstr;
  else if (part>=MTK_JACK_E_01 && part<=MTK_JACK_E_06)
    econst_=cstr;
  else
    return false;
  return true;
}


mtkDragger *mtkJackGL::clone () const
{
  mtkJackGL *ptr=new mtkJackGL();
  (*ptr)=*this;
  return ptr;
}


void mtkJackGL::draw (bool pick)
{
  // unused parameter
  pick=pick;

  if (state_ == MTK_HIDDEN)
    return;

  if (state_ == MTK_ACTIVE) {
    glPushName(MTK_DRAGGER);
    glPushName(id_);
    glPushName(0);
  }

  GLboolean lighting=glIsEnabled(GL_LIGHTING);
  if (lighting)
    glDisable(GL_LIGHTING);

  // draw vertices
  glColor3f(vertex_color_.r, vertex_color_.g, vertex_color_.b);
  glPointSize(vertex_size_);

  if (state_ == MTK_ACTIVE) {
    for (int i=0; i<7; i++)
      if (vshow_[i]) {
        glLoadName(MTK_JACK_V_0+i);
        glBegin(GL_POINTS);
        glVertex3f(v_[i][0], v_[i][1], v_[i][2]);
        glEnd();
      }
  }
  else {
    glBegin(GL_POINTS);
    for (int i=0; i<7; i++) {
      if (vshow_[i])
        glVertex3f(v_[i][0], v_[i][1], v_[i][2]);
    }
    glEnd();
  }

  // draw edges
  glColor3f(edge_color_.r, edge_color_.g, edge_color_.b);
  glLineWidth(edge_thickness_);

  if (state_ == MTK_ACTIVE) {
    if (eshow_[0]) {
      glLoadName(MTK_JACK_E_01);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glEnd();
    }
    if (eshow_[1]) {
      glLoadName(MTK_JACK_E_02);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glEnd();
    }
    if (eshow_[2]) {
      glLoadName(MTK_JACK_E_03);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glEnd();
    }
    if (eshow_[3]) {
      glLoadName(MTK_JACK_E_04);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glEnd();
    }
    if (eshow_[4]) {
      glLoadName(MTK_JACK_E_05);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glEnd();
    }
    if (eshow_[5]) {
      glLoadName(MTK_JACK_E_06);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glEnd();
    }
  }
  else {
    glBegin(GL_LINES);
    if (eshow_[0]) {
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
    }
    if (eshow_[1]) {
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
    }
    if (eshow_[2]) {
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
    }
    if (eshow_[3]) {
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
    }
    if (eshow_[4]) {
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
    }
    if (eshow_[5]) {
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
    }
    glEnd();
  }

  if (lighting)
    glEnable(GL_LIGHTING);

  if (state_ == MTK_ACTIVE) {
    glPopName();
    glPopName();
    glPopName();
  }
}


bool mtkJackGL::interaction (mtkDraggerPart part, int x, int y, mtkEvent type)

{
  bool ok, ret=false;
  gmPoint3 p;
  gmVector3 v;
  switch (type) {
    case MTK_PICK:
      if (part>=MTK_JACK_V_0 && part<=MTK_JACK_V_6) {
        // handle vertices
        p=v_[part-MTK_JACK_V_0];
        if (callback_)
          ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                           MTK_PICK, data_);
        vconst_.start(x, y, p);
      }
      else if (part>=MTK_JACK_E_01 && part<=MTK_JACK_E_06) {
        // handle edges
        gmPoint3 o1, o2;

        // get edge vertices
        switch (part) {
          case MTK_JACK_E_01:
            o1=v_[0];
            o2=v_[1];
            break;
          case MTK_JACK_E_02:
            o1=v_[0];
            o2=v_[2];
            break;
          case MTK_JACK_E_03:
            o1=v_[0];
            o2=v_[3];
            break;
          case MTK_JACK_E_04:
            o1=v_[0];
            o2=v_[4];
            break;
          case MTK_JACK_E_05:
            o1=v_[0];
            o2=v_[5];
            break;
          case MTK_JACK_E_06:
            o1=v_[0];
            o2=v_[6];
            break;
        }
        econst_.edgeIntersect(x, y, o1, o2, &p);
        if (callback_)
          ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                           MTK_PICK, data_);
        econst_.start(x, y, p);
      }
      break;

    case MTK_MOVE:
      ok=false;
      if (part>=MTK_JACK_V_0 && part<=MTK_JACK_V_6) {
        ok=vconst_.next(x, y, &p, &v);
        if (ok && callback_)
          ret=(*callback_)(id_, part, p, v, MTK_MOVE, data_);
        if (vconst_.getPhase() == MTK_SETUP_PHASE)
          vconst_.start(x, y, p+v);
      }
      else if (part>=MTK_JACK_E_01 && part<=MTK_JACK_E_06) {
        ok=econst_.next(x, y, &p, &v);
        if (ok && callback_)
          ret=(*callback_)(id_, part, p, v, MTK_MOVE, data_);
        if (econst_.getPhase() == MTK_SETUP_PHASE)
          econst_.start(x, y, p+v);
      }
      break;

    case MTK_RELEASE:
      if (part>=MTK_JACK_V_0 && part<=MTK_JACK_V_6)
        vconst_.end(&p);
      else if (part>=MTK_JACK_E_01 && part<=MTK_JACK_E_06)
        econst_.end(&p);
      if (callback_)
        ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                         MTK_RELEASE, data_);
      break;
  }
  return ret;
}


/*********************************************************************
 mtkCursor3DGL
**********************************************************************/

mtkCursor3DGL::mtkCursor3DGL () : mtkCursor3D()
{
}


bool mtkCursor3DGL::setConstraint (mtkDraggerPart part, mtkConstraintGL cstr)
{
  if (part>=MTK_CURSOR3D_V_0 && part<=MTK_CURSOR3D_V_6)
    vconst_=cstr;
  else if (part>=MTK_CURSOR3D_E_01 && part<=MTK_CURSOR3D_E_06)
    econst_=cstr;
  else
    return false;
  return true;
}


mtkDragger *mtkCursor3DGL::clone () const
{
  mtkCursor3DGL *ptr=new mtkCursor3DGL();
  (*ptr)=*this;
  return ptr;
}


void mtkCursor3DGL::draw (bool pick)
{
  // unused parameter
  pick=pick;

  if (state_ == MTK_HIDDEN)
    return;

  if (state_ == MTK_ACTIVE) {
    glPushName(MTK_DRAGGER);
    glPushName(id_);
    glPushName(0);
  }

  GLboolean lighting=glIsEnabled(GL_LIGHTING);
  if (lighting)
    glDisable(GL_LIGHTING);

  // draw vertices
  glColor3f(vertex_color_.r, vertex_color_.g, vertex_color_.b);
  glPointSize(vertex_size_);

  if (state_ == MTK_ACTIVE) {
    for (int i=0; i<7; i++)
      if (vshow_[i]) {
        glLoadName(MTK_CURSOR3D_V_0+i);
        glBegin(GL_POINTS);
        glVertex3f(v_[i][0], v_[i][1], v_[i][2]);
        glEnd();
      }
  }
  else {
    glBegin(GL_POINTS);
    for (int i=0; i<7; i++) {
      if (vshow_[i])
        glVertex3f(v_[i][0], v_[i][1], v_[i][2]);
    }
    glEnd();
  }

  // draw edges
  glColor3f(edge_color_.r, edge_color_.g, edge_color_.b);
  glLineWidth(edge_thickness_);

  if (state_ == MTK_ACTIVE) {
    if (eshow_[0]) {
      glLoadName(MTK_CURSOR3D_E_01);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
      glEnd();
    }
    if (eshow_[1]) {
      glLoadName(MTK_CURSOR3D_E_02);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
      glEnd();
    }
    if (eshow_[2]) {
      glLoadName(MTK_CURSOR3D_E_03);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
      glEnd();
    }
    if (eshow_[3]) {
      glLoadName(MTK_CURSOR3D_E_04);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
      glEnd();
    }
    if (eshow_[4]) {
      glLoadName(MTK_CURSOR3D_E_05);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
      glEnd();
    }
    if (eshow_[5]) {
      glLoadName(MTK_CURSOR3D_E_06);
      glBegin(GL_LINES);
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
      glEnd();
    }
  }
  else {
    glBegin(GL_LINES);
    if (eshow_[0]) {
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[1][0], v_[1][1], v_[1][2]);
    }
    if (eshow_[1]) {
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[2][0], v_[2][1], v_[2][2]);
    }
    if (eshow_[2]) {
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[3][0], v_[3][1], v_[3][2]);
    }
    if (eshow_[3]) {
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[4][0], v_[4][1], v_[4][2]);
    }
    if (eshow_[4]) {
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[5][0], v_[5][1], v_[5][2]);
    }
    if (eshow_[5]) {
      glVertex3f(v_[0][0], v_[0][1], v_[0][2]);
      glVertex3f(v_[6][0], v_[6][1], v_[6][2]);
    }
    glEnd();
  }

  if (lighting)
    glEnable(GL_LIGHTING);

  if (state_ == MTK_ACTIVE) {
    glPopName();
    glPopName();
    glPopName();
  }
}


bool mtkCursor3DGL::interaction (mtkDraggerPart part, int x, int y, mtkEvent type)

{
  bool ok, ret=false;
  gmPoint3 p;
  gmVector3 v;
  switch (type) {
    case MTK_PICK:
      if (part>=MTK_CURSOR3D_V_0 && part<=MTK_CURSOR3D_V_6) {
        // handle vertices
        p=v_[part-MTK_CURSOR3D_V_0];
        if (callback_)
          ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                           MTK_PICK, data_);
        vconst_.start(x, y, p);
      }
      else if (part>=MTK_CURSOR3D_E_01 && part<=MTK_CURSOR3D_E_06) {
        // handle edges
        gmPoint3 o1, o2;

        // get edge vertices
        switch (part) {
          case MTK_CURSOR3D_E_01:
            o1=v_[0];
            o2=v_[1];
            break;
          case MTK_CURSOR3D_E_02:
            o1=v_[0];
            o2=v_[2];
            break;
          case MTK_CURSOR3D_E_03:
            o1=v_[0];
            o2=v_[3];
            break;
          case MTK_CURSOR3D_E_04:
            o1=v_[0];
            o2=v_[4];
            break;
          case MTK_CURSOR3D_E_05:
            o1=v_[0];
            o2=v_[5];
            break;
          case MTK_CURSOR3D_E_06:
            o1=v_[0];
            o2=v_[6];
            break;
        }
        econst_.edgeIntersect(x, y, o1, o2, &p);
        if (callback_)
          ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                           MTK_PICK, data_);
        econst_.start(x, y, p);
      }
      break;

    case MTK_MOVE:
      ok=false;
      if (part>=MTK_CURSOR3D_V_0 && part<=MTK_CURSOR3D_V_6) {
        ok=vconst_.next(x, y, &p, &v);
        if (ok && callback_)
          ret=(*callback_)(id_, part, p, v, MTK_MOVE, data_);
        if (vconst_.getPhase() == MTK_SETUP_PHASE)
          vconst_.start(x, y, p+v);
      }
      else if (part>=MTK_CURSOR3D_E_01 && part<=MTK_CURSOR3D_E_06) {
        ok=econst_.next(x, y, &p, &v);
        if (ok && callback_)
          ret=(*callback_)(id_, part, p, v, MTK_MOVE, data_);
        if (econst_.getPhase() == MTK_SETUP_PHASE)
          econst_.start(x, y, p+v);
      }
      break;

    case MTK_RELEASE:
      if (part>=MTK_CURSOR3D_V_0 && part<=MTK_CURSOR3D_V_6)
        vconst_.end(&p);
      else if (part>=MTK_CURSOR3D_E_01 && part<=MTK_CURSOR3D_E_06)
        econst_.end(&p);
      if (callback_)
        ret=(*callback_)(id_, part, p, gmVector3(0.0, 0.0, 0.0),
                         MTK_RELEASE, data_);
      break;
  }
  return ret;
}
