// -*- mode: C++; -*-

/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef MTK_DRAGGER_H
#define MTK_DRAGGER_H

#include "gm.h"
#include "mtk_general.h"
#include "mtk_constraint.h"


enum mtkDraggerType {MTK_BOUNDING_BOX, MTK_JACK, MTK_CURSOR3D, MTK_TRACKBALL};

typedef unsigned int mtkDraggerPart;

// Bounding Box part ids
#define MTK_BOX_V_0      0x01
#define MTK_BOX_V_1      0x02
#define MTK_BOX_V_2      0x03
#define MTK_BOX_V_3      0x04
#define MTK_BOX_V_4      0x05
#define MTK_BOX_V_5      0x06
#define MTK_BOX_V_6      0x07
#define MTK_BOX_V_7      0x08
#define MTK_BOX_E_01     0x11
#define MTK_BOX_E_02     0x12
#define MTK_BOX_E_04     0x13
#define MTK_BOX_E_31     0x14
#define MTK_BOX_E_32     0x15
#define MTK_BOX_E_37     0x16
#define MTK_BOX_E_51     0x17
#define MTK_BOX_E_54     0x18
#define MTK_BOX_E_57     0x19
#define MTK_BOX_E_62     0x1a
#define MTK_BOX_E_64     0x1b
#define MTK_BOX_E_67     0x1c
#define MTK_BOX_F_0132   0x21
#define MTK_BOX_F_1573   0x22
#define MTK_BOX_F_4675   0x23
#define MTK_BOX_F_0264   0x24
#define MTK_BOX_F_2376   0x25
#define MTK_BOX_F_0451   0x26

// Jack part ids
#define MTK_JACK_V_0     0x31
#define MTK_JACK_V_1     0x32
#define MTK_JACK_V_2     0x33
#define MTK_JACK_V_3     0x34
#define MTK_JACK_V_4     0x35
#define MTK_JACK_V_5     0x36
#define MTK_JACK_V_6     0x37
#define MTK_JACK_E_01    0x41
#define MTK_JACK_E_02    0x42
#define MTK_JACK_E_03    0x43
#define MTK_JACK_E_04    0x44
#define MTK_JACK_E_05    0x45
#define MTK_JACK_E_06    0x46

// Trackball part ids
#define MTK_BALL_V       0x51
#define MTK_BALL_E       0x61
#define MTK_BALL_R_0     0x71
#define MTK_BALL_R_1     0x72
#define MTK_BALL_R_2     0x73
#define MTK_BALL_S       0x81

// Cursor3D part ids
#define MTK_CURSOR3D_V_0     0x91
#define MTK_CURSOR3D_V_1     0x92
#define MTK_CURSOR3D_V_2     0x93
#define MTK_CURSOR3D_V_3     0x94
#define MTK_CURSOR3D_V_4     0x95
#define MTK_CURSOR3D_V_5     0x96
#define MTK_CURSOR3D_V_6     0x97
#define MTK_CURSOR3D_E_01    0x9a
#define MTK_CURSOR3D_E_02    0x9b
#define MTK_CURSOR3D_E_03    0x9c
#define MTK_CURSOR3D_E_04    0x9d
#define MTK_CURSOR3D_E_05    0x9e
#define MTK_CURSOR3D_E_06    0x9f

typedef unsigned int mtkDraggerId;

typedef bool mtkBoxCallback(mtkDraggerId id, mtkDraggerPart part,
                            gmPoint3 p, gmVector3 v,
                            mtkEvent type, void *data);
typedef bool mtkTrackballCallback(mtkDraggerId id, mtkDraggerPart part,
                                  gmPoint3 p, gmVector3 v,
                                  mtkEvent type, void *data);
typedef bool mtkJackCallback(mtkDraggerId id, mtkDraggerPart part,
                             gmPoint3 p, gmVector3 v,
                             mtkEvent type, void *data);

typedef bool mtkCursor3DCallback(mtkDraggerId id, mtkDraggerPart part,
                             gmPoint3 p, gmVector3 v,
                             mtkEvent type, void *data);

/*********************************************************************
 Base class for draggers
**********************************************************************/
class mtkDragger {
public:
  mtkDragger ();

  mtkDraggerId getId ();
  void setState (mtkState state);
  mtkState getState ();
  void setSize (double size);
  double getSize();

  virtual mtkDraggerType getType () = 0;
  virtual mtkDragger *clone () const = 0;
  virtual void draw (bool pick) = 0;
  virtual bool interaction (mtkDraggerPart part, int x, int y,
                            mtkEvent type) = 0;

protected:
  static mtkDraggerId counter_;

  mtkDraggerId id_;
  mtkState state_;
  double size_;
};


/*********************************************************************
 Class BoundingBox
**********************************************************************/
class mtkBoundingBox: public mtkDragger {
public:
  mtkBoundingBox ();

  mtkDraggerType getType ();

  void setGeometry (gmPoint3 v0, gmVector3 e01, gmVector3 e02, gmVector3 e04);
  gmPoint3 getVertex (mtkDraggerPart part);
  gmPoint3 getCenter ();

  void setVisibility (mtkDraggerPart part, bool b);
  bool getVisibility (mtkDraggerPart part);

  void setVertexSize (unsigned int size);
  void setVertexColor (float r, float g, float b);

  /// Set edge attributes
  void setEdgeThickness (unsigned int thickness);
  void setEdgeColor (float r, float g, float b);

  /// Set face attributes
  void setFaceColor (float r, float g, float b, float a);

  /// Set dragger behamtkour passing a transform matrix to each part
  void applyMatrix (gmMatrix4 m);

  /// This callback function returns the result of 3D interaction
  void setCallback (mtkBoxCallback *callback, void *data);


protected:
  // Vertices position in 3D space
  gmPoint3 v_[8];
  bool vshow_, eshow_, fshow_;

  // Vertex information
  unsigned int vertex_size_;
  mtkColor vertex_color_;

  // Edge information
  unsigned int edge_thickness_;
  mtkColor edge_color_;

  // Vertex information
  mtkColor face_color_;
  float face_opacity_;

  mtkBoxCallback *callback_;
  void *data_;
};


/*********************************************************************
 Class Trackball
**********************************************************************/
class mtkTrackball: public mtkDragger {
public:
  mtkTrackball ();

  mtkDraggerType getType ();

  void setGeometry (gmPoint3 center, double radius, double arrow);
  void setResolution (int slices, int stacks);
  void getSphere (gmPoint3 *center, double *radius);
  gmPoint3 getArrow ();
  gmVector3 getRing (mtkDraggerPart part);

  void setVisibility (mtkDraggerPart part, bool b);
  bool getVisibility (mtkDraggerPart part);

  void setVertexSize (unsigned int size);
  void setVertexColor (float r, float g, float b);

  void setEdgeThickness (unsigned int thickness);
  void setEdgeColor (float r, float g, float b);

  void setFaceColor (float r, float g, float b, float a);

  /// Set dragger behamtkour passing a transform matrix to each part
  void applyMatrix (gmMatrix4 m);

  /// This callback function returns the result of 3D interaction
  void setCallback (mtkTrackballCallback *callback, void *data);



protected:
  // trackball data
  gmMatrix4 matrix_;
  double arrow_;
  int slices_, stacks_;
  bool vshow_, eshow_, rshow_[3], sshow_;

  // Vertex information
  unsigned int vertex_size_;
  mtkColor vertex_color_;

  // Edge information
  unsigned int edge_thickness_;
  mtkColor edge_color_;

  // Face information
  mtkColor face_color_;
  float face_opacity_;

  mtkTrackballCallback *callback_;
  void *data_;
};


/*********************************************************************
 Class Jack
**********************************************************************/
class mtkJack: public mtkDragger {
public:
  mtkJack();

  mtkDraggerType getType ();

  /// Set and get Jack methods
  void setGeometry (gmPoint3 v0, gmVector3 e01, gmVector3 e03, gmVector3 e05);
  gmPoint3 getVertex (mtkDraggerPart part);

  void setVisibility (mtkDraggerPart part, bool b);
  bool getVisibility (mtkDraggerPart part);

  /// Set vertex attributes
  void setVertexSize (unsigned int size);
  void setVertexColor (float r, float g, float b);

  /// Set edge attributes
  void setEdgeThickness (unsigned int thickness);
  void setEdgeColor (float r, float g, float b);

  /// Set dragger behamtkour passing a transform matrix to each part
  void applyMatrix (gmMatrix4 m);

  /// This callback function returns the result of 3D interaction
  void setCallback (mtkJackCallback *callback, void *data);


protected:
  // Vertices position in 3D space
  gmPoint3 v_[7];
  bool vshow_[7], eshow_[6];

  // Vertex information
  unsigned int vertex_size_;
  mtkColor vertex_color_;

  // Edge information
  unsigned int edge_thickness_;
  mtkColor edge_color_;

  mtkJackCallback *callback_;
  void *data_;
};


/*********************************************************************
 Class Cursor3D
**********************************************************************/
class mtkCursor3D: public mtkDragger {
public:
  mtkCursor3D();

  mtkDraggerType getType ();

  /// Set and get Cursor3D methods
  void setGeometry (gmPoint3 v0, gmVector3 e01, gmVector3 e03, gmVector3 e05);
  gmPoint3 getVertex (mtkDraggerPart part);

  void setVisibility (mtkDraggerPart part, bool b);
  bool getVisibility (mtkDraggerPart part);

  /// Set vertex attributes
  void setVertexSize (unsigned int size);
  void setVertexColor (float r, float g, float b);

  /// Set edge attributes
  void setEdgeThickness (unsigned int thickness);
  void setEdgeColor (float r, float g, float b);

  /// Set dragger behamtkour passing a transform matrix to each part
  void applyMatrix (gmMatrix4 m);

  /// This callback function returns the result of 3D interaction
  void setCallback (mtkCursor3DCallback *callback, void *data);


protected:
  // Vertices position in 3D space
  gmPoint3 v_[7];
  bool vshow_[7], eshow_[6];

  // Vertex information
  unsigned int vertex_size_;
  mtkColor vertex_color_;

  // Edge information
  unsigned int edge_thickness_;
  mtkColor edge_color_;

  mtkCursor3DCallback *callback_;
  void *data_;
};

/*********************************************************************
 Class BoundingBoxGL
**********************************************************************/
class mtkBoundingBoxGL: public mtkBoundingBox {
public:
  mtkBoundingBoxGL ();

  bool setConstraint (mtkDraggerPart part, mtkConstraintGL cstr);

  mtkDragger *clone () const;

  /// Draws the bounding box
  void draw (bool pick);

  /// Pass a 2D vector of pointer movement, which is mapped in 3D movements
  bool interaction (mtkDraggerPart part, int x, int y, mtkEvent type);

protected:
  mtkConstraintGL vconst_, econst_, fconst_;
};


/*********************************************************************
 Class TrackballGL
**********************************************************************/
class mtkTrackballGL: public mtkTrackball {
public:
  mtkTrackballGL ();

  bool setConstraint (mtkDraggerPart part, mtkConstraintGL cstr);

  mtkDragger *clone () const;

  /// Draws the trackball
  void draw (bool pick);

  /// Pass a 2D vector of pointer movement, which is mapped in 3D movements
  bool interaction (mtkDraggerPart part, int x, int y, mtkEvent type);

protected:
  mtkConstraintGL vconst_, econst_, rconst_, sconst_;
};


/*********************************************************************
 Class JackGL
**********************************************************************/
class mtkJackGL: public mtkJack {
public:
  mtkJackGL ();

  bool setConstraint (mtkDraggerPart part, mtkConstraintGL cstr);

  mtkDragger *clone () const;

  /// Draws the jack
  void draw (bool pick);

  /// Pass a 2D vector of pointer movement, which is mapped in 3D movements
  bool interaction (mtkDraggerPart part, int x, int y, mtkEvent type);

protected:
  mtkConstraintGL vconst_, econst_;
};

/*********************************************************************
 Class Cursor3DGL
**********************************************************************/
class mtkCursor3DGL: public mtkCursor3D {
public:
  mtkCursor3DGL ();

  bool setConstraint (mtkDraggerPart part, mtkConstraintGL cstr);

  mtkDragger *clone () const;

  /// Draws the cursor3D
  void draw (bool pick);

  /// Pass a 2D vector of pointer movement, which is mapped in 3D movements
  bool interaction (mtkDraggerPart part, int x, int y, mtkEvent type);

protected:
  mtkConstraintGL vconst_, econst_;
};

#endif // MTK_DRAGGER_H
