/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <GL/glu.h>

#include "mtk_guide.h"


/*********************************************************************
 mtkGuide
**********************************************************************/

mtkId mtkGuide::counter_=1;

mtkGuide::mtkGuide ()
{
  id_=counter_++;
  state_=MTK_ACTIVE;
}


/// get Guide id
mtkGuideId mtkGuide::getId ()
{
  return id_;
}


/// set Guide state
void mtkGuide::setState (mtkState state)
{
  state_=state;
}


/// get Guide state
mtkState mtkGuide::getState ()
{
  return state_;
}


/*********************************************************************
 mtkAxis
**********************************************************************/
mtkAxis::mtkAxis ()
{
  origin_.assign( -2.0,  1.0,  0.0);
  axis_[0].assign( 0.5,  0.0,  0.0);
  axis_[1].assign( 0.0,  0.5,  0.0);
  axis_[2].assign( 0.0,  0.0,  0.5);

  style_=MTK_XYZ_FULL;
  axis_thickness_=1;
  axis_color_.r=axis_color_.g=axis_color_.b=1.0;
}


mtkGuideType mtkAxis::getType ()
{
  return MTK_AXIS;
}


void mtkAxis::setOrigin(gmPoint3 origin)
{
  origin_=origin;
}


gmPoint3 mtkAxis::getOrigin()
{
  return origin_;
}


void mtkAxis::setAxis (gmVector3 axis[3])
{
  int i;

  for(i=0; i<3; i++)
    axis_[i]=axis[i];
}


void mtkAxis::getAxis (gmVector3 axis[3])
{
  int i;

  for(i=0; i<3; i++)
    axis[i]=axis_[i];
}


void mtkAxis::setAxisStyle (mtkAxisStyle style)
{
  style_=style;
}


mtkAxisStyle mtkAxis::getAxisStyle ()
{
  return style_;
}


void mtkAxis::setThickness (unsigned int thickness)
{
  axis_thickness_=thickness;
}


void mtkAxis::setColor (float r, float g, float b)
{
  axis_color_.r=r;
  axis_color_.g=g;
  axis_color_.b=b;
}


/*********************************************************************
 mtkGrid
**********************************************************************/

mtkGrid::mtkGrid ()
{
  origin_.assign(0.0, -1.2, 2.0);
  dir_[0].assign(0.3, 0.0, -0.3);
  dir_[1].assign(-0.3, 0.0, -0.3 );

  num_lines_dir1_=15;
  num_lines_dir2_=15;

  edge_thickness_=1;
  edge_color_.r=edge_color_.g=edge_color_.b=1.0;
  face_color_.r=face_color_.g=face_color_.b=1.0;
  face_opacity_=0.5;
  fill_face_=true;
}


mtkGuideType mtkGrid::getType ()
{
  return MTK_GRID;
}


void mtkGrid::setOrigin(gmPoint3 origin)
{
  origin_=origin;
}


gmPoint3 mtkGrid::getOrigin()
{
  return origin_;
}


void mtkGrid::setDir (gmVector3 dir[2])
{
  int i;

  for(i=0; i<2; i++)
    dir_[i]=dir[i];
}


void mtkGrid::getDir (gmVector3 dir[2])
{
  int i;

  for(i=0; i<2; i++)
    dir[i]=dir_[i];
}


void mtkGrid::setNumLines (int nLines1, int nLines2)
{
  num_lines_dir1_=nLines1;
  num_lines_dir2_=nLines2;
}


void mtkGrid::getNumLines (int *nLines1, int *nLines2)
{
  *nLines1=num_lines_dir1_;
  *nLines2=num_lines_dir2_;
}


void mtkGrid::setEdgeThickness (unsigned int thickness)
{
  edge_thickness_=thickness;
}


void mtkGrid::setEdgeColor (float r, float g, float b)
{
  edge_color_.r=r;
  edge_color_.g=g;
  edge_color_.b=b;
}


void mtkGrid::setFaceColor (float r, float g, float b, float a)
{
  face_color_.r=r;
  face_color_.g=g;
  face_color_.b=b;
  face_opacity_=a;
}


void mtkGrid::setFillFace(bool fill_face)
{
  fill_face_=fill_face;
}


/*********************************************************************
 mtkAxisGL
**********************************************************************/

mtkAxisGL::mtkAxisGL () : mtkAxis()
{
}


mtkGuide *mtkAxisGL::clone () const
{
  mtkAxisGL *ptr=new mtkAxisGL();
  (*ptr)=*this;
  return ptr;
}


void mtkAxisGL::draw ()
{
  gmPoint3 ax1[2], ax2[2], ax3[2];

  if (state_ == MTK_INACTIVE)
    return;

  // make the diffuse material track the current glColor
  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);

  // Draw Axis edges
  glColor3f(axis_color_.r, axis_color_.g, axis_color_.b);
  glLineWidth(axis_thickness_);

  if (state_ == MTK_ACTIVE) {
    switch(style_) {
    case MTK_XYZ_FULL:
      ax1[0].assign(origin_[0]-axis_[0][0], origin_[1]-axis_[0][1],
		 origin_[2]-axis_[0][2]);
      ax2[0].assign(origin_[0]-axis_[1][0], origin_[1]-axis_[1][1],
		 origin_[2]-axis_[1][2]);
      ax3[0].assign(origin_[0]-axis_[2][0], origin_[1]-axis_[2][1],
                 origin_[2]-axis_[2][2]);
      ax1[1].assign(origin_[0]+axis_[0][0], origin_[1]+axis_[0][1],
		 origin_[2]+axis_[0][2]);
      ax2[1].assign(origin_[0]+axis_[1][0], origin_[1]+axis_[1][1],
		 origin_[2]+axis_[1][2]);
      ax3[1].assign(origin_[0]+axis_[2][0], origin_[1]+axis_[2][1],
		 origin_[2]+axis_[2][2]);
    break;
    case MTK_XY_FULL:
      ax1[0].assign(origin_[0]-axis_[0][0], origin_[1]-axis_[0][1],
		 origin_[2]);
      ax2[0].assign(origin_[0]-axis_[1][0], origin_[1]-axis_[1][1],
		 origin_[2]);
      ax3[0].assign(origin_[0]-axis_[2][0], origin_[1]-axis_[2][1],
                 origin_[2]);
      ax1[1].assign(origin_[0]+axis_[0][0], origin_[1]+axis_[0][1],
		 origin_[2]);
      ax2[1].assign(origin_[0]+axis_[1][0], origin_[1]+axis_[1][1],
		 origin_[2]);
      ax3[1].assign(origin_[0]+axis_[2][0], origin_[1]+axis_[2][1],
		 origin_[2]);
    break;
    case MTK_XZ_FULL:
      ax1[0].assign(origin_[0]-axis_[0][0], origin_[1],
		 origin_[2]-axis_[0][2]);
      ax2[0].assign(origin_[0]-axis_[1][0], origin_[1],
		 origin_[2]-axis_[1][2]);
      ax3[0].assign(origin_[0]-axis_[2][0], origin_[1],
                 origin_[2]-axis_[2][2]);
      ax1[1].assign(origin_[0]+axis_[0][0], origin_[1],
		 origin_[2]+axis_[0][2]);
      ax2[1].assign(origin_[0]+axis_[1][0], origin_[1],
		 origin_[2]+axis_[1][2]);
      ax3[1].assign(origin_[0]+axis_[2][0], origin_[1],
		 origin_[2]+axis_[2][2]);
    break;
    case MTK_YZ_FULL:
      ax1[0].assign(origin_[0], origin_[1]-axis_[0][1],
		 origin_[2]-axis_[0][2]);
      ax2[0].assign(origin_[0], origin_[1]-axis_[1][1],
		 origin_[2]-axis_[1][2]);
      ax3[0].assign(origin_[0], origin_[1]-axis_[2][1],
                 origin_[2]-axis_[2][2]);
      ax1[1].assign(origin_[0], origin_[1]+axis_[0][1],
		 origin_[2]+axis_[0][2]);
      ax2[1].assign(origin_[0], origin_[1]+axis_[1][1],
		 origin_[2]+axis_[1][2]);
      ax3[1].assign(origin_[0], origin_[1]+axis_[2][1],
		 origin_[2]+axis_[2][2]);
    break;
    case MTK_XYZ_HALF:
      ax1[0].assign(origin_[0], origin_[1], origin_[2]);
      ax3[0]=ax2[0]=ax1[0];
      ax1[1].assign(origin_[0]+axis_[0][0], origin_[1]+axis_[0][1],
		 origin_[2]+axis_[0][2]);
      ax2[1].assign(origin_[0]+axis_[1][0], origin_[1]+axis_[1][1],
		 origin_[2]+axis_[1][2]);
      ax3[1].assign(origin_[0]+axis_[2][0], origin_[1]+axis_[2][1],
		 origin_[2]+axis_[2][2]);
    break;
    case MTK_XY_HALF:
      ax1[0].assign(origin_[0], origin_[1], origin_[2]);
      ax3[0]=ax2[0]=ax1[0];
      ax1[1].assign(origin_[0]+axis_[0][0], origin_[1]+axis_[0][1],
		 origin_[2]);
      ax2[1].assign(origin_[0]+axis_[1][0], origin_[1]+axis_[1][1],
		 origin_[2]);
      ax3[1].assign(origin_[0]+axis_[2][0], origin_[1]+axis_[2][1],
		 origin_[2]);
    break;
    case MTK_XZ_HALF:
      ax1[0].assign(origin_[0], origin_[1], origin_[2]);
      ax3[0]=ax2[0]=ax1[0];
      ax1[1].assign(origin_[0]+axis_[0][0], origin_[1],
		 origin_[2]+axis_[0][2]);
      ax2[1].assign(origin_[0]+axis_[1][0], origin_[1],
		 origin_[2]+axis_[1][2]);
      ax3[1].assign(origin_[0]+axis_[2][0], origin_[1],
		 origin_[2]+axis_[2][2]);
    break;
    case MTK_YZ_HALF:
      ax1[0].assign(origin_[0], origin_[1], origin_[2]);
      ax3[0]=ax2[0]=ax1[0];
      ax1[1].assign(origin_[0], origin_[1]+axis_[0][1],
		 origin_[2]+axis_[0][2]);
      ax2[1].assign(origin_[0], origin_[1]+axis_[1][1],
		 origin_[2]+axis_[1][2]);
      ax3[1].assign(origin_[0], origin_[1]+axis_[2][1],
		 origin_[2]+axis_[2][2]);
    break;
    case MTK_X:
    break;
    case MTK_Y:
    break;
    case MTK_Z:
    break;
    }
    glBegin(GL_LINES);
    glVertex3f(ax1[0][0], ax1[0][1], ax1[0][2]);
    glVertex3f(ax1[1][0], ax1[1][1], ax1[1][2]);
    glVertex3f(ax2[0][0], ax2[0][1], ax2[0][2]);
    glVertex3f(ax2[1][0], ax2[1][1], ax2[1][2]);
    glVertex3f(ax3[0][0], ax3[0][1], ax3[0][2]);
    glVertex3f(ax3[1][0], ax3[1][1], ax3[1][2]);
    glEnd();
  }
  glDisable(GL_COLOR_MATERIAL);
}


/*********************************************************************
 mtkGridGL
**********************************************************************/

mtkGridGL::mtkGridGL () : mtkGrid()
{
}


mtkGuide *mtkGridGL::clone () const
{
  mtkGridGL *ptr=new mtkGridGL();
  (*ptr)=*this;
  return ptr;
}


void mtkGridGL::draw ()
{
  if (state_ == MTK_INACTIVE)
    return;

  // make the diffuse material track the current glColor
  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);

  if (state_ == MTK_ACTIVE) {
    // Draws the Grid lines
    glColor3f(edge_color_.r, edge_color_.g, edge_color_.b);
    glLineWidth(edge_thickness_);

    int i;
    glBegin(GL_LINES);
    for(i=0; i<=num_lines_dir1_; i++) {
      glVertex3f(origin_[0]+dir_[0][0]*i,
		 origin_[1]+dir_[0][1]*i,
		 origin_[2]+dir_[0][2]*i);
      glVertex3f(origin_[0]+dir_[0][0]*i+dir_[1][0]*num_lines_dir2_,
		 origin_[1]+dir_[0][1]*i+dir_[1][1]*num_lines_dir2_,
		 origin_[2]+dir_[0][2]*i+dir_[1][2]*num_lines_dir2_);
    }
    glEnd();

    glBegin(GL_LINES);
    for(i=0; i<=num_lines_dir2_; i++) {
      glVertex3f(origin_[0]+dir_[1][0]*i,
		 origin_[1]+dir_[1][1]*i,
		 origin_[2]+dir_[1][2]*i);
      glVertex3f(origin_[0]+dir_[1][0]*i+dir_[0][0]*num_lines_dir1_,
		 origin_[1]+dir_[1][1]*i+dir_[0][1]*num_lines_dir1_,
		 origin_[2]+dir_[1][2]*i+dir_[0][2]*num_lines_dir1_);
    }
    glEnd();

    // Draws the Grid filled plane
    if((face_opacity_>0.01) && (face_opacity_<0.99)) {
      glPolygonMode(GL_FRONT, GL_FILL);
      glEnable(GL_BLEND);
      glDepthMask(GL_FALSE);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      glColor4f(face_color_.r, face_color_.g, face_color_.b, face_opacity_);
    }
    else
    if(face_opacity_>0.99) {
      glColor3f(face_color_.r, face_color_.g, face_color_.b);
    }
    
    if(face_opacity_>0.01) {
      glBegin(GL_QUADS);
      glVertex3f(origin_[0],origin_[1],origin_[2]);

      glVertex3f(origin_[0]+dir_[0][0]*num_lines_dir1_,
	         origin_[1]+dir_[0][1]*num_lines_dir1_,
	         origin_[2]+dir_[0][2]*num_lines_dir1_);

      glVertex3f(origin_[0]+dir_[0][0]*num_lines_dir1_+
	                    dir_[1][0]*num_lines_dir2_,
	         origin_[1]+dir_[0][1]*num_lines_dir1_+
	                    dir_[1][1]*num_lines_dir2_,
	         origin_[2]+dir_[0][2]*num_lines_dir1_+
	                    dir_[1][2]*num_lines_dir2_);

      glVertex3f(origin_[0]+dir_[1][0]*num_lines_dir2_,
	         origin_[1]+dir_[1][1]*num_lines_dir2_,
	         origin_[2]+dir_[1][2]*num_lines_dir2_);
      glEnd();
    }
    
    if((face_opacity_>0.01) && (face_opacity_<0.99)) {
      glDepthMask(GL_TRUE);
      glDisable(GL_BLEND);
    }
  }
  glDisable(GL_COLOR_MATERIAL);
}
