// -*- mode: C++; -*-

/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef MTK_GUIDE_H
#define MTK_GUIDE_H

#include "gm.h"
#include "mtk_general.h"


typedef unsigned int mtkGuideId;

enum mtkGuideType {MTK_AXIS, MTK_GRID};

enum mtkAxisStyle {MTK_XYZ_FULL, MTK_XY_FULL, MTK_XZ_FULL, MTK_YZ_FULL,
                   MTK_XYZ_HALF, MTK_XY_HALF, MTK_XZ_HALF, MTK_YZ_HALF,
                   MTK_X, MTK_Y, MTK_Z};


/*********************************************************************
 Base class for guide
**********************************************************************/
class mtkGuide {
public:
  mtkGuide ();

  mtkGuideId getId ();
  void setState (mtkState state);
  mtkState getState ();

  virtual mtkGuideType getType () = 0;
  virtual mtkGuide *clone () const = 0;
  virtual void draw () = 0;

protected:
  static mtkGuideId counter_;

  mtkGuideId id_;
  mtkState state_;
};


/*********************************************************************
 Class Axis
**********************************************************************/
class mtkAxis: public mtkGuide {
public:
  mtkAxis();

  mtkGuideType getType ();

  /// Set and get Axis methods
  void setOrigin (gmPoint3 origin);
  gmPoint3 getOrigin();
  void setAxis (gmVector3 axis[3]);
  void getAxis (gmVector3 axis[3]);
  void setAxisStyle (mtkAxisStyle style);
  mtkAxisStyle getAxisStyle ();

  /// Set axis attributes
  void setThickness (unsigned int thickness);
  void setColor (float r, float g, float b);

protected:
  // Position, orientation and length in 3D space
  gmPoint3 origin_;
  gmVector3 axis_[3];

  // Which parts of axis are drawn. It can be XYZ, XY, XZ, YZ, X, Y and Z
  mtkAxisStyle style_;

  // Edge information
  unsigned int axis_thickness_;
  mtkColor axis_color_;
};


/*********************************************************************
 Class Grid
**********************************************************************/
class mtkGrid: public mtkGuide {
public:
  mtkGrid ();

  mtkGuideType getType ();

  /// Set and get Grid methods
  void setOrigin (gmPoint3 origin);
  gmPoint3 getOrigin ();
  void setDir (gmVector3 dir[2]);
  void getDir (gmVector3 dir[2]);
  void setNumLines (int nLines1, int nLines2);
  void getNumLines (int *nLines1, int *nLines2);

  /// Set edge attributes
  void setEdgeThickness (unsigned int thickness);
  void setEdgeColor (float r, float g, float b);

  /// Set face attributes
  void setFaceColor (float r, float g, float b, float a);
  void setFillFace (bool fill_face);

protected:
  // Set position and orientation in 3D space
  gmPoint3 origin_;
  gmVector3 dir_[2];

  int num_lines_dir1_;
  int num_lines_dir2_;

  // Edge information
  unsigned int edge_thickness_;
  mtkColor edge_color_;

  //  information
  mtkColor face_color_;
  float face_opacity_;
  bool fill_face_;
};


/*********************************************************************
 Class AxisGL
**********************************************************************/
class mtkAxisGL: public mtkAxis {
public:
  mtkAxisGL ();

  mtkGuide *clone () const;

  /// Draws the axis
  void draw ();
};


/*********************************************************************
 Class GridGL
**********************************************************************/
class mtkGridGL: public mtkGrid {
public:
  mtkGridGL ();

  mtkGuide *clone () const;

  /// Draws the grid
  void draw ();
};


#endif // MTK_GUIDE_H
