/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <math.h>

#include "mtk_manip.h"


// global manipulator id counter
mtkManipulatorId mtkManipulator::counter_=1;

// records all used element ids
mtkManipulator::mtkIdSet mtkManipulator::global_registry_;


//----------------------------------------------------------------//

mtkManipulator::mtkManipulator (mtkCoreGL *core)
{
  core_=core;
  id_=counter_++;
  state_=MTK_ACTIVE;
}


/// get manipulator id
mtkManipulatorId mtkManipulator::getId ()
{
  return id_;
}


/// set manipulator state
void mtkManipulator::setState (mtkState state)
{
  state_=state;
}


/// get manipulator state
mtkState mtkManipulator::getState ()
{
  return state_;
}


/// register an element id
bool mtkManipulator::registerId (mtkId id)
{
  if (!id)
    return false;
  mtkIdSetIt it=global_registry_.find(id);
  if (it != global_registry_.end())
    return false;
  global_registry_.insert(id);
  registry_.insert(id);
  return true;
}


/// test if given id is registered
bool mtkManipulator::isRegistered (mtkId id)
{
  mtkIdSetIt it=registry_.find(id);
  if (it == registry_.end())
    return false;
  return true;
}


/// unregister id
bool mtkManipulator::unregisterId (mtkId id)
{
  mtkIdSetIt it=registry_.find(id);
  if (it == registry_.end())
    return false;
  registry_.erase(id);
  global_registry_.erase(id);
  return true;
}


//----------------------------------------------------------------//

mtkDefaultBoxManipulator::mtkDefaultBoxManipulator (mtkCoreGL *core) :
  mtkManipulator(core)
{
  translate_=0;
  scale_=0;
  rotate_=0;
  shear_=0;
  translate_data_=scale_data_=rotate_data_=shear_data_=0;
  mode_=MTK_DEFAULT;

  dragger_=core_->createBoxDragger();
  core_->setDraggerState(dragger_, MTK_HIDDEN);
  core_->setBoxCallback(dragger_, &interactCallback_, (void *)this);
}


mtkDefaultBoxManipulator::~mtkDefaultBoxManipulator ()
{
  // unselect elements
  // unregister elements
  // delete dragger
}


/// set manipulator state
void mtkDefaultBoxManipulator::setState (mtkState state)
{
  mtkManipulator::setState(state);
  core_->setDraggerState(dragger_, selected_.size()? state_: MTK_HIDDEN);
}


void mtkDefaultBoxManipulator::setMode (mtkManipulatorMode mode)
{
  mode_=mode;
}


mtkManipulatorMode mtkDefaultBoxManipulator::getMode ()
{
  return mode_;
}


mtkDraggerId mtkDefaultBoxManipulator::getDraggerId ()
{
  return dragger_;
}


/// register an element id
bool mtkDefaultBoxManipulator::registerId (mtkId id)
{
  if (mtkManipulator::registerId(id)) {
    // register selection callback for this id
    if (core_->setElementCallback(id, &selectCallback_, (void *)this))
      return true;
    mtkManipulator::unregisterId(id); // failed: forget id
  }
  return false;
}


/// unregister id
bool mtkDefaultBoxManipulator::unregisterId (mtkId id)
{
  if (mtkManipulator::unregisterId(id)) {
    // unregister selection callback for this id
    core_->setSelection(id, false);
    core_->setElementCallback(id, 0, 0);
    return true;
  }
  return false;
}


/// set callback for translation event
void mtkDefaultBoxManipulator::setTranslateCallback (mtkTranslateCallback *cb,
                                                     void *data)
{
  translate_=cb;
  translate_data_=data;
}


/// set callback for scale event
void mtkDefaultBoxManipulator::setScaleCallback (mtkScaleCallback *cb,
                                                 void *data)
{
  scale_=cb;
  scale_data_=data;
}


/// set callback for rotate event
void mtkDefaultBoxManipulator::setRotateCallback (mtkRotateCallback *cb,
                                                  void *data)
{
  rotate_=cb;
  rotate_data_=data;
}


/// set callback for shear event
void mtkDefaultBoxManipulator::setShearCallback (mtkShearCallback *cb,
                                                 void *data)
{
  shear_=cb;
  shear_data_=data;
}


// wrapper method for selection callback
bool mtkDefaultBoxManipulator::selectCallback_ (list<mtkId> on, list<mtkId> off,
                                               void *data)
{
  return ((mtkDefaultBoxManipulator *)data)->select_(on, off);
}


// actual handler for element selection event
bool mtkDefaultBoxManipulator::select_ (list<mtkId> on, list<mtkId> off)
{
  list<mtkId>::iterator i;
  for (i=on.begin(); i!=on.end(); i++)
    selected_.push_back(*i);
  for (i=off.begin(); i!=off.end(); i++)
    selected_.remove(*i);

  if (selected_.size()) {
    // update dragger
    gmPoint3 min, max;
    core_->calcBoundingBox(selected_, &min, &max);
    core_->setBoxGeometry(dragger_, min,
                          gmVector3(0.0, 0.0, max[2]-min[2]),
                          gmVector3(0.0, max[1]-min[1], 0.0),
                          gmVector3(max[0]-min[0], 0.0, 0.0));
  }

  // update dragger state
  core_->setDraggerState(dragger_, selected_.size()? state_: MTK_HIDDEN);

  return true;
}


// wrapper method for dragger callback
bool mtkDefaultBoxManipulator::interactCallback_ (mtkDraggerId id,
                                                  mtkDraggerPart part,
                                                  gmPoint3 p, gmVector3 v,
                                                  mtkEvent type, void *data)
{
  return ((mtkDefaultBoxManipulator *)data)->interact_(id, part, p, v, type);
}


// actual handler for dragger manipulation event
bool mtkDefaultBoxManipulator::interact_ (mtkDraggerId id,
                                          mtkDraggerPart part,
                                          gmPoint3 p, gmVector3 v,
                                          mtkEvent type)
{
  // unused parameters
  id=id;

  bool update=false;
  if (mode_ == MTK_DEFAULT) {
    if (part>=MTK_BOX_V_0 && part<=MTK_BOX_V_7)
      update=vertex_(part, p, v, type);
    else if (part>=MTK_BOX_E_01 && part<=MTK_BOX_E_67)
      update=edge_(part, p, v, type);
    else
      update=face_(part, p, v, type);
  }
  else {
    if (part>=MTK_BOX_V_0 && part<=MTK_BOX_V_7)
      update=vertex2_(part, p, v, type);
    else if (part>=MTK_BOX_E_01 && part<=MTK_BOX_E_67)
      update=edge2_(part, p, v, type);
    else
      update=face2_(part, p, v, type);
  }
  return update;
}


bool mtkDefaultBoxManipulator::vertex_ (mtkDraggerPart part,
                                        gmPoint3 p, gmVector3 v,
                                        mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmPoint3 c;

    core_->getBoxCenter(dragger_, &c);
    core_->setSnapParallelAxis(dragger_, part, p-c);
  }
  else if (type == MTK_MOVE) {
    if (scale_) {
      double factor;
      gmPoint3 c, vtx;
      gmMatrix4 m;

      core_->getBoxCenter(dragger_, &c);
      core_->getBoxVertex(dragger_, part, &vtx);
      factor=1.0+v[0]/(vtx[0]-c[0]); // the other coords give the same factor

      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.scaling(factor, factor, factor)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyBoxMatrix(dragger_, m);
      (*scale_)(id_, selected_, c, gmVector3(0.0, 0.0, 0.0), factor,
                scale_data_);
      update=true;
    }
  }
  return update;
}


bool mtkDefaultBoxManipulator::edge_ (mtkDraggerPart part,
                                      gmPoint3 p, gmVector3 v,
                                      mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmPoint3 p1, p2;

    switch (part) {
    case MTK_BOX_E_01:
    case MTK_BOX_E_32:
    case MTK_BOX_E_54:
    case MTK_BOX_E_67:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_1, &p2);
      break;
    case MTK_BOX_E_02:
    case MTK_BOX_E_31:
    case MTK_BOX_E_57:
    case MTK_BOX_E_64:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_2, &p2);
      break;
    case MTK_BOX_E_04:
    case MTK_BOX_E_37:
    case MTK_BOX_E_51:
    case MTK_BOX_E_62:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_4, &p2);
      break;
    }
    core_->setSnapParallelPlane(dragger_, part, p2-p1);
  }
  else if (type == MTK_MOVE) {
    if (rotate_) {
      double angle, dummy;
      gmPoint3 center, p1, p2;
      gmVector3 axis, axis2, n1, n2;
      gmMatrix4 m;

      core_->getBoxCenter(dragger_, &center);
      switch (part) {
      case MTK_BOX_E_01:
      case MTK_BOX_E_32:
      case MTK_BOX_E_54:
      case MTK_BOX_E_67:
        core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_1, &p2);
        break;
      case MTK_BOX_E_02:
      case MTK_BOX_E_31:
      case MTK_BOX_E_57:
      case MTK_BOX_E_64:
        core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_2, &p2);
        break;
      case MTK_BOX_E_04:
      case MTK_BOX_E_37:
      case MTK_BOX_E_51:
      case MTK_BOX_E_62:
        core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_4, &p2);
        break;
      }

      axis=(p2-p1).normalize();
      if (!gmCalcPlane(center, center+axis, p, &n1, &dummy))
        return false; /*!*/
      if (!gmCalcPlane(center, center+axis, p+v, &n2, &dummy))
        return false; /*!*/
      axis2=gmCross(n1, n2);
      angle=gmDegrees(asin(axis2.length()))*gmSign(gmDot(axis, axis2));

      m=m.translation(-center[0], -center[1], -center[2]);
      m=m.rotation(angle, axis)*m;
      m=m.translation(center[0], center[1], center[2])*m;
      core_->applyBoxMatrix(dragger_, m);
      (*rotate_)(id_, selected_, center, axis, angle, rotate_data_);
      update=true;
    }
  }
  return update;
}


bool mtkDefaultBoxManipulator::face_ (mtkDraggerPart part,
                                      gmPoint3 p, gmVector3 v,
                                      mtkEvent type)
{
  // unused parameters
  p=p;

  bool update=false;
  if (type == MTK_PICK) {
    double dummy;
    gmPoint3 p1, p2, p3;
    gmVector3 n;

    switch (part) {
    case MTK_BOX_F_0132:
    case MTK_BOX_F_4675:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_1, &p2);
      core_->getBoxVertex(dragger_, MTK_BOX_V_3, &p3);
      break;
    case MTK_BOX_F_0264:
    case MTK_BOX_F_1573:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_2, &p2);
      core_->getBoxVertex(dragger_, MTK_BOX_V_6, &p3);
      break;
    case MTK_BOX_F_0451:
    case MTK_BOX_F_2376:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_4, &p2);
      core_->getBoxVertex(dragger_, MTK_BOX_V_5, &p3);
      break;
    }
    gmCalcPlane(p1, p2, p3, &n, &dummy);
    core_->setSnapParallelPlane(dragger_, part, n);
  }
  else if (type == MTK_MOVE) {
    if (translate_) {
      gmMatrix4 m;

      core_->applyBoxMatrix(dragger_, m.translation(v[0], v[1], v[2]));
      (*translate_)(id_, selected_, v, translate_data_);
      update=true;
    }
  }
  return update;
}


bool mtkDefaultBoxManipulator::vertex2_ (mtkDraggerPart part,
                                         gmPoint3 p, gmVector3 v,
                                         mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmPoint3 c;

    core_->getBoxCenter(dragger_, &c);
    core_->setSnapSphere(dragger_, part, c, (p-c).length());
  }
  else if (type == MTK_MOVE) {
    if (rotate_) {
      double angle;
      gmPoint3 c;
      gmVector3 axis, v1, v2;
      gmMatrix4 m;

      core_->getBoxCenter(dragger_, &c);
      v1=(p-c).normalize();
      v2=(p+v-c).normalize();
      axis=gmCross(v1, v2);
      angle=gmDegrees(acos(gmDot(v1, v2)));

      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.rotation(angle, axis)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyBoxMatrix(dragger_, m);
      (*rotate_)(id_, selected_, c, axis, angle, rotate_data_);
      update=true;
    }
  }
  return update;
}


bool mtkDefaultBoxManipulator::edge2_ (mtkDraggerPart part,
                                       gmPoint3 p, gmVector3 v,
                                       mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmPoint3 p1, p2;

    switch (part) {
    case MTK_BOX_E_01:
    case MTK_BOX_E_32:
    case MTK_BOX_E_54:
    case MTK_BOX_E_67:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_1, &p2);
      break;
    case MTK_BOX_E_02:
    case MTK_BOX_E_31:
    case MTK_BOX_E_57:
    case MTK_BOX_E_64:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_2, &p2);
      break;
    case MTK_BOX_E_04:
    case MTK_BOX_E_37:
    case MTK_BOX_E_51:
    case MTK_BOX_E_62:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_4, &p2);
      break;
    }
    core_->setSnapParallelAxis(dragger_, part, p2-p1);
  }
  else if (type == MTK_MOVE) {
    if (scale_) {
      double dist1, dist2, factor;
      gmPoint3 c, p1, p2;
      gmVector3 n;
      gmMatrix4 m;

      core_->getBoxCenter(dragger_, &c);
      switch (part) {
      case MTK_BOX_E_01:
      case MTK_BOX_E_32:
      case MTK_BOX_E_54:
      case MTK_BOX_E_67:
        core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_1, &p2);
        break;
      case MTK_BOX_E_02:
      case MTK_BOX_E_31:
      case MTK_BOX_E_57:
      case MTK_BOX_E_64:
        core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_2, &p2);
        break;
      case MTK_BOX_E_04:
      case MTK_BOX_E_37:
      case MTK_BOX_E_51:
      case MTK_BOX_E_62:
        core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_4, &p2);
        break;
      }

      n=(p2-p1).normalize();
      dist1=gmSignedDistance(gmPoint3(p[0]-c[0], p[1]-c[1], p[2]-c[2]),
                             n, 0.0);
      if (gmAbs(dist1) < GM_EPSILON)
        return false; /*!*/
      p+=v;
      dist2=gmSignedDistance(gmPoint3(p[0]-c[0], p[1]-c[1], p[2]-c[2]),
                             n, 0.0);
      if (gmAbs(dist2) < GM_EPSILON)
        return false; /*!*/
      factor=dist2/dist1;

      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.scaling(factor, n)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyBoxMatrix(dragger_, m);
      (*scale_)(id_, selected_, c, n, factor, scale_data_);
      update=true;
    }
  }
  return update;
}


bool mtkDefaultBoxManipulator::face2_ (mtkDraggerPart part,
                                       gmPoint3 p, gmVector3 v,
                                       mtkEvent type)
{
  // unused parameters
  p=p;

  bool update=false;
  if (type == MTK_PICK) {
    double dummy;
    gmPoint3 p1, p2, p3;
    gmVector3 n;

    switch (part) {
    case MTK_BOX_F_0132:
    case MTK_BOX_F_4675:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_1, &p2);
      core_->getBoxVertex(dragger_, MTK_BOX_V_3, &p3);
      break;
    case MTK_BOX_F_0264:
    case MTK_BOX_F_1573:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_2, &p2);
      core_->getBoxVertex(dragger_, MTK_BOX_V_6, &p3);
      break;
    case MTK_BOX_F_0451:
    case MTK_BOX_F_2376:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_4, &p2);
      core_->getBoxVertex(dragger_, MTK_BOX_V_5, &p3);
      break;
    }
    gmCalcPlane(p1, p2, p3, &n, &dummy);
    core_->setSnapParallelPlane(dragger_, part, n);
  }
  else if (type == MTK_MOVE) {
    if (shear_) {
      double angle, d, dist;
      gmPoint3 c, p1, p2, p3;
      gmVector3 axis, n;
      gmMatrix4 m;

      core_->getBoxCenter(dragger_, &c);
      switch (part) {
      case MTK_BOX_F_0132:
        core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_1, &p2);
        core_->getBoxVertex(dragger_, MTK_BOX_V_3, &p3);
        break;
      case MTK_BOX_F_0264:
        core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_2, &p2);
        core_->getBoxVertex(dragger_, MTK_BOX_V_6, &p3);
        break;
      case MTK_BOX_F_0451:
        core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_4, &p2);
        core_->getBoxVertex(dragger_, MTK_BOX_V_5, &p3);
        break;
      case MTK_BOX_F_4675:
        core_->getBoxVertex(dragger_, MTK_BOX_V_4, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_6, &p2);
        core_->getBoxVertex(dragger_, MTK_BOX_V_7, &p3);
        break;
      case MTK_BOX_F_1573:
        core_->getBoxVertex(dragger_, MTK_BOX_V_1, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_5, &p2);
        core_->getBoxVertex(dragger_, MTK_BOX_V_7, &p3);
        break;
      case MTK_BOX_F_2376:
        core_->getBoxVertex(dragger_, MTK_BOX_V_2, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_3, &p2);
        core_->getBoxVertex(dragger_, MTK_BOX_V_7, &p3);
        break;
      }
      gmCalcPlane(p1, p2, p3, &n, &d);
      dist=gmAbs(gmSignedDistance(c, n, d));
      gmVector3 ref(0.0, 1.0, 0.0);
      if (gmDot(n, ref) < 0.0)
        n=-n; // angle betwen ref and n should be <= 90 degrees
      else
        v=-v; // swap motion direction
      axis=gmCross(n, ref);
      angle=gmDegrees(asin(axis.length()));
      if (axis.length() < GM_EPSILON) {
        axis=ref;
        angle=0.0;
      }

      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.rotation(angle, axis)*m;
      v=m.transform(v);
      m=m.shearing(v[0]/dist, 0.0, 0.0, 0.0, 0.0, v[2]/dist)*m;
      m=m.rotation(-angle, axis)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyBoxMatrix(dragger_, m);
      (*shear_)(id_, selected_, m, shear_data_);
      update=true;
    }
  }
  return update;
}


//----------------------------------------------------------------//

mtkDefaultTrackballManipulator::mtkDefaultTrackballManipulator (mtkCoreGL *core) :
  mtkManipulator(core)
{
  translate_=0;
  scale_=0;
  rotate_=0;
  translate_data_=scale_data_=rotate_data_=0;

  dragger_=core_->createTrackballDragger();
  core_->setDraggerState(dragger_, MTK_HIDDEN);
  core_->setBallCallback(dragger_, &interactCallback_, (void *)this);
}


mtkDefaultTrackballManipulator::~mtkDefaultTrackballManipulator ()
{
  // unselect elements
  // unregister elements
  // delete dragger
}


/// set manipulator state
void mtkDefaultTrackballManipulator::setState (mtkState state)
{
  mtkManipulator::setState(state);
  core_->setDraggerState(dragger_, selected_.size()? state_: MTK_HIDDEN);
}


mtkDraggerId mtkDefaultTrackballManipulator::getDraggerId ()
{
  return dragger_;
}


/// register an element id
bool mtkDefaultTrackballManipulator::registerId (mtkId id)
{
  if (mtkManipulator::registerId(id)) {
    // register selection callback for this id
    if (core_->setElementCallback(id, &selectCallback_, (void *)this))
      return true;
    mtkManipulator::unregisterId(id); // failed: forget id
  }
  return false;
}


/// unregister id
bool mtkDefaultTrackballManipulator::unregisterId (mtkId id)
{
  if (mtkManipulator::unregisterId(id)) {
    // unregister selection callback for this id
    core_->setSelection(id, false);
    core_->setElementCallback(id, 0, 0);
    return true;
  }
  return false;
}


/// set callback for scale event
void mtkDefaultTrackballManipulator::setTranslateCallback (mtkTranslateCallback *cb,
                                                          void *data)
{
  translate_=cb;
  translate_data_=data;
}


/// set callback for scale event
void mtkDefaultTrackballManipulator::setScaleCallback (mtkScaleCallback *cb,
                                                      void *data)
{
  scale_=cb;
  scale_data_=data;
}


/// set callback for rotate event
void mtkDefaultTrackballManipulator::setRotateCallback (mtkRotateCallback *cb,
                                                       void *data)
{
  rotate_=cb;
  rotate_data_=data;
}


// wrapper method for selection callback
bool mtkDefaultTrackballManipulator::selectCallback_ (list<mtkId> on, list<mtkId> off,
                                                     void *data)
{
  return ((mtkDefaultTrackballManipulator *)data)->select_(on, off);
}


// actual handler for element selection event
bool mtkDefaultTrackballManipulator::select_ (list<mtkId> on, list<mtkId> off)
{
  list<mtkId>::iterator i;
  for (i=on.begin(); i!=on.end(); i++)
    selected_.push_back(*i);
  for (i=off.begin(); i!=off.end(); i++)
    selected_.remove(*i);

  if (selected_.size()) {
    // update dragger
    gmPoint3 center;
    double radius;
    core_->calcBoundingSphere(selected_, &center, &radius);
    core_->setBallGeometry(dragger_, center, radius, 0.5);
  }

  // update dragger state
  core_->setDraggerState(dragger_, selected_.size()? state_: MTK_HIDDEN);

  return true;
}


// wrapper method for dragger callback
bool mtkDefaultTrackballManipulator::interactCallback_ (mtkDraggerId id,
                                                        mtkDraggerPart part,
                                                        gmPoint3 p,
                                                        gmVector3 v,
                                                        mtkEvent type,
                                                        void *data)
{
  return ((mtkDefaultTrackballManipulator *)data)->interact_(id, part, p, v,
                                                             type);
}


// actual handler for dragger manipulation event
bool mtkDefaultTrackballManipulator::interact_ (mtkDraggerId id,
                                                mtkDraggerPart part,
                                                gmPoint3 p, gmVector3 v,
                                                mtkEvent type)
{
  // unused parameters
  id=id;

  bool update=false;
  if (part == MTK_BALL_V)
    update=vertex_(part, p, v, type);
  else if (part == MTK_BALL_E)
    update=edge_(part, p, v, type);
  else if (part>=MTK_BALL_R_0 && part<=MTK_BALL_R_2)
    update=ring_(part, p, v, type);
  else if (part == MTK_BALL_S)
    update=surface_(part, p, v, type);
  return update;
}


bool mtkDefaultTrackballManipulator::vertex_ (mtkDraggerPart part,
                                              gmPoint3 p, gmVector3 v,
                                              mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    double r;
    gmPoint3 c;

    core_->getBallSphere(dragger_, &c, &r);
    core_->setSnapParallelAxis(dragger_, part, p-c);
  }
  else if (type == MTK_MOVE) {
    if (scale_) {
      double r, s;
      gmPoint3 a, c;
      gmMatrix4 m;

      core_->getBallSphere(dragger_, &c, &r);
      core_->getBallArrow(dragger_, &a);
      s=1.0+gmSign(gmDot(v, a-c))*v.length()/gmDistance(a, c);

      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.scaling(s, s, s)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyBallMatrix(dragger_, m);
      (*scale_)(id_, selected_, c, gmVector3(0.0, 0.0, 0.0), s, scale_data_);
      update=true;
    }
  }
  return update;
}


bool mtkDefaultTrackballManipulator::edge_ (mtkDraggerPart part,
                                            gmPoint3 p, gmVector3 v,
                                            mtkEvent type)
{
  // unused parameters
  p=p;

  bool update=false;
  if (type == MTK_PICK) {
    double r;
    gmPoint3 a, c;

    core_->getBallSphere(dragger_, &c, &r);
    core_->getBallArrow(dragger_, &a);
    core_->setSnapParallelAxis(dragger_, part, a-c);
  }
  else if (type == MTK_MOVE) {
    if (translate_) {
      gmMatrix4 m;

      core_->applyBallMatrix(dragger_, m.translation(v[0], v[1], v[2]));
      (*translate_)(id_, selected_, v, translate_data_);
      update=true;
    }
  }
  return update;
}


bool mtkDefaultTrackballManipulator::ring_ (mtkDraggerPart part,
                                            gmPoint3 p, gmVector3 v,
                                            mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmVector3 n;

    core_->getBallRing(dragger_, part, &n);
    core_->setSnapParallelPlane(dragger_, part, n);
  }
  else if (type == MTK_MOVE) {
    if (rotate_) {
      double angle, r;
      gmPoint3 c;
      gmVector3 axis, v1, v2;
      gmMatrix4 m;

      core_->getBallSphere(dragger_, &c, &r);
      v1=(p-c).normalize();
      v2=(p+v-c).normalize();
      core_->getBallRing(dragger_, part, &axis);
      axis*=gmSign(gmDot(axis, gmCross(v1, v2)));
      angle=gmDegrees(acos(gmDot(v1, v2)));

      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.rotation(angle, axis)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyBallMatrix(dragger_, m);
      (*rotate_)(id_, selected_, c, axis, angle, rotate_data_);
      update=true;
    }
  }
  return update;
}


bool mtkDefaultTrackballManipulator::surface_ (mtkDraggerPart part,
                                               gmPoint3 p, gmVector3 v,
                                               mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmPoint3 c;
    double r;

    core_->getBallSphere(dragger_, &c, &r);
    core_->setSnapSphere(dragger_, part, c, r);
  }
  else if (type == MTK_MOVE) {
    if (rotate_) {
      double angle, r;
      gmPoint3 c;
      gmVector3 axis, v1, v2;
      gmMatrix4 m;

      core_->getBallSphere(dragger_, &c, &r);
      v1=(p-c).normalize();
      v2=(p+v-c).normalize();
      axis=gmCross(v1, v2);
      angle=gmDegrees(acos(gmDot(v1, v2)));

      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.rotation(angle, axis)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyBallMatrix(dragger_, m);
      (*rotate_)(id_, selected_, c, axis, angle, rotate_data_);
      update=true;
    }
  }
  return update;
}


//----------------------------------------------------------------//

mtkDefaultJackManipulator::mtkDefaultJackManipulator (mtkCoreGL *core) :
  mtkManipulator(core)
{
  translate_=0;
  scale_=0;
  rotate_=0;
  translate_data_=scale_data_=rotate_data_=0;

  dragger_=core_->createJackDragger();
  core_->setDraggerState(dragger_, MTK_HIDDEN);
  core_->setJackCallback(dragger_, &interactCallback_, (void *)this);
}


mtkDefaultJackManipulator::~mtkDefaultJackManipulator ()
{
  // unselect elements
  // unregister elements
  // delete dragger
}


/// set manipulator state
void mtkDefaultJackManipulator::setState (mtkState state)
{
  mtkManipulator::setState(state);
  core_->setDraggerState(dragger_, selected_.size()? state_: MTK_HIDDEN);
}


mtkDraggerId mtkDefaultJackManipulator::getDraggerId ()
{
  return dragger_;
}


/// register an element id
bool mtkDefaultJackManipulator::registerId (mtkId id)
{
  if (mtkManipulator::registerId(id)) {
    // register selection callback for this id
    if (core_->setElementCallback(id, &selectCallback_, (void *)this))
      return true;
    mtkManipulator::unregisterId(id); // failed: forget id
  }
  return false;
}


/// unregister id
bool mtkDefaultJackManipulator::unregisterId (mtkId id)
{
  if (mtkManipulator::unregisterId(id)) {
    // unregister selection callback for this id
    core_->setSelection(id, false);
    core_->setElementCallback(id, 0, 0);
    return true;
  }
  return false;
}


/// set callback for translation event
void mtkDefaultJackManipulator::setTranslateCallback (mtkTranslateCallback *cb,
                                                      void *data)
{
  translate_=cb;
  translate_data_=data;
}


/// set callback for scale event
void mtkDefaultJackManipulator::setScaleCallback (mtkScaleCallback *cb,
                                                  void *data)
{
  scale_=cb;
  scale_data_=data;
}


/// set callback for rotate event
void mtkDefaultJackManipulator::setRotateCallback (mtkRotateCallback *cb,
                                                   void *data)
{
  rotate_=cb;
  rotate_data_=data;
}


// wrapper method for selection callback
bool mtkDefaultJackManipulator::selectCallback_ (list<mtkId> on,
                                                 list<mtkId> off,
                                                 void *data)
{
  return ((mtkDefaultJackManipulator *)data)->select_(on, off);
}


// actual handler for element selection event
bool mtkDefaultJackManipulator::select_ (list<mtkId> on, list<mtkId> off)
{
  list<mtkId>::iterator i;
  for (i=on.begin(); i!=on.end(); i++)
    selected_.push_back(*i);
  for (i=off.begin(); i!=off.end(); i++)
    selected_.remove(*i);

  if (selected_.size()) {
    // update dragger
    gmPoint3 c, min, max;
    core_->calcBoundingBox(selected_, &min, &max);
    c=gmLerp(0.5, min, max);
    core_->setJackGeometry(dragger_, c,
                           gmVector3(max[0]-min[0], 0.0, 0.0),
                           gmVector3(0.0, max[1]-min[1], 0.0),
                           gmVector3(0.0, 0.0, max[2]-min[2]));
  }

  // update dragger state
  core_->setDraggerState(dragger_, selected_.size()? state_: MTK_HIDDEN);

  return true;
}


// wrapper method for dragger callback
bool mtkDefaultJackManipulator::interactCallback_ (mtkDraggerId id,
                                                   mtkDraggerPart part,
                                                   gmPoint3 p, gmVector3 v,
                                                   mtkEvent type, void *data)
{
  return ((mtkDefaultJackManipulator *)data)->interact_(id, part, p, v, type);
}


// actual handler for dragger manipulation event
bool mtkDefaultJackManipulator::interact_ (mtkDraggerId id,
                                           mtkDraggerPart part,
                                           gmPoint3 p, gmVector3 v,
                                           mtkEvent type)
{
  // unused parameters
  id=id;

  bool update=false;
  if (part>=MTK_JACK_V_0 && part<=MTK_JACK_V_6)
    update=vertex_(part, p, v, type);
  else if (part>=MTK_JACK_E_01 && part<=MTK_JACK_E_06)
    update=edge_(part, p, v, type);
  return update;
}


bool mtkDefaultJackManipulator::vertex_ (mtkDraggerPart part,
                                         gmPoint3 p, gmVector3 v,
                                         mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmPoint3 c, v1, v2, v3;

    core_->getJackVertex(dragger_, MTK_JACK_V_0, &c);
    core_->getJackVertex(dragger_, MTK_JACK_V_1, &v1);
    core_->getJackVertex(dragger_, MTK_JACK_V_3, &v2);
    core_->getJackVertex(dragger_, MTK_JACK_V_5, &v3);
    core_->setSnapNearestParallelAxis(dragger_, part, v1-c, v2-c, v3-c);
  }
  else if (type == MTK_MOVE) {
    if((part==MTK_JACK_V_0) && (translate_)) {
      gmMatrix4 m;

      core_->applyJackMatrix(dragger_, m.translation(v[0], v[1], v[2]));
      (*translate_)(id_, selected_, v, translate_data_);
    }
    else if(rotate_) {
      gmPoint3 c;
      gmVector3 direction(gmAbs(v[0]), gmAbs(v[1]), gmAbs(v[2]));

      direction=direction.normalize();
      core_->getJackVertex(dragger_, MTK_JACK_V_0, &c);

      // If movement is parallel to an axis different from the
      // selected one, does rotation.
      gmVector3 j[6];
      gmPoint3 jv[6];
      double angle=0;

      for (int i=0; i<6; i++) {
        core_->getJackVertex(dragger_, MTK_JACK_V_1+i, &jv[i]);
        j[i]=jv[i]-c;
        j[i][0]=gmAbs(j[i][0]);
        j[i][1]=gmAbs(j[i][1]);
        j[i][2]=gmAbs(j[i][2]);
        j[i]=j[i].normalize();
      }
      // Rotation about Y or Z
      if ((part==MTK_JACK_V_1) || (part==MTK_JACK_V_2)) {
        if (direction==j[2])
          axis_=jv[4]-c;
        else if (direction==j[4])
          axis_=jv[2]-c;
      }
      // Rotation about X or Z
      if ( (part==MTK_JACK_V_3) || (part==MTK_JACK_V_4)) {
        if (direction==j[0])
          axis_=jv[4]-c;
        else if (direction==j[4])
          axis_=jv[0]-c;
      }
      // Rotation about X or Y
      if ((part==MTK_JACK_V_5) || (part==MTK_JACK_V_6)) {
        if (direction==j[0])
          axis_=jv[2]-c;
        else if (direction==j[2])
          axis_=jv[0]-c;
      }
      core_->setSnapParallelPlane(dragger_, part, axis_);

      gmVector3 axis2, n1, n2;
      double dummy;
      core_->getJackVertex(dragger_, MTK_JACK_V_0, &c);
      if (!gmCalcPlane(c, c+axis_, p, &n1, &dummy))
        return false; /*!*/
      if (!gmCalcPlane(c, c+axis_, p+v, &n2, &dummy))
        return false; /*!*/
      axis2=gmCross(n1, n2);
      angle=gmDegrees(asin(axis2.length()))*gmSign(gmDot(axis_, axis2));

      gmMatrix4 m;
      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.rotation(angle, axis_)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyJackMatrix(dragger_, m);
      (*rotate_)(id_, selected_, c, axis_, angle, rotate_data_);
    }
    update=true;
  }
  return update;
}


bool mtkDefaultJackManipulator::edge_ (mtkDraggerPart part,
                                       gmPoint3 p, gmVector3 v,
                                       mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmPoint3 p1, p2;

    switch (part) {
    case MTK_JACK_E_01:
    case MTK_JACK_E_02:
      core_->getJackVertex(dragger_, MTK_JACK_V_0, &p1);
      core_->getJackVertex(dragger_, MTK_JACK_V_1, &p2);
      break;
    case MTK_JACK_E_03:
    case MTK_JACK_E_04:
      core_->getJackVertex(dragger_, MTK_JACK_V_0, &p1);
      core_->getJackVertex(dragger_, MTK_JACK_V_3, &p2);
      break;
    case MTK_JACK_E_05:
    case MTK_JACK_E_06:
      core_->getJackVertex(dragger_, MTK_JACK_V_0, &p1);
      core_->getJackVertex(dragger_, MTK_JACK_V_5, &p2);
      break;
    }
    core_->setSnapParallelAxis(dragger_, part, p2-p1);
  }
  else if (type == MTK_MOVE) {
    if (scale_) {
      double dist1, dist2, factor;
      gmPoint3 c, p1, p2;
      gmVector3 n;
      gmMatrix4 m;

      core_->getJackVertex(dragger_, MTK_JACK_V_0, &c);
      switch (part) {
        case MTK_JACK_E_01:
        case MTK_JACK_E_02:
        core_->getJackVertex(dragger_, MTK_JACK_V_0, &p1);
        core_->getJackVertex(dragger_, MTK_JACK_V_1, &p2);
        break;
        case MTK_JACK_E_03:
        case MTK_JACK_E_04:
        core_->getJackVertex(dragger_, MTK_JACK_V_0, &p1);
        core_->getJackVertex(dragger_, MTK_JACK_V_3, &p2);
        break;
        case MTK_JACK_E_05:
        case MTK_JACK_E_06:
        core_->getJackVertex(dragger_, MTK_JACK_V_0, &p1);
        core_->getJackVertex(dragger_, MTK_JACK_V_5, &p2);
        break;
      }

      n=(p2-p1).normalize();
      dist1=gmSignedDistance(gmPoint3(p[0]-c[0], p[1]-c[1], p[2]-c[2]),
                             n, 0.0);
      if (gmAbs(dist1) < GM_EPSILON)
        return false; /*!*/
      p+=v;
      dist2=gmSignedDistance(gmPoint3(p[0]-c[0], p[1]-c[1], p[2]-c[2]),
                             n, 0.0);
      if (gmAbs(dist2) < GM_EPSILON)
        return false; /*!*/
      factor=dist2/dist1;

      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.scaling(factor, n)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyJackMatrix(dragger_, m);
      (*scale_)(id_, selected_, c, n, factor, scale_data_);
      update=true;
    }
  }
  return update;
}

//----------------------------------------------------------------//

mtkDefaultCursor3DManipulator::mtkDefaultCursor3DManipulator (mtkCoreGL *core) :
  mtkManipulator(core)
{
  translate_=0;
  scale_=0;
  rotate_=0;
  translate_data_=scale_data_=rotate_data_=0;

  dragger_=core_->createCursor3DDragger();
  core_->setDraggerState(dragger_, MTK_HIDDEN);
  core_->setCursor3DCallback(dragger_, &interactCallback_, (void *)this);
}


mtkDefaultCursor3DManipulator::~mtkDefaultCursor3DManipulator ()
{
  // unselect elements
  // unregister elements
  // delete dragger
}


/// set manipulator state
void mtkDefaultCursor3DManipulator::setState (mtkState state)
{
  mtkManipulator::setState(state);
  core_->setDraggerState(dragger_, selected_.size()? state_: MTK_HIDDEN);
}


mtkDraggerId mtkDefaultCursor3DManipulator::getDraggerId ()
{
  return dragger_;
}


/// register an element id
bool mtkDefaultCursor3DManipulator::registerId (mtkId id)
{
  if (mtkManipulator::registerId(id)) {
    // register selection callback for this id
    if (core_->setElementCallback(id, &selectCallback_, (void *)this))
      return true;
    mtkManipulator::unregisterId(id); // failed: forget id
  }
  return false;
}


/// unregister id
bool mtkDefaultCursor3DManipulator::unregisterId (mtkId id)
{
  if (mtkManipulator::unregisterId(id)) {
    // unregister selection callback for this id
    core_->setSelection(id, false);
    core_->setElementCallback(id, 0, 0);
    return true;
  }
  return false;
}


/// set callback for translation event
void mtkDefaultCursor3DManipulator::setTranslateCallback (mtkTranslateCallback *cb,
                                                      void *data)
{
  translate_=cb;
  translate_data_=data;
}


/// set callback for scale event
void mtkDefaultCursor3DManipulator::setScaleCallback (mtkScaleCallback *cb,
                                                  void *data)
{
  scale_=cb;
  scale_data_=data;
}


/// set callback for rotate event
void mtkDefaultCursor3DManipulator::setRotateCallback (mtkRotateCallback *cb,
                                                   void *data)
{
  rotate_=cb;
  rotate_data_=data;
}


// wrapper method for selection callback
bool mtkDefaultCursor3DManipulator::selectCallback_ (list<mtkId> on,
                                                 list<mtkId> off,
                                                 void *data)
{
  return ((mtkDefaultCursor3DManipulator *)data)->select_(on, off);
}


// actual handler for element selection event
bool mtkDefaultCursor3DManipulator::select_ (list<mtkId> on, list<mtkId> off)
{
  list<mtkId>::iterator i;
  for (i=on.begin(); i!=on.end(); i++)
    selected_.push_back(*i);
  for (i=off.begin(); i!=off.end(); i++)
    selected_.remove(*i);

  if (selected_.size()) {
    // update dragger
    gmPoint3 c, min, max;
    core_->calcBoundingBox(selected_, &min, &max);
    c=gmLerp(0.5, min, max);
    core_->setCursor3DGeometry(dragger_, c,
                           gmVector3(max[0]-min[0], 0.0, 0.0),
                           gmVector3(0.0, max[1]-min[1], 0.0),
                           gmVector3(0.0, 0.0, max[2]-min[2]));
  }

  // update dragger state
  core_->setDraggerState(dragger_, selected_.size()? state_: MTK_HIDDEN);

  return true;
}


// wrapper method for dragger callback
bool mtkDefaultCursor3DManipulator::interactCallback_ (mtkDraggerId id,
                                                   mtkDraggerPart part,
                                                   gmPoint3 p, gmVector3 v,
                                                   mtkEvent type, void *data)
{
  return ((mtkDefaultCursor3DManipulator *)data)->interact_(id, part, p, v, type);
}


// actual handler for dragger manipulation event
bool mtkDefaultCursor3DManipulator::interact_ (mtkDraggerId id,
                                           mtkDraggerPart part,
                                           gmPoint3 p, gmVector3 v,
                                           mtkEvent type)
{
  // unused parameters
  id=id;

  bool update=false;
  if (part>=MTK_CURSOR3D_V_0 && part<=MTK_CURSOR3D_V_6)
    update=vertex_(part, p, v, type);
  else if (part>=MTK_CURSOR3D_E_01 && part<=MTK_CURSOR3D_E_06)
    update=edge_(part, p, v, type);
  return update;
}


bool mtkDefaultCursor3DManipulator::vertex_ (mtkDraggerPart part,
                                         gmPoint3 p, gmVector3 v,
                                         mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmPoint3 c, v1, v2, v3;

    core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_0, &c);
    core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_1, &v1);
    core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_3, &v2);
    core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_5, &v3);
    core_->setSnapNearestParallelAxis(dragger_, part, v1-c, v2-c, v3-c);
  }
  else if (type == MTK_MOVE) {
    if((part==MTK_CURSOR3D_V_0) && (translate_)) {
      gmMatrix4 m;

      core_->applyCursor3DMatrix(dragger_, m.translation(v[0], v[1], v[2]));
      (*translate_)(id_, selected_, v, translate_data_);
    }
    else if(rotate_) {
      gmPoint3 c;
      gmVector3 direction(gmAbs(v[0]), gmAbs(v[1]), gmAbs(v[2]));

      direction=direction.normalize();
      core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_0, &c);

      // If movement is parallel to an axis different from the
      // selected one, does rotation.
      gmVector3 j[6];
      gmPoint3 jv[6];
      double angle=0;

      for (int i=0; i<6; i++) {
        core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_1+i, &jv[i]);
        j[i]=jv[i]-c;
        j[i][0]=gmAbs(j[i][0]);
        j[i][1]=gmAbs(j[i][1]);
        j[i][2]=gmAbs(j[i][2]);
        j[i]=j[i].normalize();
      }
      // Rotation about Y or Z
      if ((part==MTK_CURSOR3D_V_1) || (part==MTK_CURSOR3D_V_2)) {
        if (direction==j[2])
          axis_=jv[4]-c;
        else if (direction==j[4])
          axis_=jv[2]-c;
      }
      // Rotation about X or Z
      if ( (part==MTK_CURSOR3D_V_3) || (part==MTK_CURSOR3D_V_4)) {
        if (direction==j[0])
          axis_=jv[4]-c;
        else if (direction==j[4])
          axis_=jv[0]-c;
      }
      // Rotation about X or Y
      if ((part==MTK_CURSOR3D_V_5) || (part==MTK_CURSOR3D_V_6)) {
        if (direction==j[0])
          axis_=jv[2]-c;
        else if (direction==j[2])
          axis_=jv[0]-c;
      }
      core_->setSnapParallelPlane(dragger_, part, axis_);

      gmVector3 axis2, n1, n2;
      double dummy;
      core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_0, &c);
      if (!gmCalcPlane(c, c+axis_, p, &n1, &dummy))
        return false; /*!*/
      if (!gmCalcPlane(c, c+axis_, p+v, &n2, &dummy))
        return false; /*!*/
      axis2=gmCross(n1, n2);
      angle=gmDegrees(asin(axis2.length()))*gmSign(gmDot(axis_, axis2));

      gmMatrix4 m;
      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.rotation(angle, axis_)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyCursor3DMatrix(dragger_, m);
      (*rotate_)(id_, selected_, c, axis_, angle, rotate_data_);
    }
    update=true;
  }
  return update;
}


bool mtkDefaultCursor3DManipulator::edge_ (mtkDraggerPart part,
                                       gmPoint3 p, gmVector3 v,
                                       mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmPoint3 p1, p2;

    switch (part) {
    case MTK_CURSOR3D_E_01:
    case MTK_CURSOR3D_E_02:
      core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_0, &p1);
      core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_1, &p2);
      break;
    case MTK_CURSOR3D_E_03:
    case MTK_CURSOR3D_E_04:
      core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_0, &p1);
      core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_3, &p2);
      break;
    case MTK_CURSOR3D_E_05:
    case MTK_CURSOR3D_E_06:
      core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_0, &p1);
      core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_5, &p2);
      break;
    }
    core_->setSnapParallelAxis(dragger_, part, p2-p1);
  }
  else if (type == MTK_MOVE) {
    if (scale_) {
      double dist1, dist2, factor;
      gmPoint3 c, p1, p2;
      gmVector3 n;
      gmMatrix4 m;

      core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_0, &c);
      switch (part) {
        case MTK_CURSOR3D_E_01:
        case MTK_CURSOR3D_E_02:
        core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_0, &p1);
        core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_1, &p2);
        break;
        case MTK_CURSOR3D_E_03:
        case MTK_CURSOR3D_E_04:
        core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_0, &p1);
        core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_3, &p2);
        break;
        case MTK_CURSOR3D_E_05:
        case MTK_CURSOR3D_E_06:
        core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_0, &p1);
        core_->getCursor3DVertex(dragger_, MTK_CURSOR3D_V_5, &p2);
        break;
      }

      n=(p2-p1).normalize();
      dist1=gmSignedDistance(gmPoint3(p[0]-c[0], p[1]-c[1], p[2]-c[2]),
                             n, 0.0);
      if (gmAbs(dist1) < GM_EPSILON)
        return false; /*!*/
      p+=v;
      dist2=gmSignedDistance(gmPoint3(p[0]-c[0], p[1]-c[1], p[2]-c[2]),
                             n, 0.0);
      if (gmAbs(dist2) < GM_EPSILON)
        return false; /*!*/
      factor=dist2/dist1;

      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.scaling(factor, n)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyCursor3DMatrix(dragger_, m);
      (*scale_)(id_, selected_, c, n, factor, scale_data_);
      update=true;
    }
  }
  return update;
}


//----------------------------------------------------------------//

mtkDefaultTrackBoxManipulator::mtkDefaultTrackBoxManipulator (mtkCoreGL
*core) :
  mtkManipulator(core)
{
  translate_=0;
  scale_=0;
  rotate_=0;
  shear_=0;
  translate_data_=scale_data_=rotate_data_=shear_data_=0;

  dragger_=core_->createBoxDragger();
  dragger2_=core_->createTrackballDragger();
  core_->setDraggerState(dragger_, MTK_HIDDEN);
  core_->setDraggerState(dragger2_, MTK_HIDDEN);
  core_->setBoxCallback(dragger_, &interactCallback_, (void *)this);
  core_->setBallCallback(dragger2_, &interactCallback2_, (void *)this);
  core_->setBallVisibility(dragger2_,MTK_BALL_V,false);
  core_->setBallVisibility(dragger2_,MTK_BALL_E,false);
}


mtkDefaultTrackBoxManipulator::~mtkDefaultTrackBoxManipulator ()
{
  // unselect elements
  // unregister elements
  // delete dragger
}


/// set manipulator state
void mtkDefaultTrackBoxManipulator::setState (mtkState state)
{
  mtkManipulator::setState(state);
  core_->setDraggerState(dragger_, selected_.size()? state_: MTK_HIDDEN);
  core_->setDraggerState(dragger2_, selected_.size()? state_: MTK_HIDDEN);
}



mtkDraggerId mtkDefaultTrackBoxManipulator::getDraggerId ()
{

  // Returns only the box dragger
  return dragger_;
}


/// register an element id
bool mtkDefaultTrackBoxManipulator::registerId (mtkId id)
{
  if (mtkManipulator::registerId(id)) {
    // register selection callback for this id
    if (core_->setElementCallback(id, &selectCallback_, (void *)this))
      return true;
    mtkManipulator::unregisterId(id); // failed: forget id
  }
  return false;
}


/// unregister id
bool mtkDefaultTrackBoxManipulator::unregisterId (mtkId id)
{
  if (mtkManipulator::unregisterId(id)) {
    // unregister selection callback for this id
    core_->setSelection(id, false);
    core_->setElementCallback(id, 0, 0);
    return true;
  }
  return false;
}


/// set callback for translation event
void mtkDefaultTrackBoxManipulator::setTranslateCallback
(mtkTranslateCallback *cb, void *data)
{
  translate_=cb;
  translate_data_=data;
}


/// set callback for scale event
void mtkDefaultTrackBoxManipulator::setScaleCallback (mtkScaleCallback
*cb, void *data)
{
  scale_=cb;
  scale_data_=data;
}


/// set callback for rotate event
void mtkDefaultTrackBoxManipulator::setRotateCallback (mtkRotateCallback
*cb, void *data)
{
  rotate_=cb;
  rotate_data_=data;
}


/// set callback for shear event
void mtkDefaultTrackBoxManipulator::setShearCallback (mtkShearCallback
*cb, void *data)
{
  shear_=cb;
  shear_data_=data;
}


// wrapper method for selection callback
bool mtkDefaultTrackBoxManipulator::selectCallback_ (list<mtkId> on,
list<mtkId> off, void *data)
{
  return ((mtkDefaultTrackBoxManipulator *)data)->select_(on, off);
}


// actual handler for element selection event
bool mtkDefaultTrackBoxManipulator::select_ (list<mtkId> on, list<mtkId>
off)
{
  list<mtkId>::iterator i;
  for (i=on.begin(); i!=on.end(); i++)
    selected_.push_back(*i);
  for (i=off.begin(); i!=off.end(); i++)
    selected_.remove(*i);

  if (selected_.size()) {
    // update dragger
    double boxsize,radius;
    gmPoint3 min, max, center, corner;
    core_->calcBoundingBox(selected_, &min, &max);
    center[0] = (min[0]+max[0])/2;
    center[1] = (min[1]+max[1])/2;
    center[2] = (min[2]+max[2])/2;
    boxsize = max[0]-min[0];
    if(max[1]-min[1] > boxsize)
       boxsize = max[1]-min[1];
    if(max[2]-min[2] > boxsize)
       boxsize = max[2]-min[2];
    radius = boxsize * 0.64;
    corner[0] = center[0] - boxsize/2;
    corner[1] = center[1] - boxsize/2;
    corner[2] = center[2] - boxsize/2;

    core_->setBoxGeometry(dragger_, corner,
                          gmVector3(0.0, 0.0, boxsize),
                          gmVector3(0.0, boxsize, 0.0),
                          gmVector3(boxsize, 0.0, 0.0));
    core_->setBoxFaceColor(dragger_,1.0,0.5,0.5,0.6);
    core_->setBallGeometry(dragger2_, center, radius,0.0);
  }

  // update dragger state
  core_->setDraggerState(dragger_, selected_.size()? state_: MTK_HIDDEN);
  core_->setDraggerState(dragger2_, selected_.size()? state_: MTK_HIDDEN);

  return true;
}




// wrapper method for dragger callback
bool mtkDefaultTrackBoxManipulator::interactCallback_ (mtkDraggerId id,
                                                  mtkDraggerPart part,
                                                  gmPoint3 p, gmVector3 v,
                                                  mtkEvent type, void *data)
{
  return ((mtkDefaultTrackBoxManipulator *)data)->interact_(id, part, p,
v, type);
}


// actual handler for dragger manipulation event
bool mtkDefaultTrackBoxManipulator::interact_ (mtkDraggerId id,
                                          mtkDraggerPart part,
                                          gmPoint3 p, gmVector3 v,
                                          mtkEvent type)
{
  // unused parameters
  id=id;

  bool update=false;
  if (part>=MTK_BOX_V_0 && part<=MTK_BOX_V_7)
    update=vertex_(part, p, v, type);
  else if (part>=MTK_BOX_E_01 && part<=MTK_BOX_E_67)
    update=edge_(part, p, v, type);
  else
    update=face_(part, p, v, type);
  return update;
}


// wrapper method for dragger callback
bool mtkDefaultTrackBoxManipulator::interactCallback2_ (mtkDraggerId id,
                                                        mtkDraggerPart part,
                                                        gmPoint3 p,
                                                        gmVector3 v,
                                                        mtkEvent type,
                                                        void *data)
{
  return ((mtkDefaultTrackBoxManipulator *)data)->interact2_(id, part, p, 
v, type);
}


// actual handler for dragger manipulation event
bool mtkDefaultTrackBoxManipulator::interact2_ (mtkDraggerId id,
                                                mtkDraggerPart part,
                                                gmPoint3 p, gmVector3 v,
                                                mtkEvent type)
{
  // unused parameters
  id=id;

  bool update=false;
  
  if (part>=MTK_BALL_R_0 && part<=MTK_BALL_R_2)
    update=ring_(part, p, v, type);
  else if (part == MTK_BALL_S)
    update=surface_(part, p, v, type);
  return update;
}



bool mtkDefaultTrackBoxManipulator::vertex_ (mtkDraggerPart part,
                                        gmPoint3 p, gmVector3 v,
                                        mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmPoint3 c;

    core_->getBoxCenter(dragger_, &c);
    core_->setSnapParallelAxis(dragger_, part, p-c);
  }
  else if (type == MTK_MOVE) {
    if (scale_) {
      double factor;
      gmPoint3 c, vtx;
      gmMatrix4 m;

      core_->getBoxCenter(dragger_, &c);
      core_->getBoxVertex(dragger_, part, &vtx);
      factor=1.0+v[0]/(vtx[0]-c[0]); // the other coords give the same factor

      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.scaling(factor, factor, factor)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyBoxMatrix(dragger_, m);
      
      // Updates Trackball
      core_->applyBallMatrix(dragger2_,m);
      
      (*scale_)(id_, selected_, c, gmVector3(0.0, 0.0, 0.0), factor,
                scale_data_);
      update=true;
    }
  }
  return update;
}


bool mtkDefaultTrackBoxManipulator::edge_ (mtkDraggerPart part,
                                      gmPoint3 p, gmVector3 v,
                                      mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmPoint3 p1, p2;

    switch (part) {
    case MTK_BOX_E_01:
    case MTK_BOX_E_32:
    case MTK_BOX_E_54:
    case MTK_BOX_E_67:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_1, &p2);
      break;
    case MTK_BOX_E_02:
    case MTK_BOX_E_31:
    case MTK_BOX_E_57:
    case MTK_BOX_E_64:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_2, &p2);
      break;
    case MTK_BOX_E_04:
    case MTK_BOX_E_37:
    case MTK_BOX_E_51:
    case MTK_BOX_E_62:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_4, &p2);
      break;
    }
    core_->setSnapParallelPlane(dragger_, part, p2-p1);
  }
  else if (type == MTK_MOVE) {
    if (rotate_) {
      double angle, dummy;
      gmPoint3 center, p1, p2;
      gmVector3 axis, axis2, n1, n2;
      gmMatrix4 m;

      core_->getBoxCenter(dragger_, &center);
      switch (part) {
      case MTK_BOX_E_01:
      case MTK_BOX_E_32:
      case MTK_BOX_E_54:
      case MTK_BOX_E_67:
        core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_1, &p2);
        break;
      case MTK_BOX_E_02:
      case MTK_BOX_E_31:
      case MTK_BOX_E_57:
      case MTK_BOX_E_64:
        core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_2, &p2);
        break;
      case MTK_BOX_E_04:
      case MTK_BOX_E_37:
      case MTK_BOX_E_51:
      case MTK_BOX_E_62:
        core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
        core_->getBoxVertex(dragger_, MTK_BOX_V_4, &p2);
        break;
      }

      axis=(p2-p1).normalize();
      if (!gmCalcPlane(center, center+axis, p, &n1, &dummy))
        return false; /*!*/
      if (!gmCalcPlane(center, center+axis, p+v, &n2, &dummy))
        return false; /*!*/
      axis2=gmCross(n1, n2);
      angle=gmDegrees(asin(axis2.length()))*gmSign(gmDot(axis, axis2));

      m=m.translation(-center[0], -center[1], -center[2]);
      m=m.rotation(angle, axis)*m;
      m=m.translation(center[0], center[1], center[2])*m;
      core_->applyBoxMatrix(dragger_, m);


      // Just for testing purposes, let's not rotate the object
      // (*rotate_)(id_, selected_, center, axis, angle, rotate_data_);
      update=true;
    }
  }
  return update;
}


bool mtkDefaultTrackBoxManipulator::face_ (mtkDraggerPart part,
                                      gmPoint3 p, gmVector3 v,
                                      mtkEvent type)
{
  // unused parameters
  p=p;

  bool update=false;
  if (type == MTK_PICK) {
    double dummy;
    gmPoint3 p1, p2, p3;
    gmVector3 n;

    switch (part) {
    case MTK_BOX_F_0132:
    case MTK_BOX_F_4675:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_1, &p2);
      core_->getBoxVertex(dragger_, MTK_BOX_V_3, &p3);
      break;
    case MTK_BOX_F_0264:
    case MTK_BOX_F_1573:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_2, &p2);
      core_->getBoxVertex(dragger_, MTK_BOX_V_6, &p3);
      break;
    case MTK_BOX_F_0451:
    case MTK_BOX_F_2376:
      core_->getBoxVertex(dragger_, MTK_BOX_V_0, &p1);
      core_->getBoxVertex(dragger_, MTK_BOX_V_4, &p2);
      core_->getBoxVertex(dragger_, MTK_BOX_V_5, &p3);
      break;
    }
    gmCalcPlane(p1, p2, p3, &n, &dummy);
    core_->setSnapParallelPlane(dragger_, part, n);
  }
  else if (type == MTK_MOVE) {
    if (translate_) {
      gmMatrix4 m;

      core_->applyBoxMatrix(dragger_, m.translation(v[0], v[1], v[2]));

      // Updates ball
      core_->applyBallMatrix(dragger2_, m.translation(v[0], v[1], v[2]));
     
 
      (*translate_)(id_, selected_, v, translate_data_);
      update=true;
    }
  }
  return update;
}


bool mtkDefaultTrackBoxManipulator::ring_ (mtkDraggerPart part,
                                            gmPoint3 p, gmVector3 v,
                                            mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmVector3 n;

    core_->getBallRing(dragger2_, part, &n);
    core_->setSnapParallelPlane(dragger2_, part, n);
  }
  else if (type == MTK_MOVE) {
    if (rotate_) {
      double angle, r;
      gmPoint3 c;
      gmVector3 axis, v1, v2;
      gmMatrix4 m;

      core_->getBallSphere(dragger2_, &c, &r);
      v1=(p-c).normalize();
      v2=(p+v-c).normalize();
      core_->getBallRing(dragger2_, part, &axis);
      axis*=gmSign(gmDot(axis, gmCross(v1, v2)));
      angle=gmDegrees(acos(gmDot(v1, v2)));

      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.rotation(angle, axis)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyBallMatrix(dragger2_, m);
      (*rotate_)(id_, selected_, c, axis, angle, rotate_data_);
      update=true;
    }
  }
  return update;
}


bool mtkDefaultTrackBoxManipulator::surface_ (mtkDraggerPart part,
                                               gmPoint3 p, gmVector3 v,
                                               mtkEvent type)
{
  bool update=false;
  if (type == MTK_PICK) {
    gmPoint3 c;
    double r;

    core_->getBallSphere(dragger2_, &c, &r);
    core_->setSnapSphere(dragger2_, part, c, r);
  }
  else if (type == MTK_MOVE) {
    if (rotate_) {
      double angle, r;
      gmPoint3 c;
      gmVector3 axis, v1, v2;
      gmMatrix4 m;

      core_->getBallSphere(dragger2_, &c, &r);
      v1=(p-c).normalize();
      v2=(p+v-c).normalize();
      axis=gmCross(v1, v2);
      angle=gmDegrees(acos(gmDot(v1, v2)));

      m=m.translation(-c[0], -c[1], -c[2]);
      m=m.rotation(angle, axis)*m;
      m=m.translation(c[0], c[1], c[2])*m;
      core_->applyBallMatrix(dragger2_, m);
      (*rotate_)(id_, selected_, c, axis, angle, rotate_data_);
      update=true;
    }
  }
  return update;
}

