// -*- mode: C++; -*-

/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef MTK_MANIP_H
#define MTK_MANIP_H

#include <set.h>
#include <list.h>

#include "mtk.h"


/// manipulator identification
typedef unsigned int mtkManipulatorId;

enum mtkManipulatorMode {MTK_DEFAULT, MTK_ALTERNATE};

/// callback for translation
typedef void mtkTranslateCallback (mtkManipulatorId id, list<mtkId> id_list,
                                   gmVector3 v, void *data);

/// callback for scaling
typedef void mtkScaleCallback (mtkManipulatorId id, list<mtkId> id_list,
                               gmPoint3 p, gmVector3 v, double factor,
                               void *data);

/// callback for rotation
typedef void mtkRotateCallback (mtkManipulatorId id, list<mtkId> id_list,
                                gmPoint3 p, gmVector3 v, double angle,
                                void *data);

/// callback for shearing
typedef void mtkShearCallback (mtkManipulatorId id, list<mtkId> id_list,
                               gmMatrix4 m, void *data);


//----------------------------------------------------------------//

/** This is the generic manipulator base class.
 */
class mtkManipulator {
public:
  /// binds manipulator to the mtkew
  mtkManipulator (mtkCoreGL *core);

  /// get manipulator id
  mtkManipulatorId getId ();
  /// set manipulator state
  void setState (mtkState state);
  /// get manipulator state
  mtkState getState ();

  /// register an element id
  bool registerId (mtkId id);
  /// test if given id is registered
  bool isRegistered (mtkId id);
  /// unregister id
  bool unregisterId (mtkId id);

  // some useful typedefs
  typedef set<mtkId, less<mtkId> > mtkIdSet;
  typedef mtkIdSet::iterator mtkIdSetIt;

protected:
  // global manipulator id counter
  static mtkManipulatorId counter_;
  // records all used element ids
  static mtkIdSet global_registry_;

  // points to the mtkew
  mtkCoreGL *core_;
  // manipulator id
  mtkManipulatorId id_;
  // current manipulator state
  mtkState state_;
  // records elements handled by this manipulator
  mtkIdSet registry_;
  // records selected elements
  list<mtkId> selected_;
};


//----------------------------------------------------------------//

/** This is a specialized {\bf Bounding Box} class with default behamtkor.
 */
class mtkDefaultBoxManipulator : public mtkManipulator {
public:
  /// binds manipulator to the mtkew
  mtkDefaultBoxManipulator (mtkCoreGL *core);
  ~mtkDefaultBoxManipulator ();

  /// set manipulator state
  void setState (mtkState state);
  void setMode (mtkManipulatorMode mode);
  mtkManipulatorMode getMode ();
  mtkDraggerId getDraggerId ();

  /// register an element id
  bool registerId (mtkId id);
  /// unregister id
  bool unregisterId (mtkId id);

  /// set callback for translation event
  void setTranslateCallback (mtkTranslateCallback *cb, void *data);
  /// set callback for scale event
  void setScaleCallback (mtkScaleCallback *cb, void *data);
  /// set callback for rotate event
  void setRotateCallback (mtkRotateCallback *cb, void *data);
  /// set callback for shear event
  void setShearCallback (mtkShearCallback *cb, void *data);

private:
  // wrapper method for selection callback
  static bool selectCallback_ (list<mtkId> on, list<mtkId> off, void *data);
  // actual handler for element selection event
  bool select_ (list<mtkId> on, list<mtkId> off);

  // wrapper method for dragger callback
  static bool interactCallback_ (mtkDraggerId id, mtkDraggerPart part,
                                 gmPoint3 p, gmVector3 v,
                                 mtkEvent type, void *data);

  // actual handler for dragger manipulation event
  bool interact_ (mtkDraggerId id, mtkDraggerPart part,
                  gmPoint3 p, gmVector3 v, mtkEvent type);

  bool vertex_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);
  bool edge_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);
  bool face_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);

  bool vertex2_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);
  bool edge2_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);
  bool face2_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);

  mtkManipulatorMode mode_;
  mtkDraggerId dragger_;

  // external callback holders
  mtkTranslateCallback *translate_;
  void *translate_data_;
  mtkScaleCallback *scale_;
  void *scale_data_;
  mtkRotateCallback *rotate_;
  void *rotate_data_;
  mtkShearCallback *shear_;
  void *shear_data_;
};


/** This is a specialized {\bf Trackball} class with default behamtkor.
 */
class mtkDefaultTrackballManipulator : public mtkManipulator {
public:
  /// binds manipulator to the mtkew
  mtkDefaultTrackballManipulator (mtkCoreGL *core);
  ~mtkDefaultTrackballManipulator ();

  /// set manipulator state
  void setState (mtkState state);
  mtkDraggerId getDraggerId ();

  /// register an element id
  bool registerId (mtkId id);
  /// unregister id
  bool unregisterId (mtkId id);

  /// set callback for translation event
  void setTranslateCallback (mtkTranslateCallback *cb, void *data);
  /// set callback for scale event
  void setScaleCallback (mtkScaleCallback *cb, void *data);
  /// set callback for rotate event
  void setRotateCallback (mtkRotateCallback *cb, void *data);

private:
  // wrapper method for selection callback
  static bool selectCallback_ (list<mtkId> on, list<mtkId> off, void *data);
  // actual handler for element selection event
  bool select_ (list<mtkId> on, list<mtkId> off);

  // wrapper method for dragger callback
  static bool interactCallback_ (mtkDraggerId id, mtkDraggerPart part,
                                 gmPoint3 p, gmVector3 v,
				 mtkEvent type, void *data);

  // actual handler for dragger manipulation event
  bool interact_ (mtkDraggerId id, mtkDraggerPart part,
                  gmPoint3 p, gmVector3 v, mtkEvent type);

  bool vertex_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);
  bool edge_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);
  bool ring_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);
  bool surface_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);

  // dragger
  mtkDraggerId dragger_;

  // external callback holders
  mtkTranslateCallback *translate_;
  void *translate_data_;
  mtkScaleCallback *scale_;
  void *scale_data_;
  mtkRotateCallback *rotate_;
  void *rotate_data_;
};


/** This is a specialized {\bf Jack} class with default behamtkor.
 */
class mtkDefaultJackManipulator : public mtkManipulator {
public:
  /// binds manipulator to the mtkew
  mtkDefaultJackManipulator (mtkCoreGL *core);
  ~mtkDefaultJackManipulator ();

  /// set manipulator state
  void setState (mtkState state);
  mtkDraggerId getDraggerId ();

  /// register an element id
  bool registerId (mtkId id);
  /// unregister id
  bool unregisterId (mtkId id);

  /// set callback for translation event
  void setTranslateCallback (mtkTranslateCallback *cb, void *data);
  /// set callback for scale event
  void setScaleCallback (mtkScaleCallback *cb, void *data);
  /// set callback for rotate event
  void setRotateCallback (mtkRotateCallback *cb, void *data);

private:
  // wrapper method for selection callback
  static bool selectCallback_ (list<mtkId> on, list<mtkId> off, void *data);
  // actual handler for element selection event
  bool select_ (list<mtkId> on, list<mtkId> off);

  // wrapper method for dragger callback
  static bool interactCallback_ (mtkDraggerId id, mtkDraggerPart part,
                                 gmPoint3 p, gmVector3 v,
				 mtkEvent type, void *data);

  // actual handler for dragger manipulation event
  bool interact_ (mtkDraggerId id, mtkDraggerPart part,
                  gmPoint3 p, gmVector3 v, mtkEvent type);

  bool vertex_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);
  bool edge_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);

  // dragger
  mtkDraggerId dragger_;

  gmVector3 axis_;

  // external callback holders
  mtkTranslateCallback *translate_;
  void *translate_data_;
  mtkScaleCallback *scale_;
  void *scale_data_;
  mtkRotateCallback *rotate_;
  void *rotate_data_;
};


/** This is a specialized {\bf Cursor3D} class with default behamtkor.
 */
class mtkDefaultCursor3DManipulator : public mtkManipulator {
public:
  /// binds manipulator to the mtkew
  mtkDefaultCursor3DManipulator (mtkCoreGL *core);
  ~mtkDefaultCursor3DManipulator ();

  /// set manipulator state
  void setState (mtkState state);
  mtkDraggerId getDraggerId ();

  /// register an element id
  bool registerId (mtkId id);
  /// unregister id
  bool unregisterId (mtkId id);

  /// set callback for translation event
  void setTranslateCallback (mtkTranslateCallback *cb, void *data);
  /// set callback for scale event
  void setScaleCallback (mtkScaleCallback *cb, void *data);
  /// set callback for rotate event
  void setRotateCallback (mtkRotateCallback *cb, void *data);

private:
  // wrapper method for selection callback
  static bool selectCallback_ (list<mtkId> on, list<mtkId> off, void *data);
  // actual handler for element selection event
  bool select_ (list<mtkId> on, list<mtkId> off);

  // wrapper method for dragger callback
  static bool interactCallback_ (mtkDraggerId id, mtkDraggerPart part,
                                 gmPoint3 p, gmVector3 v,
				 mtkEvent type, void *data);

  // actual handler for dragger manipulation event
  bool interact_ (mtkDraggerId id, mtkDraggerPart part,
                  gmPoint3 p, gmVector3 v, mtkEvent type);

  bool vertex_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);
  bool edge_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);

  // dragger
  mtkDraggerId dragger_;

  gmVector3 axis_;

  // external callback holders
  mtkTranslateCallback *translate_;
  void *translate_data_;
  mtkScaleCallback *scale_;
  void *scale_data_;
  mtkRotateCallback *rotate_;
  void *rotate_data_;
};


//----------------------------------------------------------------//

/** This is a new "trackbox" manipulator by Leonardo "Bubble".
    Last modification: 24/01/2000
 */
class mtkDefaultTrackBoxManipulator : public mtkManipulator {
public:
  /// binds manipulator to the mtkew
  mtkDefaultTrackBoxManipulator (mtkCoreGL *core);
  ~mtkDefaultTrackBoxManipulator ();

  /// set manipulator state
  void setState (mtkState state);
  mtkDraggerId getDraggerId ();

  /// register an element id
  bool registerId (mtkId id);
  /// unregister id
  bool unregisterId (mtkId id);

  /// set callback for translation event
  void setTranslateCallback (mtkTranslateCallback *cb, void *data);
  /// set callback for scale event
  void setScaleCallback (mtkScaleCallback *cb, void *data);
  /// set callback for rotate event
  void setRotateCallback (mtkRotateCallback *cb, void *data);
  /// set callback for shear event
  void setShearCallback (mtkShearCallback *cb, void *data);

private:
  // wrapper method for selection callback
  static bool selectCallback_ (list<mtkId> on, list<mtkId> off, void *data);
  // actual handler for element selection event
  bool select_ (list<mtkId> on, list<mtkId> off);

  // wrapper method for dragger callback
  static bool interactCallback_ (mtkDraggerId id, mtkDraggerPart part,
                                 gmPoint3 p, gmVector3 v,
                                 mtkEvent type, void *data);

  static bool interactCallback2_ (mtkDraggerId id, mtkDraggerPart part,
                                 gmPoint3 p, gmVector3 v,
                                 mtkEvent type, void *data);

  // actual handler for dragger manipulation event
  bool interact_ (mtkDraggerId id, mtkDraggerPart part,
                  gmPoint3 p, gmVector3 v, mtkEvent type);

  bool interact2_ (mtkDraggerId id, mtkDraggerPart part,
                  gmPoint3 p, gmVector3 v, mtkEvent type);

  bool vertex_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);
  bool edge_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);
  bool face_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);
  bool ring_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);
  bool surface_ (mtkDraggerPart part, gmPoint3 p, gmVector3 v, mtkEvent type);

  // draggers
  mtkDraggerId dragger_;
  mtkDraggerId dragger2_;

  // external callback holders
  mtkTranslateCallback *translate_;
  void *translate_data_;
  mtkScaleCallback *scale_;
  void *scale_data_;
  mtkRotateCallback *rotate_;
  void *rotate_data_;
  mtkShearCallback *shear_;
  void *shear_data_;
};


#endif // MTK_MANIP_H
