/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/glu.h>

#include "mtk_selection.h"


/*********************************************************************
 Selection Implementation
**********************************************************************/

mtkSelection::mtkSelection ()
{
  pick_mode_=MTK_PICK_NEAREST;
  selection_mode_=MTK_SEL_EXCLUSIVE;
  width_=height_=5;
}


/// Defines the type of picking
void mtkSelection::setPickMode (mtkPickMode mode)
{
  pick_mode_=mode;
}


mtkPickMode mtkSelection::getPickMode ()
{
  return pick_mode_;
}


/// Defines the picking area, in pixels
void mtkSelection::setPickArea (unsigned int width, unsigned int height)
{
  width_=width;
  height_=height;
}


/// Defines selection mode
void mtkSelection::setSelectionMode (mtkSelectionMode mode)
{
  selection_mode_=mode;
}


mtkSelectionMode mtkSelection::getSelectionMode ()
{
  return selection_mode_;
}


/// Selects an element givens its identifier
void mtkSelection::select (mtkId id)
{
  if (!id)
    return;
  selection_set_.insert(id);
}


/// Unselects an element givens its identifier
void mtkSelection::unselect (mtkId id)
{
  if (!id)
    return;
  selection_set_.erase(id);
}


// Verifies if an element is selected, given its identifier
bool mtkSelection::isSelected (mtkId id)
{
  mtkIdSetIt i=selection_set_.find(id);
  if (i == selection_set_.end())
    return false;
  return true;
}


// Unselects all selected elements
void mtkSelection::unselectAll ()
{
  selection_set_.erase(selection_set_.begin(), selection_set_.end());
}


// Changes the state of an element, from selected to unselected and mtkce versa
void mtkSelection::toggle (mtkId id)
{
  if(isSelected(id))
    unselect(id);
  else
    select(id);
}


// Returns a list of all selected elements
list<mtkId> mtkSelection::getAll ()
{
  list<mtkId> ret;
  for (mtkIdSetIt it=selection_set_.begin(); it!=selection_set_.end(); it++)
    ret.push_back(*it);
  return ret;
}


//----------------------------------------------------------------//

mtkSelectionDataGL::mtkSelectionDataGL ()
{
  z=0;
  type=MTK_NONE;
  id=part=0;
  selected=false;
}

mtkSelectionDataGL& mtkSelectionDataGL::operator= (const mtkSelectionDataGL& data)
{
  z=data.z;
  type=data.type;
  id=data.id;
  part=data.part;
  selected=data.selected;

  return *this;
}

bool operator< (const mtkSelectionDataGL& d1, const mtkSelectionDataGL& d2)
{
  if (d1.type<d2.type || d1.id<d2.id || d1.part<d2.part)
    return true;
  return false;
}


void mtkSelectionGL::startPicking(int x, int y)
{
  GLint mtkewport[4];
  glGetIntegerv(GL_VIEWPORT, mtkewport);

  glSelectBuffer(MTK_SEL_BUFFER_SIZE, buffer_);
  glRenderMode(GL_SELECT);
  glInitNames();

  GLfloat proj_matrix[16];
  glMatrixMode(GL_PROJECTION);
  glGetFloatv(GL_PROJECTION_MATRIX, proj_matrix);
  glPushMatrix();

  // create a picking region near pick location, with size width x height
  glLoadIdentity();
  gluPickMatrix((GLdouble)x, (GLdouble)(mtkewport[3]-y), width_, height_,
                mtkewport);
  glMultMatrixf(proj_matrix);

  // revert matrix mode
  glMatrixMode(GL_MODELVIEW);
}


typedef set<mtkSelectionDataGL, less<mtkSelectionDataGL> > DataSet;


// After "initPicking" and before "endPicking" there must be the
// display list and dragger "draw" methods
list<mtkSelectionDataGL> mtkSelectionGL::endPicking()
{
  // end OpenGL picking
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glFlush();
  GLint hits=glRenderMode(GL_RENDER);

  mtkSelectionDataGL tmp;
  list<mtkSelectionDataGL> changed;

  // test if just empty space was selected
  if (!hits) {
    if (selection_mode_ == MTK_SEL_EXCLUSIVE) {
      // unselect everything
      for (mtkIdSetIt it=selection_set_.begin(); it!=selection_set_.end();
           it++) {
        tmp.z=0;
        tmp.type=MTK_ELEMENT;
        tmp.id=(*it);
        tmp.part=0;
        tmp.selected=false;
        changed.push_back(tmp);
      }
      unselectAll();
    }
    return changed;
  }

  DataSet picked;
  DataSet::iterator pit;
  mtkSelectionDataGL nearest_elem, nearest_aux;
  GLuint cnt, *ptr=buffer_;
  bool first_elem=true, first_aux=true;
  for (GLint i=0; i<hits; i++) {
    cnt=*ptr++; // number of names for this fragment
    tmp.z=*ptr++; // minimum z for this fragment
    ptr++; // ignore maximum z
    if (cnt < 3) {
      // not enough names; skip fragment
      ptr+=cnt;
      continue;
    }
    tmp.type=*ptr++;
    tmp.id=*ptr++;
    tmp.part=*ptr++;
    tmp.selected=true;

    // store both the nearest element and auxiliary object fragments
    if (tmp.type == MTK_ELEMENT) {
      if (first_elem || tmp.z<nearest_elem.z) {
        first_elem=false;
        nearest_elem=tmp;
      }
    }
    else if (first_aux || tmp.z<nearest_aux.z) {
      first_aux=false;
      nearest_aux=tmp;
    }

    // test if fragment has not been picked
    if (picked.find(tmp) == picked.end())
      picked.insert(tmp);
  }

  // test if the nearest fragment belongs to an auxiliary object
  if (!first_aux && (first_elem || nearest_aux.z<=nearest_elem.z)) {
    // handle a dragger
    if (!first_aux && nearest_aux.type==MTK_DRAGGER) {
      mtkSelectionDataGL smallest=nearest_aux;
      // test if there are also picked elements
      if (!first_elem) {
        // pick smallest dragger part still mtksible
        for (pit=picked.begin(); pit!=picked.end(); pit++)
          if ((*pit).type==MTK_DRAGGER && (*pit).id==smallest.id &&
              ((*pit).part&0xf0)<(smallest.part&0xf0) &&
              (*pit).z<=nearest_elem.z)
            smallest=*pit;
      }
      else {
        // just pick the smallest dragger part
        for (pit=picked.begin(); pit!=picked.end(); pit++)
          if ((*pit).type==MTK_DRAGGER && (*pit).id==smallest.id &&
              ((*pit).part&0xf0)<(smallest.part&0xf0))
            smallest=*pit;
      }
      changed.push_back(smallest);
      return changed;
    }
    //#endif
    // handle other auxiliary objects
    changed.push_back(nearest_aux);
    return changed;
  }

  // handle display list elements
  if (pick_mode_ == MTK_PICK_NEAREST) {
    switch (selection_mode_) {
      case MTK_SEL_EXCLUSIVE:
        // test if the nearest element is not selected
        if (selection_set_.find(nearest_elem.id) == selection_set_.end()) {
          // unselect everything
          for (mtkIdSetIt it=selection_set_.begin(); it!=selection_set_.end();
               it++) {
            tmp.z=0;
            tmp.type=MTK_ELEMENT;
            tmp.id=(*it);
            tmp.part=0;
            tmp.selected=false;
            changed.push_back(tmp);
          }
          unselectAll();

          // now select only the nearest
          select(nearest_elem.id);
          changed.push_back(nearest_elem);
        }
        break;

      case MTK_SEL_INCLUSIVE:
        // test if the nearest element is not selected
        if (selection_set_.find(nearest_elem.id) == selection_set_.end()) {
          select(nearest_elem.id);
          changed.push_back(nearest_elem);
        }
        break;

      case MTK_SEL_TOGGLE:
        // test if the nearest element is not selected
        if (selection_set_.find(nearest_elem.id) == selection_set_.end())
          select(nearest_elem.id);
        else {
          unselect(nearest_elem.id);
          nearest_elem.selected=false;
        }
        changed.push_back(nearest_elem);
        break;
    }
  }
  else if (pick_mode_ == MTK_PICK_ALL) {
    // remove any draggers left in the picked set
    switch (selection_mode_) {
      case MTK_SEL_EXCLUSIVE:
        break;

      case MTK_SEL_INCLUSIVE:
        break;

      case MTK_SEL_TOGGLE:
        break;
    }
  }

  return changed;
}
