// -*- mode: C++; -*-

/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef MTK_SELECTION_H
#define MTK_SELECTION_H

#include <set.h>
#include <list.h>

#include <GL/gl.h>

#include "mtk_general.h"


#define MTK_SEL_BUFFER_SIZE 256

enum mtkPickMode {MTK_PICK_NEAREST, MTK_PICK_ALL, MTK_PICK_STEP};

enum mtkSelectionMode {MTK_SEL_EXCLUSIVE, MTK_SEL_INCLUSIVE, MTK_SEL_TOGGLE};


/*********************************************************************
 Class Selection
**********************************************************************/
class mtkSelection {
public:
  mtkSelection ();

  /// Defines the type of picking
  void setPickMode (mtkPickMode mode);
  mtkPickMode getPickMode ();
  /// Defines the picking area, in pixels
  void setPickArea (unsigned int width, unsigned int height);

  /// Defines selection mode
  void setSelectionMode (mtkSelectionMode mode);
  mtkSelectionMode getSelectionMode ();
  /// Selects an element givens its identifier
  void select (mtkId id);
  /// Unselects an element givens its identifier
  void unselect (mtkId id);
  // Unselects all selected elements
  void unselectAll ();
  // Changes the state of an element,
  // from selected to unselected and mtkce versa
  void toggle (mtkId id);
  // Verifies if an element is selected, given its identifier
  bool isSelected (mtkId id);
  // Returns a list with the identifier of each selected element
  list<mtkId> getAll ();

protected:
  // types used internally
  typedef set<mtkId, less<mtkId> > mtkIdSet;
  typedef mtkIdSet::iterator mtkIdSetIt;

  /* Defines how picking is made:
     PICK_NEAREST - picks only the most near object under the cursor
     PICK_ALL   - picks all objects under the cursor
     PICK_STEP  - steps through all objects under the cursor,
                  one at a time */
  mtkPickMode pick_mode_;

  /* Defines how elements are selected:
     SEL_EXCLUSIVE - selects only one element at a time,
                  clearing the selection of other elements
     SEL_TOGGLE - if an element is not selected,
                  it becomes selected, and mtkce versa
     SEL_INCLUSIVE    - only selects new elements; the other elements are
                  not changed */
  mtkSelectionMode selection_mode_;

  // Width and height defines the area of picking in pixel
  unsigned int width_;
  unsigned int height_;

  // Data about each picked element
   mtkIdSet selection_set_;
};


/**=========================================================
   Derivates Class mtkDisplayList to include OpenGL functions
 ===========================================================
*/

class mtkSelectionDataGL {
public:
  mtkSelectionDataGL ();
  mtkSelectionDataGL& operator= (const mtkSelectionDataGL& data);
  friend bool operator< (const mtkSelectionDataGL& d1,
                         const mtkSelectionDataGL& d2);

  GLuint z;
  mtkType type;
  mtkId id, part;
  bool selected;
};



class mtkSelectionGL : public mtkSelection {
public:
  /// Pick elements on the x,y cursor position in the display
  void startPicking (int x, int y);
  // Process elements picked by OpenGL
  list<mtkSelectionDataGL> endPicking ();

private:
  GLuint buffer_[MTK_SEL_BUFFER_SIZE];
};


#endif // MTK_SELECTION_H
